# Fix_GeometryErrors.py [Python 3.6]
# Fixes geometry errors detected by arcpy.EnableNetworkTopology_un
#
# ERRORTYPE   ERRORTEXT
# 20          Self-intersecting line
# 21          Duplicate vertices

print("Fix_GeometryErrors.py")
print("Initializing...")

import arcpy
import arcpy.da
import os
from Migrate_helper import MigrateHelper

class GeometryHelper(object):
    # Constants
    _sErrorFC = "LineErrors"
    _sTargetFC = "PipelineLine"

    # Private fields
    __sErrorPath = None
    __sTargetPath = None

    def FixGeometryErrors(self, bReadOnly):
        print("Running...")
        mh = MigrateHelper()
        sTargetGDB = mh._sWorkingDir + "/" + mh._sDBConnection
        sID = mh.GetUNID();
        sErrorFC = "UN_" + sID + "_" + self._sErrorFC
        self.__sErrorPath = sTargetGDB + "/" + sErrorFC
        self.__sTargetPath = sTargetGDB + "/" + self._sTargetFC

        # Disable editor tracking
        print("   Disabling editor tracking...")
        mh.SetEditorTracking(self.__sTargetPath, False);

        # Query error features
        print("   Querying errors...")
        Fields = ["ORIGINGLOBALID", "ERRORTYPE"]
        sWhere = "errortype IN (20, 21)"
        RepairList = []
        Errors = {}
        with arcpy.da.SearchCursor(self.__sErrorPath, Fields, sWhere) as cursor:
            for row in cursor:
                sGID = row[0]
                iErr = row[1]
                RepairList.append(sGID)
                Errors[sGID] = iErr
        if len(RepairList) == 0:
            print("No errors found.")
            return

        # Repair features
        print("   Repairing features...")
        self.__RepairFeatures(RepairList, Errors, bReadOnly);

        # Enable editor tracking
        print("   Enabling editor tracking...")
        mh.SetEditorTracking(self.__sTargetPath, True);

        print("Done.")
        return

    def __RepairFeatures(self, RepairList, Errors, bReadOnly):
        # Build query
        sWhere = "globalid IN ( "
        bFirst = True
        for sGID in RepairList:
            if bFirst:
                bFirst = False
            else:
                sWhere += ", "
            sWhere += "'" + sGID + "'"
        sWhere += " )"

        # Start edit session
        if not bReadOnly:
            sWS = os.path.dirname(self.__sTargetPath)
            edit = arcpy.da.Editor(sWS)
            edit.startEditing(False, False)
            edit.startOperation()

        # Update features
        iCount = 0
        iUpdated = 0
        Fields = ["SHAPE@", "GLOBALID"]
        if bReadOnly:
            cursor = arcpy.da.SearchCursor(self.__sTargetPath, Fields, sWhere)
        else:
            cursor = arcpy.da.UpdateCursor(self.__sTargetPath, Fields, sWhere)
        for row in cursor:
            iCount += 1
            apGeomOld = row[0]
            sGID = row[1]
            iErr = Errors[sGID]
            apGeomNew = self.__RepairGeometry(apGeomOld, sGID, iErr)
            if apGeomNew is None:
                continue
            iUpdated += 1
            if bReadOnly:
                continue
            row[0] = apGeomNew
            cursor.updateRow(row)
        del cursor

        # Finish up
        if not bReadOnly:
            edit.stopOperation()
            edit.stopEditing(True)
        print("Updates: " + str(iUpdated) + " of " + str(iCount))
        return

    def __RepairGeometry(self, apGeomOld, sGID, iErr):

        # Generalize
        bChange = False
        sOldJSON = apGeomOld.JSON
        iNumPtOld = apGeomOld.pointCount
        dLenOld = apGeomOld.length
        apGeomNew = apGeomOld.generalize(0.02)
        sNewJSON = apGeomNew.JSON
        iNumPtNew = apGeomNew.pointCount
        dLenNew = apGeomNew.length
        if iNumPtOld != iNumPtNew:
            bChange = True
            print("   " + sGID + "(err " + str(iErr) + "): Vertex change [" + str(iNumPtOld) + " to " + str(iNumPtNew) + "]")
        elif dLenOld != dLenNew:
            bChange = True
            print("   " + sGID + "(err " + str(iErr) + "): Length change [" + str(dLenOld) + " to " + str(dLenNew) + "]")
        if iErr == 21:
            if bChange:
                return apGeomNew
            else:
                return None
            
        # Repair intersection
        print("   Attempting to repair self-intersection: " + sGID)
        Points = apGeomNew.getPart(0)
        bInt = self.__RepairIntersection(Points)
        if bInt:
            print("      Overlap detected.")
        array = arcpy.Array([Points])
        sr = apGeomOld.spatialReference
        apGeomNew = arcpy.Polyline(array, sr)
        return apGeomNew

    def __RepairIntersection(self, Points):

        # Find adjacent overlapping segments
        iPos = -1
        iCount = Points.count
        for i in range(iCount - 2):
            pt0 = Points[i]
            pt1 = Points[i + 1]
            pt2 = Points[i + 2]
            pl0 = arcpy.Polyline(arcpy.Array([pt0, pt1]))
            pl1 = arcpy.Polyline(arcpy.Array([pt1, pt2]))
            # arcpy.Geometry.overlaps method doesn't work!
            pl2 = pl0.intersect(pl1, 2)
            if pl2 is None:
                continue # This should never happen
            bOverlaps = (pl2.length > 0)
            if bOverlaps:
                iPos = i
                break
        if iPos == -1:
            return False
        Points.remove(iPos)
        return True

print("Ready.")
#if __name__ == "__main__":
#    gh = GeometryHelper()
#    gh.FixGeometryErrors(True)
