# Fix_PointErrors.py [Python 3.6]
# Fixes stacked junction features
#
# ERRORTYPE   ERRORTEXT
# 25          Stacked point features

# If two point features are stacked, there will be two errors:
#   A -> B
#   B -> A
# If three point features are stacked, there will be six errors:
#   A -> B
#   B -> A
#   C -> B
#   B -> C
#   C -> A
#   A -> C

print("Fix_PointErrors.py")
print("Initializing...")

import arcpy
import arcpy.da
from Migrate_helper import MigrateHelper

class PointHelper(object):
    # Constants
    _sErrorFC = "PointErrors"
    _sJunctionFC = "PipelineJunction"
    _iJunctionFC = 9
    _iPipelineDeviceID = 6 # UtilityNetwork_nnn_FeatureSourceID
    _iConnectionPoint = 21
    _iTee = 7

    # Private fields
    __AssetGroups = None
    __AssetTypes = None
    __DeleteList = None

    def FixPointErrors(self, bReadOnly):
        print("Running...")
        mh = MigrateHelper()
        sTargetGDB = mh._sWorkingDir + "/" + mh._sDBConnection
        sID = mh.GetUNID();
        sErrorFC = "UN_" + sID + "_" + self._sErrorFC
        sErrorPath = sTargetGDB + "/" + sErrorFC
        sJunctionPath = sTargetGDB + "/" + self._sJunctionFC

        # Query error features
        print("Querying errors...")
        Fields = ["ORIGINFEATURESOURCEID", "ORIGINGLOBALID", "DESTINATIONFEATURESOURCEID", "DESTINATIONGLOBALID"]
        sWhere = "errortype = 25 AND originfeaturesourceid = " + str(self._iJunctionFC)
        ErrorRows = []
        with arcpy.da.SearchCursor(sErrorPath, Fields, sWhere) as cursor:
            for row in cursor:
                ErrorRows.append(row)
        if len(ErrorRows) == 0:
            print("No errors found.")
            return

        # Query junction asset groups
        print("Querying asset groups...")
        self.__AssetGroups = {}
        self.__AssetTypes = {}
        self.__DeleteList = {}
        Fields = ["GLOBALID", "ASSETGROUP", "ASSETTYPE"]
        sWhere = "globalid IN ( "
        bFirst = True
        for row in ErrorRows:
            sGID = row[1]
            if bFirst:
                bFirst = False
            else:
                sWhere += ", "
            sWhere += "'" + sGID + "'"
        sWhere += " )"
        with arcpy.da.SearchCursor(sJunctionPath, Fields, sWhere) as cursor:
            for row in cursor:
                sGID = row[0]
                iGroup = row[1]
                iType = row[2]
                self.__AssetGroups[sGID] = iGroup
                self.__AssetTypes[sGID] = iType
                self.__DeleteList[sGID] = False

        # Compare error features
        print("Comparing error features...")
        for row in ErrorRows:
            iOriginFC = row[0]
            sOriginGID = row[1]
            iDestFC = row[2]
            sDestGID = row[3]
            self.__CompareStackedPoints(iOriginFC, sOriginGID, iDestFC, sDestGID)

        # Build list of junctions to delete
        print("JUNCTIONS TO DELETE:")
        iCount = 0
        sWhere = "globalid IN ( "
        bFirst = True
        for sGID in self.__DeleteList:
            bDelete = self.__DeleteList[sGID]
            if not bDelete:
                continue
            if bFirst:
                bFirst = False
            else:
                sWhere += ", "
            sWhere += "'" + sGID + "'"
            print("   " + sGID + ": " + str(self.__AssetGroups[sGID]))
            iCount += 1
        sWhere += " )"
        print("Total = " + str(iCount))
        if bReadOnly:
            print("Done.")
            return

        # Delete junctions
        print("Deleting junctions...")
        Fields = ["GLOBALID"]
        edit = arcpy.da.Editor(sTargetGDB)
        edit.startEditing(False, False)
        edit.startOperation()
        with arcpy.da.UpdateCursor(sJunctionPath, Fields, sWhere) as cursor:
            for row in cursor:
                cursor.deleteRow()
        edit.stopOperation()
        edit.stopEditing(True)

        print("Done.")
        return

    def __CompareStackedPoints(self, iFC1, sGID1, iFC2, sGID2):
        # Compare two stacked point features and determine which one should be deleted
        
        # If one feature is a device, the other will be a junction and will be deleted
        if iFC1 == self._iPipelineDeviceID:
            self.__DeleteList[sGID2] = True
            return
        elif iFC2 == self._iPipelineDeviceID:
            self.__DeleteList[sGID1] = True
            return

        # If at least one feature is already marked for deletion, continue
        bDelA = False
        if sGID1 in self.__DeleteList:
            bDelA = self.__DeleteList[sGID1]
        bDelB = False
        if sGID2 in self.__DeleteList:
            bDelB = self.__DeleteList[sGID2]
        if bDelA or bDelB:
            return

        # Compare asset groups; if they are the same, then A wins
        iGroupA = self.__AssetGroups[sGID1]
        iGroupB = self.__AssetGroups[sGID2]
        if iGroupA == iGroupB and iGroupA != self._iConnectionPoint:
            self.__DeleteList[sGID1] = False
            self.__DeleteList[sGID2] = True
            return

        # A connection point always loses
        if iGroupA == self._iConnectionPoint and iGroupB != self._iConnectionPoint:
            self.__DeleteList[sGID1] = True
            self.__DeleteList[sGID2] = False
            return
        elif iGroupB == self._iConnectionPoint and iGroupA != self._iConnectionPoint:
            self.__DeleteList[sGID2] = True
            self.__DeleteList[sGID1] = False
            return

        # A riser connection point loses to a system connection point
        # [This will happen when a meter is on a farm gate:
        # as a result, the meter will lose its connection]
        if iGroupA == self._iConnectionPoint and iGroupB == self._iConnectionPoint:
            iTypeA = self.__AssetTypes[sGID1]
            iTypeB = self.__AssetTypes[sGID2]
            if iTypeA == iTypeB:
                self.__DeleteList[sGID2] = True
                self.__DeleteList[sGID1] = False
                return
            elif iTypeA > iTypeB:
                self.__DeleteList[sGID2] = True
                self.__DeleteList[sGID1] = False
                return
            else:
                self.__DeleteList[sGID1] = True
                self.__DeleteList[sGID2] = False
                return

        # A tee always wins
        if iGroupA == self._iTee:
            self.__DeleteList[sGID1] = False
            self.__DeleteList[sGID2] = True
            return

        # Otherwise, the lower asset group value wins
        # [This should not happen]
        if iGroupA < iGroupB:
            self.__DeleteList[sGID1] = False
            self.__DeleteList[sGID2] = True
        else:
            self.__DeleteList[sGID1] = True
            self.__DeleteList[sGID2] = False
        return

print("Ready.")
#if __name__ == "__main__":
#    ph = PointHelper()
#    ph.FixPointErrors(True)
