# Migrate_device.py [Python 3.6]

# Feature classes migrated [-> PipelineDevice]: 
#   TownBorderStation [POC Valves]
#   GasValve
#   RegStation/Valve relationship records [not necessarily containment]
#   ReliefValve
#   ControllableFitting

print("Migrate_device.py")
print("Initalizing...")

import arcpy
from datetime import datetime
import traceback
import uuid
from Migrate_helper import MigrateHelper

class MigrateDevice(object):
    # Constants
    _bSkipProposed = True

    # Private fields
    __helper = None
    __sSourceGDB = ""
    __sTargetGDB = ""
    __sTempGDB = arcpy.env.scratchGDB
    __sTargetFC = ""
    __sTargetRelTab = ""
    __sValveInspectionTab = ""
    __sLineFC = ""
    __cursor = None

    def __init__(self):
        self.__helper = MigrateHelper()
        self.__sSourceGDB = MigrateHelper._sSourceGDB
        self.__sTargetGDB = MigrateHelper._sWorkingDir + "/" + MigrateHelper._sTargetGDBName
        self.__sTargetFC = self.__sTargetGDB + "/PipelineDevice"
        self.__sTargetRelTab = self.__sTargetGDB + "/UNSGPipelineAssembly_PipelineDevice"
        self.__sValveInspectionTab = self.__sTargetGDB + "/ValveInspection"
        self.__sLineFC = self.__sTargetGDB + "/PipelineLine"
        return

    def DoIt(self, bDeleteOnly = False):
        print("Running...")
        dtStart = datetime.now()

        # Populate gas valves
        with arcpy.da.Editor(self.__sTargetGDB) as edit:
            self.__MigrateGasValve()

        # Populate valve relationships
        with arcpy.da.Editor(self.__sTargetGDB) as edit:
            self.__MigrateGasValveRelationships()

        # Populate relief valves
        with arcpy.da.Editor(self.__sTargetGDB) as edit:
            self.__MigrateReliefValve()

        # Populate controllable fittings
        with arcpy.da.Editor(self.__sTargetGDB) as edit:
            self.__MigrateControllableFitting()

        # Calculate terminalconfiguration field
        arcpy.CalculateField_management(self.__sTargetFC, "terminalconfiguration", "'Default'")

        dtEnd = datetime.now()
        print("Done.")
        print("Elapsed time: " + str(dtEnd - dtStart))
        return

    def __MigrateGasValve(self):
        sFCName = "GasValve"
        print("Migrating " + sFCName + "...")

        # Create InsertCursor
        out_fields = ["ASSETGROUP", "ASSETTYPE", "CREATIONDATE", "CREATOR", 
                      "LASTUPDATE", "UPDATEDBY", "DEVICESTATUS", "DESIGNTYPE",
                      "INSULATORDEVICE", "INSTALLATIONDATE", "INSERVICEDATE", "ASSETID", 
                      "LOCATIONDESCRIPTION", "LIFECYCLESTATUS", "MAOPRECORD", "DIAMETER",
                      "ROTATION", "MATERIAL", "OPERATINGCLASSIFICATION", "WORKORDERID", 
                      "VALVEUSE", "ANSIPRESSURERATING", "BONDEDINDICATOR", "GROUNDLEVELINDICATOR", 
                      "STORES_STOCK_NUMBER", "SYSTEM_NAME", "SEQUENTIAL_FILE_NUMBER", "CHART_DATE", 
                      "LS_MAINLINE", "SERIES", "REPORT", "SHAPE@", "LEGACY_OID",
                      "GLOBALID"
                      ]
        self.__cursor = arcpy.da.InsertCursor(self.__sTargetFC, out_fields)

        # Get projected shapes
        print("   Getting projected shapes...")
        ProjectedShapes = self.__helper.GetProjectedShapes(sFCName)

        # Get connected lines
        print("   Getting connected lines...")
        ConnectedLines, LineGeoms = self.__helper.GetConnectedLines(sFCName)

        # Create SearchCursor
        print("   Processing rows...")
        sSourceFC = self.__sSourceGDB + "/" + sFCName
        in_fields = ["DATECREATED", "CREATIONUSER", "DATEMODIFIED", "LASTUSER", 
                     "PRESENTPOSITION", "VALVETYPE", "INSULATEDINDICATOR", "DATE_INSTALLED", 
                     "INSERVICEDATE", "FACILITYID", "LOCATIONDESCRIPTION", "WORKFUNCTION", 
                     "MAOP", "VALVEDIAMETER", "SYMBOLROTATION", "MATERIAL", 
                     "OPERATINGCLASSIFICATION", "WORKORDERID", "VALVEUSE", "ANSIPRESSURERATING", 
                     "BONDEDINDICATOR", "GROUNDLEVELINDICATOR", "STORES_STOCK_NUMBER", "SYSTEM_NAME", 
                     "SEQUENTIAL_FILE_NUMBER", "CHART_DATE", "LS_MAINLINE", "SERIES", 
                     "REPORT", "GLOBALID", "OBJECTID"
                     ]
        iPortOne = 8
        iPortTwo = 9
        Devices = set()
        Terminals = {}
        iCount = 0
        with arcpy.da.SearchCursor(sSourceFC, in_fields) as cursorIn:
            for in_row in cursorIn:
                iWorkFunction = in_row[11]
                if self._bSkipProposed and iWorkFunction == 1:
                    continue
                sValveGID = in_row[29]
                if sValveGID not in ProjectedShapes:
                    continue
                shape = ProjectedShapes[sValveGID]

                # Process row
                bEmergency = self.__ProcessValveRow(in_row, shape)
                iCount += 1
                if iCount % 1000 == 0:
                    print("   GasValve row " + str(iCount) + "...")
                if not bEmergency:
                    continue

                # Determine terminal assignments
                if sValveGID not in ConnectedLines:
                    continue # This should never happen
                LineRecs = ConnectedLines[sValveGID]
                for LineRec in LineRecs:
                    sLineGID = LineRec[0]
                    if sLineGID not in Terminals:
                        Terminals[sLineGID] = []
                    if sValveGID not in Devices:
                        Devices.add(sValveGID)
                        Terminals[sLineGID].append([shape, iPortOne])
                    else:
                        Terminals[sLineGID].append([shape, iPortTwo])

        del self.__cursor
        del ProjectedShapes
        del ConnectedLines
        if len(Terminals) == 0:
            return

        # Update terminal assignments
        print("   Updating terminals...")
        Fields = ["FROMDEVICETERMINAL", "TODEVICETERMINAL", "GLOBALID"]
        with arcpy.da.UpdateCursor(self.__sLineFC, Fields) as cursor:
            for row in cursor:
                sLineGID = row[2]
                if sLineGID not in Terminals:
                    continue
                Entries = Terminals[sLineGID]
                if sLineGID not in LineGeoms:
                    continue # This should never happen
                shape = LineGeoms[sLineGID]
                for entry in Entries:
                    geomValve = entry[0]
                    iPort = entry[1]
                    iFromOrTo = self.__helper.FromOrTwo(shape, geomValve)
                    row[iFromOrTo] = iPort
                cursor.updateRow(row)
        del LineGeoms
        del Terminals
        return

    def __ProcessValveRow(self, in_row, shape):
        bEmergency = True
        iAssetGroup = 3 # "Controllable Valve"
        iAssetType = 483 # "Emergency"
        dtCreated = in_row[0]
        sCreationUser = in_row[1]
        dtModified = in_row[2]
        sLastUser = in_row[3]

        # Domain: UPDM_DEVICESTATUS
        iDeviceStatus = None
        iPresentPosition = in_row[4] # 0 = Closed, 1 = Open
        if iPresentPosition == 0:
            iDeviceStatus = 2
        elif iPresentPosition == 1:
            iDeviceStatus = 1

        # Domain: UPDM_ValveControllableType
        iDesignType = None
        sValveType = in_row[5]
        if sValveType == "BAL":
            iDesignType = 1 # "Ball"
        elif sValveType == "CHK":
            # Terminals will need to be assigned manually
            bEmergency = False
            iAssetGroup = 4 # "Flow Valve"
            iAssetType = 261 # "Check"
        elif sValveType == "EFV":
            bEmergency = False
            iAssetGroup = 1 # "Excess Flow Valve"
            iAssetType = 1 # "Automatic Reset"
        elif sValveType == "GAT":
            iDesignType = 3 # "Gate"
        elif sValveType == "MSV":
            iDesignType = 101 # "Manual Shutoff"
        elif sValveType == "PLG":
            iDesignType = 6 # "Plug"
        elif sValveType == "UNK":
            iDesignType = 0 # "Unknown"

        # Domain: UPDM_INSULATORDEVICE
        iInsulated = None
        sInsulated = in_row[6]
        if sInsulated == "No":
            iInsulated = 1
        elif sInsulated == "Yes":
            iInsulated = 2;

        dtInstalled = in_row[7]
        dtInService = in_row[8]
        sFacilityID = in_row[9]
        sLocation = in_row[10]

        # Domain: UPDM_LifeCycleStatus
        iLifeCycle = 3 # "In Service"
        iWorkFunction = in_row[11]
        if iWorkFunction == 1:
            iLifeCycle = 1 # "Proposed"

        dMAOP = in_row[12]

        # Domain: UPDM_EFVDiameter, UPDM_ValveDiameter
        dDiameter = None
        sDiameter = in_row[13]
        if sDiameter == "UNK":
            dDiameter = 0
        else:
            dDiameter = self.__helper.StringToFloat(sDiameter)
        
        dSymbolRotation = in_row[14]
                
        # Domain: UPDM_ASTMComponentMaterial
        sMaterial = None
        sMat = in_row[15]
        if sMat == "ST":
            sMaterial = "M" # "Steel"
        elif sMat == "PL":
            sMaterial = "X" # "Other"
        elif sMat == "UNK":
            sMaterial = "UN" # "Unknown"

        sClassification = in_row[16]
        sWorkOrderID = in_row[17]
        sValveUse = in_row[18]
        sRating = in_row[19]

        # Domain: UPDM_YesNoIndicator
        iBonded = None
        sBonded = in_row[20]
        if sBonded == "Yes":
            iBonded = 1 # "Yes"
        elif sBonded == "No":
            iBonded = 2 # No

        sGroundLevel = in_row[21]
        sSSN = in_row[22]
        sSystem = in_row[23]
        sSFN = in_row[24]
        dtChart = in_row[25]

        iMainline = None
        sML = in_row[26]
        if sML == "Yes":
            iMainline = 1
        elif sML == "No":
            iMainline = 2
        sSeries = in_row[27]
        sReport = in_row[28]
        sGlobalID = in_row[29]
        iObjectID = in_row[30]

        out_row = [iAssetGroup, iAssetType, dtCreated, sCreationUser, 
                    dtModified, sLastUser, iDeviceStatus, iDesignType, 
                    iInsulated, dtInstalled, dtInService, sFacilityID,
                    sLocation, iLifeCycle, dMAOP, dDiameter,
                    dSymbolRotation, sMaterial, sClassification, sWorkOrderID,
                    sValveUse, sRating, iBonded, sGroundLevel,
                    sSSN, sSystem, sSFN, dtChart,
                    iMainline, sSeries, sReport, shape, iObjectID,
                    sGlobalID
                    ]
        self.__cursor.insertRow(out_row)
        return bEmergency

    def __MigrateGasValveRelationships(self):
        print("Migrating valve relationships...")

        # Get valve global IDs
        print("   Building valve lookup tables...")
        SystemLookup = {}
        EmergencyLookup = {}
        in_fields = ["LEGACY_OID", "GLOBALID"]
        sFC = self.__sTargetGDB + "/PipelineDevice"
        for i in range(2):
            if i == 0:
                Lookup = SystemLookup
                sWhere = "assetgroup = 3 AND assettype = 485"
            else:
                Lookup = EmergencyLookup
                sWhere = "assetgroup = 3 AND assettype = 483"
            with arcpy.da.SearchCursor(sFC, in_fields, sWhere) as cursorIn:
                for in_row in cursorIn:
                    Lookup[in_row[0]] = in_row[1]

        # Get station global IDs
        print("   Building station lookup tables...")
        RegLookup = {}
        TBLookup = {}
        sFC = self.__sTargetGDB + "/PipelineAssembly"
        for i in range(2):
            if i == 0:
                Lookup = RegLookup
                sWhere = "assetgroup = 3"
            elif i == 1:
                Lookup = TBLookup
                sWhere = "assetgroup = 5"
            with arcpy.da.SearchCursor(sFC, in_fields, sWhere) as cursorIn:
                for in_row in cursorIn:
                    Lookup[in_row[0]] = in_row[1]

        # Migrate relationship table rows
        out_fields = ["ASSEMBLY_GID", "DEVICE_GID", "CREATIONDATE", "CREATOR", "LASTUPDATE", "UPDATEDBY"]
        cursorOut = arcpy.da.InsertCursor(self.__sTargetRelTab, out_fields)

        for i in range(3):
            if i == 0:
                sName = "RegStation_TBS"
                rel_fields = ["REGSTATIONOBJECTID", "TBSOBJECTID"]
                DeviceLookup = SystemLookup
                AssemblyLookup = RegLookup
            elif i == 1:
                sName = "RegStation_GasValve"
                rel_fields = ["REGSTATIONOBJECTID", "GASVALVEOBJECTID"]
                DeviceLookup = EmergencyLookup
                AssemblyLookup = RegLookup
            elif i == 2:
                sName = "TBS_GasValve"
                rel_fields = ["TBSOBJECTID", "GASVALVEOBJECTID"]
                DeviceLookup = EmergencyLookup
                AssemblyLookup = TBLookup
            print("   Migrating " + sName + "...")

            sSourceTab = self.__sSourceGDB + "/" + sName
            with arcpy.da.SearchCursor(sSourceTab, rel_fields) as cursorIn:
                for in_row in cursorIn:
                    iAssemblyOID = in_row[0]
                    if iAssemblyOID not in AssemblyLookup:
                        continue
                    sAssemblyGID = AssemblyLookup[iAssemblyOID]
                    iDeviceOID = in_row[1]
                    if iDeviceOID not in DeviceLookup:
                        continue
                    sDeviceGID = DeviceLookup[iDeviceOID]
                    dtNow = datetime.now()
                    out_row = [sAssemblyGID, sDeviceGID, dtNow, "UPDM", dtNow, "UPDM"]
                    iOID = cursorOut.insertRow(out_row)
        del cursorOut
        return

    def __MigrateReliefValve(self):
        sFCName = "ReliefValve"
        print("Migrating " + sFCName + "...")

        # Create InsertCursor
        out_fields = ["ASSETGROUP", "ASSETTYPE", "CREATIONDATE", "CREATOR", 
                      "LASTUPDATE", "UPDATEDBY", "ASSETID", "LIFECYCLESTATUS", 
                      "ROTATION", "COMMENTS", "SHAPE@", "GLOBALID"
                      ]
        self.__cursor = arcpy.da.InsertCursor(self.__sTargetFC, out_fields)

        # Get projected shapes
        print("   Getting projected shapes...")
        ProjectedShapes = self.__helper.GetProjectedShapes(sFCName)

        # Create SearchCursor
        print("   Processing rows...")
        sSourceFC = self.__sSourceGDB + "/" + sFCName
        in_fields = ["DATECREATED", "CREATIONUSER", "DATEMODIFIED", "LASTUSER", 
                     "FACILITYID", "WORKFUNCTION", "SYMBOLROTATION", "COMMENTS", 
                     "GLOBALID"
                     ]
        with arcpy.da.SearchCursor(sSourceFC, in_fields) as cursorIn:
            for in_row in cursorIn:
                sGlobalID = in_row[8]
                if sGlobalID not in ProjectedShapes:
                    continue
                shape = ProjectedShapes[sGlobalID]
                self.__ProcessRVRow(in_row, shape)
                # Fewer than 1000 records
        del self.__cursor
        return

    def __ProcessRVRow(self, in_row, shape):
        iAssetGroup = 5 # "Relief Valve"
        iAssetType = 21 # "Blow off"
        dtCreated = in_row[0]
        sCreationUser = in_row[1]
        dtModified = in_row[2]
        sLastUser = in_row[3]
        sFacilityID = in_row[4]

        # Domain: UPDM_LifeCycleStatus
        iLifeCycle = 3 # "In Service"
        iWorkFunction = in_row[5]
        if iWorkFunction == 1:
            if self._bSkipProposed:
                return
            iLifeCycle = 1 # "Proposed"

        dRotation = in_row[6]
        sComments = in_row[7]
        sGlobalID = in_row[8]
        out_row = [iAssetGroup, iAssetType, dtCreated, sCreationUser, 
                    dtModified, sLastUser, sFacilityID, iLifeCycle,
                    dRotation, sComments, shape, sGlobalID
                    ]
        self.__cursor.insertRow(out_row)
        return

    def __MigrateControllableFitting(self):
        sFCName = "ControllableFitting"
        print("Migrating " + sFCName + "...")

        # Create InsertCursor
        out_fields = ["ASSETGROUP", "ASSETTYPE", "CREATIONDATE", "CREATOR", 
                      "LASTUPDATE", "UPDATEDBY", "DEVICESTATUS", "INSULATORDEVICE", 
                      "INSERVICEDATE", "LIFECYCLESTATUS", "DIAMETER", "ROTATION", 
                      "MATERIAL", "WORKORDERID", "STORES_STOCK_NUMBER", "SEQUENTIAL_FILE_NUMBER", 
                      "CHART_DATE", "SHAPE@", "GLOBALID"
                      ]
        self.__cursor = arcpy.da.InsertCursor(self.__sTargetFC, out_fields)

        # Get projected shapes
        print("   Getting projected shapes...")
        ProjectedShapes = self.__helper.GetProjectedShapes(sFCName)

        # Create SearchCursor
        print("   Processing rows...")
        sSourceFC = self.__sSourceGDB + "/" + sFCName
        in_fields = ["SUBTYPECD", "DATECREATED", "CREATIONUSER", "DATEMODIFIED", 
                     "LASTUSER", "PRESENTPOSITION", "INSULATEDINDICATOR", "INSERVICEDATE", 
                     "WORKFUNCTION", "FITTINGDIAMETER", "SYMBOLROTATION", "MATERIAL", 
                     "WORKORDERID", "STORES_STOCK_NUMBER", "SEQUENTIAL_FILE_NUMBER", "CHART_DATE", 
                     "GLOBALID"
                     ]
        iCount = 0
        with arcpy.da.SearchCursor(sSourceFC, in_fields) as cursorIn:
            for in_row in cursorIn:
                sGlobalID = in_row[16]
                if sGlobalID not in ProjectedShapes:
                    continue
                shape = ProjectedShapes[sGlobalID]
                self.__ProcessCFRow(in_row, shape)
                iCount += 1
                if iCount % 1000 == 0:
                    print("   ControllableFitting row " + str(iCount) + "...")
        del self.__cursor
        return

    def __ProcessCFRow(self, in_row, shape):

        # Subtype: Line Stopper -> Short Stop, Short Stop -> Controllable Tee
        iAssetGroup = 0
        iAssetType = 0
        iSubtype = in_row[0]
        if iSubtype == 1: # "Short Stop"
            iAssetGroup = 12 # "Controllable Tee"
            iAssetType = 241 # "Controllable Tee"
        elif iSubtype == 3: # "Line Stopper"
            iAssetGroup = 13 # "Short Stop"
            iAssetType = 461 # "Short Stop"
        dtCreated = in_row[1]
        sCreationUser = in_row[2]
        dtModified = in_row[3]
        sLastUser = in_row[4]

        # Domain: UPDM_DEVICESTATUS
        iDeviceStatus = None
        iPresentPosition = in_row[5] # 0 = Closed, 1 = Open
        if iPresentPosition == 0:
            iDeviceStatus = 2
        elif iPresentPosition == 1:
            iDeviceStatus = 1

        # Domain: UPDM_INSULATORDEVICE
        iInsulated = None
        sInsulated = in_row[6]
        if sInsulated == "No":
            iInsulated = 1
        elif sInsulated == "Yes":
            iInsulated = 2;

        dtInService = in_row[7]

        # Domain: UPDM_LifeCycleStatus
        iLifeCycle = 3 # "In Service"
        iWorkFunction = in_row[8]
        if iWorkFunction == 1:
            if self._bSkipProposed:
                return
            iLifeCycle = 1 # "Proposed"

        # Diameter: NxUxU -> N (double)
        sDiameter = in_row[9]
        if sDiameter.find("x") > -1:
            sDiameter = sDiameter.split("x")[0]
        dDiameter = self.__helper.StringToFloat(sDiameter)

        dSymbolRotation = in_row[10]
                
        # Domain: UPDM_ASTMComponentMaterial
        sMaterial = None
        sMat = in_row[11]
        if sMat == "MTL":
            sMaterial = "M" # "Steel"
        elif sMat == "PL":
            sMaterial = "X" # "Other"

        sWorkOrderID = in_row[12]
        sSSN = in_row[13]
        sSFN = in_row[14]
        dtChart = in_row[15]
        sGlobalID = in_row[16]

        out_row = [iAssetGroup, iAssetType, dtCreated, sCreationUser, 
                    dtModified, sLastUser, iDeviceStatus, iInsulated,
                    dtInService, iLifeCycle, dDiameter, dSymbolRotation, 
                    sMaterial, sWorkOrderID, sSSN, sSFN, 
                    dtChart, shape, sGlobalID
                    ]
        self.__cursor.insertRow(out_row)
        return

print("Ready.")
if __name__ == "__main__":
    md = MigrateDevice()
    md.DoIt()
