﻿# Migrate_helper.py [Python 3.6]
# Prepares data for Utility Network migration
#
# BEFORE MIGRATION:
# 1. Make sure SDE connection is set for branch versioning
# 
# MIGRATION STEPS:
# 1. Project existing data to WGS 1984 UTM Zone 12N: MigrateHelper.Prepare_UTM84
# 2. Copy base asset package and set spatial reference: MigrateHelper.Create_Base
# 3. Apply UPDM modifications: MigrateHelper.Apply_UPDM, MigrateHelper.Fix_UPDM
# 4. Apply AZGas modifications: MigrateHelper.Apply_AZGas
# 5. Migrate AZGas data to asset package: MigrateHelper.Migrate_Data
# 6. Stage utility network and apply asset package: MigrateHelper.Create_AZGas
# 7. Load data: MigrateHelper.Load_AZGasData
# 8. Generate or refresh connectivity errors: MigrateHelper.Enable_Topology
# 9. Fix geometry and point errors: MigrateHelper.Fix_Errors 
# 10. Register as versioned: MigrateHelper.Register_Versioned
# 11. Enable topology: MigrateHelper.Enable_Topology
# 12. Manual tasks: fix remaining errors and update subnetworks

print("Migrate_helper.py")
print("Initializing...")

import arcpy
import arcpy.da
from datetime import datetime
import re
import traceback
import uuid

class MigrateHelper(object):
    # Constants
    _bTopoBug = True
    _bShortTest = False
    _sShortTestTown = "FLAGSTAFF"
    _bNewAssetGroup = True
    _sNewAssetGroupName = "Bogus Controller"
    _sWorkingDir = "C:/apps/Data/UPDM_Migration"
    _sSourceGDB = "C:/apps/Data/AZGas/AZGas.gdb"
    _sUTM84GDBName = "AZGas_UTM84.gdb"
    _sTargetGDBName = "AssetPackage_AZGas.gdb"
    _sInstance = ""
    _sKeyCodeFile = "keycodes.ecp"
    _sAdminConnection = "AZGas (sa).sde"
    _sDBConnection = "AZGas (UES).sde"
    _sUNDSName = "PipeSystem"
    _sUNName = "UtilityNetwork"
    _sLineName = "PipelineLine"
    _sSubnetLineName = "PipelineSubnetLine"
    _sPortalUrl = "" 
    _sPortalUser = ""
    _sPortalPass = ""
    _iWKID = 32612 # WGS_1984_UTM_Zone_12N
    _sTransform = "WGS_1984_(ITRF00)_To_NAD_1983"

    # Private fields
    __SpatialRef = None
    __bSignedIn = False
    __EditTrackingDatasets = [
        "PipelineLine",
        "PipelineServiceSummary",
        "PipelineAssembly",
        "METER",
        "PipelineDevice",
        "UNSGPipelineAssembly_PipelineDevice",
        "ValveInspection",
        "PipelineJunction"]

    def __init__(self):
        # Set up standard UTM spatial reference
        self.__SpatialRef = arcpy.SpatialReference(self._iWKID)
        return

    # Public methods

    def SetEditorTracking(self, sDS, bEnable):
        if bEnable:
            arcpy.EnableEditorTracking_management(sDS, "CREATOR", "CREATIONDATE", "UPDATEDBY", "LASTUPDATE");
        else:
            arcpy.DisableEditorTracking_management(sDS);
        return

    def AddEditorTracking(self, sTabName):
        arcpy.AddField_management(sTabName, "CREATIONDATE", "DATE", field_alias = "Creation date")
        arcpy.AddField_management(sTabName, "CREATOR", "TEXT", field_length = 255,
                                  field_alias = "Creator")
        arcpy.AddField_management(sTabName, "LASTUPDATE", "DATE", field_alias = "Last update")
        arcpy.AddField_management(sTabName, "UPDATEDBY", "TEXT", field_length = 255,
                                  field_alias = "Updated by")
        self.SetEditorTracking(sTabName, True)
        return

    def StringToFloat(self, sVal):
        if not sVal:
            return None
        dVal = None
        sParseVal = re.sub('[^.0-9]', '', sVal)
        try:
            dVal = float(sParseVal)
            return dVal
        except Exception:
            dVal = None
        return dVal

    def GetProjectedLayer(self, sFCName):
        sProjGDB = self._sWorkingDir + "/" + self._sUTM84GDBName
        sProjFC = sProjGDB  + "/" + sFCName
        sFeatLayer = "lyr_" + str(uuid.uuid4())
        arcpy.MakeFeatureLayer_management(sProjFC, sFeatLayer)
        if self._bShortTest:
            # Select geometries for short test
            sTownLayer = "lyr_" + str(uuid.uuid4())
            sTownFC = sProjGDB + "/general_town_limit"
            sWhere = "town_name = '" + self._sShortTestTown + "'"
            arcpy.MakeFeatureLayer_management(sTownFC, sTownLayer, sWhere)
            arcpy.SelectLayerByLocation_management(sFeatLayer, "WITHIN", sTownLayer)
        result = arcpy.GetCount_management(sFeatLayer)
        iCount = int(result[0])
        return sFeatLayer

    def GetProjectedShapes(self, sFCName):
        ProjectedShapes = {}
        sFeatLayer = self.GetProjectedLayer(sFCName)
        Fields = ["GLOBALID", "SHAPE@"]
        with arcpy.da.SearchCursor(sFeatLayer, Fields) as c:
            for in_row in c:
                ProjectedShapes[in_row[0]] = in_row[1]
        return ProjectedShapes

    def GetConnectedLines(self, sPointFCName):
        ConnectedLines = {}

        # Create temporary point FC
        print("   Copying data...")
        sLayer = self.GetProjectedLayer(sPointFCName)
        sTempGDB = arcpy.env.scratchGDB
        sPointFC = sTempGDB + "/" + sPointFCName
        if arcpy.Exists(sPointFC):
            arcpy.Delete_management(sPointFC)
        desc = arcpy.Describe(sLayer)
        sr = desc.spatialReference
        sType = desc.shapeType.upper()
        arcpy.CreateFeatureclass_management(sTempGDB, sPointFCName, sType, spatial_reference = sr)
        arcpy.AddField_management(sPointFC, "GLOBALID", "GUID")
        arcpy.Append_management(sLayer, sPointFC, "NO_TEST")

        # Perform spatial join
        print("   Performing spatial join...")
        sTargetGDB = self._sWorkingDir + "/" + self._sTargetGDBName
        sLineFC = sTargetGDB + "/" + self._sLineName
        sJoinFC = sPointFC + "_Join"
        if arcpy.Exists(sJoinFC):
            arcpy.Delete_management(sJoinFC)
        arcpy.SpatialJoin_analysis(sPointFC, sLineFC, sJoinFC, "JOIN_ONE_TO_MANY")

        # Build connectivity table
        print("   Querying connected lines...")
        Fields = ["GLOBALID", "GLOBALID_1", "ASSETGROUP", "MAOPRECORD", "SYSTEM_NAME"]
        LineIDs = set()
        with arcpy.da.SearchCursor(sJoinFC, Fields) as cursor:
            for row in cursor:
                sPointGID = row[0]
                sLineGID = row[1]
                LineIDs.add(sLineGID)
                iAssetGroup = row[2]
                dMAOP = row[3]
                sSystemName = row[4]
                LineRec = [sLineGID, iAssetGroup, dMAOP, sSystemName]
                if sPointGID in ConnectedLines:
                    ConnectedLines[sPointGID].append(LineRec)
                else:
                    ConnectedLines[sPointGID] = [LineRec]

        # Get line geometries
        print("   Getting line geometries...")
        Fields = ["GLOBALID", "SHAPE@"]
        LineGeoms = {}
        with arcpy.da.SearchCursor(sLineFC, Fields) as cursor:
            for row in cursor:
                sLineGID = row[0]
                if sLineGID not in LineIDs:
                    continue
                shape = row[1]
                LineGeoms[sLineGID] = shape
        del LineIDs
        return ConnectedLines, LineGeoms

    def FromOrTwo(self, geomLine, geomPoint):
        iIndex = -1
        sr = geomLine.spatialReference
        if self.Overlaps(geomLine.firstPoint, sr, geomPoint):
            iIndex = 0
        elif self.Overlaps(geomLine.lastPoint, sr, geomPoint):
            iIndex = 1
        return iIndex

    def Overlaps(self, ptEndpoint, sr, geomValve):
        geomPt = arcpy.PointGeometry(ptEndpoint, sr)
        bufPt = geomPt.buffer(0.01)
        return not bufPt.disjoint(geomValve)

    def GetUNID(self):
        sTargetGDB = self._sWorkingDir + "/" + self._sDBConnection
        arcpy.env.workspace = sTargetGDB
        Tables = arcpy.ListTables("*UN_*_TOPOLOGY")
        if len(Tables) == 0:
            print("Topology table not found.")
            return None
        sTabName = Tables[0]
        Tok = sTabName.split("_")
        sID = Tok[1]
        return sID

    def Prepare_UTM84(self):
        # Create projected version of AZGas to retrieve geometries
        print("Preparing UTM84 GDB...")
        dtStart = datetime.now()

        # Copy AZGas to temp GDB
        print("Copying AZGas to temp GDB...")
        sTempGDB = "C:/Temp/AZGas_Temp.gdb"
        if arcpy.Exists(sTempGDB):
            arcpy.Delete_management(sTempGDB)
        arcpy.Copy_management(self._sSourceGDB, sTempGDB)
        arcpy.env.workspace = sTempGDB

        # Copy GasGeomNetwork_Junctions to backup feature class
        print("Copying GasGeomNetwork_Junctions...")
        sInFC = sTempGDB + "/GasGeomNetwork_Junctions"
        sOutName = "GasGeomNetwork_Junctions_bak"
        arcpy.FeatureClassToFeatureClass_conversion(sInFC, sTempGDB, sOutName)
        #print("Adding global IDs...")
        #arcpy.AddGlobalIDs_management(sTempGDB + "/" + sOutName)

        # Remove geometric network
        print("Removing geometric network...")
        arcpy.Delete_management(sTempGDB + "/GasDataset/GasGeomNetwork")

        # Create UTM84 GDB
        print("Creating UTM84 GDB...")
        sUTMGDB = self._sWorkingDir + "/" + self._sUTM84GDBName
        if arcpy.Exists(sUTMGDB):
            arcpy.Delete_management(sUTMGDB)
        arcpy.CreateFileGDB_management(self._sWorkingDir, self._sUTM84GDBName);

        # Project datasets and feature classes
        print("Projecting datasets...")
        ProjectList = []
        for sName in arcpy.ListDatasets():
            desc = arcpy.Describe(sName)
            if desc.datasetType == "RasterDataset":
                continue
            ProjectList.append(sName)
        ProjectList.extend(arcpy.ListFeatureClasses())
        for sName in ProjectList:
            print("   " + sName)
            sTarget = sUTMGDB + "/" + sName
            arcpy.Project_management(sName, sTarget, self.__SpatialRef, self._sTransform)

        # Clean up
        print("Finishing up...")
        arcpy.ClearEnvironment("workspace")
        arcpy.Delete_management(sTempGDB)
        print("Done.")
        dtEnd = datetime.now()
        print("Elapsed time: " + str(dtEnd - dtStart))
        return

    def Create_Base(self):
        # Create base asset package for migration
        # [After running this, execute Create_UPDM]
        print("Creating base asset package...")
        dtStart = datetime.now()

		# Copy asset package GDB
        print("   Copying asset package...")
        sSourceGDB = self._sWorkingDir + "/UPDM_2018_zip/AssetPackage/Pipeline_AssetPackage.gdb"
        sTargetGDB = self._sWorkingDir + "/Pipeline_AssetPackage.gdb"
        if arcpy.Exists(sTargetGDB):
            arcpy.Delete_management(sTargetGDB)
        arcpy.Copy_management(sSourceGDB, sTargetGDB)

        # Delete sample data
        print("   Deleting sample data...")
        arcpy.env.workspace = sTargetGDB
        DeleteList = arcpy.ListFeatureClasses(feature_dataset = "UtilityNetwork")
        DeleteList.append("ValveInspection")
        DeleteList.append("C_Associations")
        DeleteList.append("C_SubnetworkControllers")
        for sName in DeleteList:
            arcpy.TruncateTable_management(sName)

        # Change spatial reference
        print("   Changing spatial reference...")
        sSourceGDB = sTargetGDB
        sTargetGDB = self._sWorkingDir + "/" + self._sTargetGDBName 
        if arcpy.Exists(sTargetGDB):
            arcpy.Delete_management(sTargetGDB)
        arcpy.pt.ChangeGDBSpatialReference(sSourceGDB, self.__SpatialRef,
            self._sWorkingDir, self._sTargetGDBName)

        # Populate ServiceTerritory
        print("Populating ServiceTerritory...")
        sSourceFC = self._sWorkingDir + "/" + self._sUTM84GDBName + "/District"
        sTargetGDB = self._sWorkingDir + "/" + self._sTargetGDBName
        sTargetFC = sTargetGDB + "/ServiceTerritory"
        in_fields = ["SHAPE@"]
        out_fields = ["SHAPE@", "globalid", "GlobalID_1", "GlobalID_2"]
        arcpy.TruncateTable_management(sTargetFC)
        with arcpy.da.Editor(sTargetGDB) as editST:
            out_cursor = arcpy.da.InsertCursor(sTargetFC, out_fields)
            with arcpy.da.SearchCursor(sSourceFC, in_fields) as cursorIn:
                for in_row in cursorIn:
                    sGID = "{" + str(uuid.uuid4()).upper() + "}"
                    sGID1 = "{" + str(uuid.uuid4()).upper() + "}"
                    sGID2 = "{" + str(uuid.uuid4()).upper() + "}"
                    out_row = [in_row[0], sGID, sGID1, sGID2]
                    out_cursor.insertRow(out_row)
            del out_cursor

        # Clean up
        print("Finishing up...")
        arcpy.Delete_management(sSourceGDB)
        sBackupGDB = self._sWorkingDir + "/AssetPackage_Base.gdb"
        if arcpy.Exists(sBackupGDB):
            arcpy.Delete_management(sBackupGDB)
        arcpy.Copy_management(sTargetGDB, sBackupGDB)
        print("Done.")
        dtEnd = datetime.now()
        print("Elapsed time: " + str(dtEnd - dtStart))
        return

    def Apply_UPDM(self):
        # Apply UPDM modifications to base asset package
        print("Applying UPDM to base asset package...")
        dtStart = datetime.now()
        exec(open("./01CreateUPDMBaseAdd-On_2018.py").read())
        exec(open("./02CreateIntegrityModule_2018.py").read())
        exec(open("./03CreateAssetInspectionModule_2018.py").read())

        print("Finishing up...")
        sTargetGDB = self._sWorkingDir + "/" + self._sTargetGDBName
        sBackupGDB = self._sWorkingDir + "/AssetPackage_UPDM.gdb"
        if arcpy.Exists(sBackupGDB):
            arcpy.Delete_management(sBackupGDB)
        arcpy.Copy_management(sTargetGDB, sBackupGDB)
        print("Done.")
        dtEnd = datetime.now()
        print("Elapsed time: " + str(dtEnd - dtStart))
        return

    def Fix_UPDM(self):
        # Fix UPDM asset package in order to enable versioning and network topology
        print("Fixing UPDM asset package...")
        dtStart = datetime.now()
        sTargetGDB = self._sWorkingDir + "/" + self._sTargetGDBName
        arcpy.env.workspace = sTargetGDB

        # Enable editor tracking for ValveInspection
        mh.SetEditorTracking("ValveInspection", True)

        # Assign correct domains to designtype field
        # [01CreateUPDMBaseAdd-On_2018.py line 2152]
        print("   Fixing DESIGNTYPE domains...")
        arcpy.AssignDomainToField_management("PipelineDevice", "designtype", "UPDM_ValveControllableType", 3)
        arcpy.AssignDomainToField_management("PipelineDevice", "designtype", "UPDM_ValveFlowType", 4)

        # Fix description for "Replaced Regulator" coded value in UPDM_LeakWorkPerformed domain
        # [03CreateInspectionModule_2018.py line 392]
        print("   Fixing UPDM_LeakWorkPerformed coded value description...")
        arcpy.DeleteCodedValueFromDomain_management(sTargetGDB, "UPDM_LeakWorkPerformed", "Replaced Regulator")
        arcpy.AddCodedValueToDomain_management(sTargetGDB, "UPDM_LeakWorkPerformed", "Replaced Regulator", "Replaced Regulator")

        # Disable and re-enable attachments
        # [03CreateInspectionModule_2018.py line 1541]
        print("   Fixing P_LineValveInspection attachments...")
        arcpy.DisableAttachments_management("P_LineValveInspection")
        arcpy.EnableAttachments_management("P_LineValveInspection")

        # Add editor tracking to PipelineServiceSummary
        # [01CreateUPDMBaseAdd-On_2018.py]
        print("   Adding editor tracking to PipelineServiceSummary...")
        self.AddEditorTracking("PipelineServiceSummary")

        # Add editor tracking to attachment tables
        # [03CreateInspectionModule_2018.py]
        print("   Adding editor tracking to attachment tables...")
        self.AddEditorTracking("P_RegulatorInspect__ATTACH")
        self.AddEditorTracking("P_StrainerInspect__ATTACH")
        self.AddEditorTracking("P_ValveInspection__ATTACH")
        self.AddEditorTracking("P_LineValveInspection__ATTACH")
        self.AddEditorTracking("P_ReliefValveInspection__ATTACH")
        self.AddEditorTracking("P_StationVisualInspection__ATTACH")

        print("Finishing up...")
        sBackupGDB = self._sWorkingDir + "/AssetPackage_UPDM2.gdb"
        if arcpy.Exists(sBackupGDB):
            arcpy.Delete_management(sBackupGDB)
        arcpy.Copy_management(sTargetGDB, sBackupGDB)
        print("Done.")
        dtEnd = datetime.now()
        print("Elapsed time: " + str(dtEnd - dtStart))
        return

    def Apply_AZGas(self):
        # Modify UPDM asset package for AZGas migration
        print("Modifying UPDM asset package...")
        dtStart = datetime.now()
        sTargetGDB = self._sWorkingDir + "/" + self._sTargetGDBName
        sBackupGDB = self._sWorkingDir + "/AssetPackage_UPDM2.gdb"
        if arcpy.Exists(sBackupGDB):
            if arcpy.Exists(sTargetGDB):
                arcpy.Delete_management(sTargetGDB)
            arcpy.Copy_management(sBackupGDB, sTargetGDB)
        arcpy.env.workspace = sTargetGDB

        # Add station/valve relationship class
        sRelClassName = "UNSGPipelineAssembly_PipelineDevice"
        print("   Adding PipelineAssembly/PipelineDevice relationship class...")
        arcpy.CreateRelationshipClass_management("PipelineAssembly", "PipelineDevice", sRelClassName,
                                                    "SIMPLE", "Device", "Assembly", "NONE", "MANY_TO_MANY", "NONE",
                                                    "GLOBALID", "ASSEMBLY_GID", "GLOBALID", "DEVICE_GID")
        arcpy.AddGlobalIDs_management(sRelClassName)
        self.AddEditorTracking(sRelClassName)

        #**** LINE UPDATES ****
        #   DistributionMain 
        #   SubTransmissionMain
        #   TransmissionMain
        #   AbandonedDistributionMain
        #   AbandonedSubTransmissionMain
        #   AbandonedTransmissionMain
        #   Service
        #   AbandonedService
        #   CPBondWire

        # Add work crew domain
        print("   Adding work crew domain...")
        sDomainName = "UNSG_WorkCrew"
        sDesc = "UNS Gas work crew"
        Values = ["NPL", "UES"]
        arcpy.CreateDomain_management(sTargetGDB, sDomainName, sDesc, "TEXT", "CODED")
        for sVal in Values:
            arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, sVal, sVal)

        # Add USNG CP color domain
        print("   Adding CP color domain...")
        sDomainName = "UNSG_CPColor"
        sDesc = "UNS Gas CP circuit color"
        Values = ["blue", "brown", "cyan", "green", "grey", "magenta", "orange", "red", "yellow"]
        arcpy.CreateDomain_management(sTargetGDB, sDomainName, sDesc, "TEXT", "CODED")
        for sVal in Values:
            arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, sVal, sVal)
        
        # Add missing transmission pipe diameter domain values
        print("   Adding transmission diameter values...")
        sDomainName = "UPDM_LineDiameterTransmission"
        Values = ["1", "2", "4", "6", "8", "10"]
        for sVal in Values:
            dVal = self.StringToFloat(sVal)
            arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, dVal, sVal)

        # Add missing service pipe diameter domain values
        print("   Adding service diameter values...")
        sDomainName = "UPDM_LineDiameterServices"
        Values = ["8", "16"]
        for sVal in Values:
            dVal = self.StringToFloat(sVal)
            arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, dVal, sVal)

        # Add PipelineLine fields
        print("   Adding PipelineLine fields...")
        arcpy.AddField_management("PipelineLine", "STORES_STOCK_NUMBER", "TEXT", field_length = 10,
                                  field_alias = "Stores Stock Number (R)")
        arcpy.AddField_management("PipelineLine", "SYSTEM_NAME", "TEXT", field_length = 40,
                                  field_alias = "System Name")
        arcpy.AddField_management("PipelineLine", "SYSTEM_SECTION", "LONG", field_alias = "System Section")
        arcpy.AddField_management("PipelineLine", "SEQUENTIAL_FILE_NUMBER", "TEXT", field_length = 10,
                                  field_alias = "Sequential File Number")
        arcpy.AddField_management("PipelineLine", "INSTALLED_BY", "TEXT", field_length = 20,
                                  field_alias = "Work Crew", field_domain = "UNSG_WorkCrew")
        arcpy.AddField_management("PipelineLine", "STATUS_DATE", "DATE", field_alias = "Status Change Date")
        arcpy.AddField_management("PipelineLine", "LOCATABLE", "SHORT", 
                                  field_alias = "Locatable", field_domain = "UPDM_YesNoIndicator")
        arcpy.AssignDefaultToField_management("PipelineLine", "LOCATABLE", 1)
        arcpy.AddField_management("PipelineLine", "LS_MAINLINE", "SHORT", 
                                  field_alias = "Leak Survey Mainline", field_domain = "UPDM_YesNoIndicator")
        arcpy.AddField_management("PipelineLine", "CP_TOWN_NAME", "TEXT", field_length = 40,
                                  field_alias = "CP Town Name")
        arcpy.AddField_management("PipelineLine", "CP_CIRCUIT_NUMBER", "TEXT", field_length = 10,
                                  field_alias = "CP Circuit Number")
        arcpy.AddField_management("PipelineLine", "CP_COLOR", "TEXT", field_length = 10,
                                  field_alias = "CP Color", field_domain = "UNSG_CPColor")
        arcpy.AddField_management("PipelineLine", "ISOLATED_MISMATCH", "SHORT", 
                                  field_alias = "Mismatch", field_domain = "UPDM_YesNoIndicator")
        arcpy.AddField_management("PipelineLine", "COMMENTS", "TEXT", field_length = 128,
                                  field_alias = "Comments")
        arcpy.AddField_management("PipelineLine", "SERVICE_CARD", "TEXT", field_length = 255,
                                  field_alias = "Service Card")
        arcpy.AddField_management("PipelineLine", "SLEEVED", "SHORT", 
                                  field_alias = "Sleeved", field_domain = "UPDM_YesNoIndicator")
        arcpy.AddField_management("PipelineLine", "STUB", "SHORT", 
                                  field_alias = "Is Stub", field_domain = "UPDM_YesNoIndicator")

        # Set new domains for subtypes
        print("   Setting new domains for subtypes...")
        arcpy.AssignDomainToField_management("PipelineLine", "INSTALLED_BY", "UNSG_WorkCrew", [1, 2, 3])
        arcpy.AssignDomainToField_management("PipelineLine", "LOCATABLE", "UPDM_YesNoIndicator", [1, 2, 3])
        arcpy.AssignDomainToField_management("PipelineLine", "LS_MAINLINE", "UPDM_YesNoIndicator", [1, 2, 3])
        arcpy.AssignDomainToField_management("PipelineLine", "CP_COLOR", "UNSG_CPColor", [1, 2, 3, 8])
        arcpy.AssignDomainToField_management("PipelineLine", "ISOLATED_MISMATCH", "UPDM_YesNoIndicator", 1)
        arcpy.AssignDomainToField_management("PipelineLine", "SLEEVED", "UPDM_YesNoIndicator", 1)
        arcpy.AssignDomainToField_management("PipelineLine", "STUB", "UPDM_YesNoIndicator", 1)

        #**** ASSEMBLY UPDATES ****
        #   TownBorderStation [except POC Valve]
        #   RuralTap
        #   RegulatorStation
        #   MeterSetting

        # Increase PipelineAssembly station name field length
        print("   Increasing station name field length...")
        arcpy.AlterField_management("PipelineAssembly", "name", field_length = 40)

        # Add station type domain
        print("   Adding station type domain...")
        sDomainName = "UNSG_StationType"
        sDesc = "UNS Gas station type"
        #Values = ["District Regulator Left (05)", "District Regulator Right (05)", "Farm Gate (06)", "Farm Tap (03)", 
        #          "Low Pressure Left (04)", "Low Pressure Right (04)", "Other (07)", "Point Of Custody (POC)", 
        #          "Rural Tap (02)", "Tank Regulator (Tank)", "Town Border (01)"]
        Values = ["District Regulator Left (05)", "District Regulator Right (05)", "Farm Gate (06)", "Farm Tap (03)", 
                  "Low Pressure Left (04)", "Low Pressure Right (04)", "Other (07)", "Point Of Custody (POC)", 
                  "Rural Tap (02)", "Tank Regulator (Tank)", "City Gate (01)"]
        arcpy.CreateDomain_management(sTargetGDB, sDomainName, sDesc, "TEXT", "CODED")
        for sVal in Values:
            arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, sVal, sVal)

        # Add station supplier domain
        print("   Adding station supplier domain...")
        sDomainName = "UNSG_StationSupplier"
        sDesc = "UNS Gas station supplier"
        Values = ["El Paso Natural Gas (18)", "Other (11)", "Questar (20)", "Unisource (10)", "Transwestern (19)"]
        arcpy.CreateDomain_management(sTargetGDB, sDomainName, sDesc, "TEXT", "CODED")
        for sVal in Values:
            arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, sVal, sVal)

        # Add meter setting status domain
        print("   Adding meter setting status domain...")
        sDomainName = "UNSG_MeterSettingStatus"
        sDesc = "UNS Gas meter setting status"
        Values = [1, 2, 3, 4]
        Labels = ["Meter Setting", "Pre-Install", "Installed", "GPS Collected"]
        arcpy.CreateDomain_management(sTargetGDB, sDomainName, sDesc, "LONG", "CODED")
        for i in range(len(Values)):
            arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, Values[i], Labels[i])

        # Add PipelineAssembly fields
        print("   Adding PipelineAssembly fields...")
        arcpy.AddField_management("PipelineAssembly", "TYPE", "TEXT", field_length = 35,
                                  field_alias = "Type", field_domain = "UNSG_StationType")
        arcpy.AddField_management("PipelineAssembly", "STATION_NUMBER", "LONG", field_alias = "Station Number")
        arcpy.AddField_management("PipelineAssembly", "STATION_SUPPLIER", "TEXT", field_length = 30,
                                  field_alias = "Supplier", field_domain = "UNSG_StationSupplier")
        arcpy.AddField_management("PipelineAssembly", "SUPPLIER_IDENTIFICATION_NUMBER", "TEXT", field_length = 15,
                                  field_alias = "Supplier ID Number")
        arcpy.AddField_management("PipelineAssembly", "SEQUENTIAL_FILE_NUMBER", "TEXT", field_length = 10,
                                  field_alias = "Sequential File Number")
        arcpy.AddField_management("PipelineAssembly", "PHOTOS", "TEXT", field_length = 255,
                                  field_alias = "Photos")
        arcpy.AddField_management("PipelineAssembly", "CHART_DATE", "DATE", field_alias = "Chart Date")
        arcpy.AddField_management("PipelineAssembly", "METERSETTING_STATUS", "LONG", 
                                  field_alias = "Meter Setting Status", field_domain = "UNSG_MeterSettingStatus")
        arcpy.AddField_management("PipelineAssembly", "FARM_TAP_GID", "GUID", field_alias = "Farm Tap GlobalID")
        arcpy.AddField_management("PipelineAssembly", "STATUS_DATE", "DATE", field_alias = "Status Change Date")
        arcpy.AddField_management("PipelineAssembly", "LABEL_TEXT", "TEXT", field_length = 128,
                                  field_alias = "Label Text")
        arcpy.AddField_management("PipelineAssembly", "LEGACY_OID", "LONG", field_alias = "Legacy Relationship OID")

        # Set new domains for subtypes
        print("   Setting new domains for subtypes...")
        arcpy.AssignDomainToField_management("PipelineAssembly", "TYPE", "UNSG_StationType", [3, 4, 5])
        arcpy.AssignDomainToField_management("PipelineAssembly", "STATION_SUPPLIER", "UNSG_StationSupplier", [3, 4, 5])
        arcpy.AssignDomainToField_management("PipelineAssembly", "METERSETTING_STATUS", "UNSG_MeterSettingStatus", 2)

        # Enable attachments
        # print("   Enabling PipelineAssembly attachments...")
        # arcpy.EnableAttachments_management("PipelineAssembly")

        #**** DEVICE UPDATES ****
        #   TownBorderStation [POC Valve]
        #   GasValve
        #   ReliefValve
        #   ControllableFitting

        # Increase PipelineDevice location description field length
        print("   Increasing location description field length...")
        arcpy.AlterField_management("PipelineDevice", "locationdescription", field_length = 255)

        # Add MSV value to UPDM_ValveControllableType domain
        print("   Adding MSV value to UPDM_ValveControllableType domain...")
        arcpy.AddCodedValueToDomain_management(sTargetGDB, "UPDM_ValveControllableType", 101, "Manual Shutoff")

        # Add valve use domain
        print("   Adding valve use domain...")
        sDomainName = "UNSG_ValveUse"
        sDesc = "UNS Gas Valve Use"
        Values = ["BYPS", "ISO", "PURG", "RELF", "IN", "OUT", "BLDN"]
        Labels = ["Bypass", "Isolation", "Purge", "Relief", "Inlet", "Outlet", "Blow Down"]
        arcpy.CreateDomain_management(sTargetGDB, sDomainName, sDesc, "TEXT", "CODED")
        for i in range(len(Values)):
            arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, Values[i], Labels[i])

        # Add valve rating domain
        print("   Adding valve rating domain...")
        sDomainName = "UNSG_ValveRating"
        sDesc = "UNS Gas ANSI valve rating"
        Values = ["1", "2", "5", "6", 
                  "7", "8", "A", "C", 
                  "E", "F", "K", "L", 
                  "M", "N", "O", "P", 
                  "UNK"]
        Labels = ["7 w.c.", "8-10 psig", "20-25 psig", "50-55 psig", 
                  "70-76 psig", "70-95 psig", "125 psig or less", "175 psig or less", 
                  "215 psig or less", "250 psig or less", "400 psig or less", "450 psig or less", 
                  "500 psig or less", "550 psig or less", "600 psig or less", "650 psig or less", 
                  "Unknown"]
        arcpy.CreateDomain_management(sTargetGDB, sDomainName, sDesc, "TEXT", "CODED")
        for i in range(len(Values)):
            arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, Values[i], Labels[i])

        # Add ground level domain
        print("   Adding ground level domain...")
        sDomainName = "UNSG_GroundLevel"
        sDesc = "UNS Gas ground level"
        Values = ["A", "B"]
        Labels = ["Above Ground", "Below Ground"]
        arcpy.CreateDomain_management(sTargetGDB, sDomainName, sDesc, "TEXT", "CODED")
        for i in range(len(Values)):
            arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, Values[i], Labels[i])

        # Add EFV series domain
        print("   Adding EFV series domain...")
        sDomainName = "UNSG_EFVSeries"
        sDesc = "UNS Gas EFV series"
        Values = ["400", "800", "1100", "1800", "5500", "300", "475", "525", "700", "775", "1200", "3900", "2600"]
        arcpy.CreateDomain_management(sTargetGDB, sDomainName, sDesc, "TEXT", "CODED")
        for sVal in Values:
            arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, sVal, sVal)

        # Add PipelineDevice fields
        # [Odorizer field for POC Valve not added, as it may be modelled by a separate feature]
        print("   Adding PipelineDevice fields...")
        arcpy.AddField_management("PipelineDevice", "VALVEUSE", "TEXT", field_length = 5,
                                  field_alias = "Valve Use", field_domain = "UNSG_ValveUse")
        arcpy.AddField_management("PipelineDevice", "ANSIPRESSURERATING", "TEXT", field_length = 5,
                                  field_alias = "ANSI Pressure Rating", field_domain = "UNSG_ValveRating")
        arcpy.AddField_management("PipelineDevice", "BONDEDINDICATOR", "SHORT", 
                                  field_alias = "Bonded Indicator", field_domain = "UPDM_YesNoIndicator")
        arcpy.AddField_management("PipelineDevice", "GROUNDLEVELINDICATOR", "TEXT", field_length = 5,
                                  field_alias = "Ground Level Indicator", field_domain = "UNSG_GroundLevel")
        arcpy.AddField_management("PipelineDevice", "STORES_STOCK_NUMBER", "TEXT", field_length = 10,
                                  field_alias = "Stores Stock Number (R)")
        arcpy.AddField_management("PipelineDevice", "SYSTEM_NAME", "TEXT", field_length = 40,
                                  field_alias = "System Name")
        arcpy.AddField_management("PipelineDevice", "SEQUENTIAL_FILE_NUMBER", "TEXT", field_length = 10,
                                  field_alias = "Sequential File Number")
        arcpy.AddField_management("PipelineDevice", "CHART_DATE", "DATE", field_alias = "Chart Date")
        arcpy.AddField_management("PipelineDevice", "LS_MAINLINE", "SHORT", 
                                  field_alias = "Leak Survey Mainline", field_domain = "UPDM_YesNoIndicator")
        arcpy.AddField_management("PipelineDevice", "SERIES", "TEXT", field_length = 5,
                                  field_alias = "EFV Series", field_domain = "UNSG_EFVSeries")
        arcpy.AddField_management("PipelineDevice", "REPORT", "TEXT", field_length = 255,
                                  field_alias = "EFV Report")
        arcpy.AddField_management("PipelineDevice", "LEGACY_OID", "LONG", field_alias = "Legacy Relationship OID")

        # Set new domains for subtypes
        print("   Setting new domains for subtypes...")
        arcpy.AssignDomainToField_management("PipelineDevice", "VALVEUSE", "UNSG_ValveUse", [3, 4])
        arcpy.AssignDomainToField_management("PipelineDevice", "ANSIPRESSURERATING", "UNSG_ValveRating", [3, 4])
        arcpy.AssignDomainToField_management("PipelineDevice", "BONDEDINDICATOR", "UPDM_YesNoIndicator", [3, 4])
        arcpy.AssignDomainToField_management("PipelineDevice", "GROUNDLEVELINDICATOR", "UNSG_GroundLevel", [3, 4])
        arcpy.AssignDomainToField_management("PipelineDevice", "LS_MAINLINE", "UPDM_YesNoIndicator", [3, 4])
        arcpy.AssignDomainToField_management("PipelineDevice", "SERIES", "UNSG_EFVSeries", 1)

        # Add outage status domain
        print("   Adding outage status domain...")
        sDomainName = "UNSG_OutageStatus"
        sDesc = "UNS Gas premise outage status"
        Values = ["Essential", "Interruptable", "Normal"]
        arcpy.CreateDomain_management(sTargetGDB, sDomainName, sDesc, "TEXT", "CODED")
        for sVal in Values:
            arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, sVal, sVal)

        # Create METER table
        print("   Creating METER table...")
        arcpy.CreateTable_management(sTargetGDB, "METER")
        arcpy.AddGlobalIDs_management("METER")
        arcpy.AddField_management("METER", "CREATIONDATE", "DATE", field_alias = "Creation date")
        arcpy.AddField_management("METER", "CREATOR", "TEXT", field_length = 255,
                                  field_alias = "Creator")
        arcpy.AddField_management("METER", "LASTUPDATE", "DATE", field_alias = "Last update")
        arcpy.AddField_management("METER", "UPDATEDBY", "TEXT", field_length = 255,
                                  field_alias = "Updated by")
        arcpy.AddField_management("METER", "DEVICE_GID", "GUID", field_alias = "Device GlobalID")
        arcpy.AddField_management("METER", "METER_NUMBER", "TEXT", field_length = 30,
                                  field_alias = "Meter Number")
        arcpy.AddField_management("METER", "METER_INSTALL_DATE", "DATE", field_alias = "Meter Install Date")
        arcpy.AddField_management("METER", "METER_CYCLE", "TEXT", field_length = 4,
                                  field_alias = "Meter Cycle")
        arcpy.AddField_management("METER", "ROUTE_NUMBER", "TEXT", field_length = 8,
                                  field_alias = "Meter Route")
        arcpy.AddField_management("METER", "STOP_NUMBER", "LONG", field_alias = "Meter Stop")
        arcpy.AddField_management("METER", "LOCATION_ON_PREMISE", "TEXT", field_length = 60,
                                  field_alias = "Meter Location")
        arcpy.AddField_management("METER", "PREMISE_NUMBER", "TEXT", field_length = 10,
                                  field_alias = "Premise Number")
        arcpy.AddField_management("METER", "OUTAGE_STATUS", "TEXT", field_length = 15,
                                  field_alias = "Outage Status", field_domain = "UNSG_OutageStatus")
        arcpy.AddField_management("METER", "REFUNDED", "SHORT", 
                                  field_alias = "Refunded", field_domain = "UPDM_YesNoIndicator")
        arcpy.AddField_management("METER", "CSLA", "TEXT", field_length = 255,
                                  field_alias = "CSLA")
        arcpy.AddField_management("METER", "IMPORT_DATE", "DATE", field_alias = "Import Date")
        arcpy.AddField_management("METER", "PREMISE_STATUS", "TEXT", field_length = 2,
                                  field_alias = "Premise Status")
        arcpy.AddField_management("METER", "STREET_ADDRESS", "TEXT", field_length = 64,
                                  field_alias = "Street Address")
        arcpy.AddField_management("METER", "CITY", "TEXT", field_length = 30,
                                  field_alias = "Postal City")
        arcpy.AddField_management("METER", "ZIP_CODE", "TEXT", field_length = 12,
                                  field_alias = "Zip Code")
        arcpy.AddField_management("METER", "TAXING_CITY", "TEXT", field_length = 16,
                                  field_alias = "Taxing City")
        arcpy.AddField_management("METER", "COMMENTS", "TEXT", field_length = 250,
                                  field_alias = "Comments")
        arcpy.AddField_management("METER", "COMMENT_FLAG", "SHORT", 
                                  field_alias = "Comment Flag", field_domain = "UPDM_YesNoIndicator")
        arcpy.AddField_management("METER", "ELDR_DISBL", "TEXT", field_length = 2,
                                  field_alias = "Elderly/Disabled")
        arcpy.AddField_management("METER", "CUSTOMER_NAME", "TEXT", field_length = 64,
                                  field_alias = "Customer Name")
        arcpy.AddField_management("METER", "CUSTOMER_HOME_PHONE", "TEXT", field_length = 24,
                                  field_alias = "Customer Home Phone")
        arcpy.AddField_management("METER", "CUSTOMER_WORK_PHONE", "TEXT", field_length = 24,
                                  field_alias = "Customer Work Phone")
        arcpy.AddField_management("METER", "CUSTOMER_MOBILE_PHONE", "TEXT", field_length = 24,
                                  field_alias = "Customer Mobile Phone")
        self.SetEditorTracking("METER", True)

        # Add METER relationship class
        print("   Adding PipelineDevice/METER relationship class...")
        sRelClassName = "UNSGPipelineDevice_METER"
        arcpy.CreateRelationshipClass_management("PipelineDevice", "METER", sRelClassName,
                                                 "SIMPLE", "Meter Record", "Customer Meter", "NONE", "ONE_TO_ONE", "NONE",
                                                 "GLOBALID", "DEVICE_GID")

        #**** JUNCTION UPDATES ****
        #   NonControllableFitting
        #   MainJobSeparator
        #   GasGeomNetwork_Junctions

        # Add riser and system asset types to connection point
        print("   Adding riser and system asset types...")
        sDomainName = "Asset Type Pipeline Junction Connection Point"
        arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, 802, "Riser Connection")
        arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, 803, "System Connection")

        # Add values to UPDM_ReducerFittingSize
        print("   Adding fitting size values...")
        sDomainName = "UPDM_ReducerFittingSize"
        Values = [101, 102, 103, 104, 105, 106, 107]
        Labels = ['2" x 1/2"', '2" x 3/4"', '3" x 1"', '4" x 1"', '6" x 2"', '10" x 4"', '16" x 6"']
        for i in range(len(Values)):
            arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, Values[i], Labels[i])

        # Add PipelineJunction fields
        print("   Adding PipelineJunction fields...")
        arcpy.AddField_management("PipelineJunction", "STORES_STOCK_NUMBER", "TEXT", field_length = 10,
                                  field_alias = "Stores Stock Number (R)")
        arcpy.AddField_management("PipelineJunction", "SEQUENTIAL_FILE_NUMBER", "TEXT", field_length = 10,
                                  field_alias = "Sequential File Number")
        arcpy.AddField_management("PipelineJunction", "CHART_DATE", "DATE", field_alias = "Chart Date")

        #**** NETWORK UPDATES ****

        # Add connectivity rules
        #arcpy.AddRule_un(sUN, "JUNCTION_EDGE_CONNECTIVITY", "PipelineJunction", "Plastic Fusion", "Plastic Fusion", "PipelineLine", "Distribution Pipe", "Plastic PVC")
        #arcpy.AddRule_un(sUN, "JUNCTION_EDGE_CONNECTIVITY", "PipelineJunction", "Coupling", "Plastic Coupling", "PipelineLine", "Distribution Pipe", "Plastic PE")
        #arcpy.AddRule_un(sUN, "JUNCTION_JUNCTION_CONNECTIVITY", "PipelineDevice", "Meter", "Customer", "PipelineJunction", "Connection Point", "*")
        print("   Adding connectivity rules...")
        Fields = ["rule_type", "from_domain_network", "from_feature_class", "from_asset_group", "from_asset_type", 
                  "from_terminal", "to_domain_network", "to_feature_class", "to_asset_group", "to_asset_type"]
        cursor = arcpy.da.InsertCursor("B_Rules", Fields)
        row = ("Junction Edge Connectivity", 1, "Junction", "Plastic Fusion", "Plastic Fusion", 
               None, 1, "Line", "Distribution Pipe", "Plastic PVC")
        cursor.insertRow(row)
        row = ("Junction Edge Connectivity", 1, "Junction", "Coupling", "Plastic Coupling", 
               None, 1, "Line", "Distribution Pipe", "Plastic PE")
        cursor.insertRow(row)
        row = ("Junction Junction Connectivity", 1, "Device", "Meter", "Customer",
               None, 1, "Junction", "Connection Point", "*")
        cursor.insertRow(row)

        # Device connections for points of custody
        row = ("Junction Junction Connectivity", 1, "Device", "Controllable Valve", "System",
                "*", 1, "Junction", "Connection Point", "*")
        cursor.insertRow(row)
        if not self._bNewAssetGroup:
            row = ("Junction Junction Connectivity", 1, "Device", "Regulator", "Regulator",
                   "*", 1, "Junction", "Connection Point", "*")
            cursor.insertRow(row)
        del cursor

        # Add terminal configuration to system valve
        print("   Adding terminal configuration...")
        Fields = ["domain_network", "asset_group", "asset_type", "terminal_configuration_name"]
        cursor = arcpy.da.InsertCursor("B_TerminalConfiguration_Assignment", Fields)
        row = (1, "Controllable Valve", "System", "Pipe Bidirectional Dual Terminal")
        cursor.insertRow(row)
        del cursor

        # Add subnetwork controller and tier group terminator categories to system valve
        print("   Adding subnetwork controller category...")
        Fields = ["domain_network", "feature_class", "asset_group", "asset_type", "category_name"]
        cursor = arcpy.da.InsertCursor("B_NetworkCategory_Assignment", Fields)
        row = (1, "Device", "Controllable Valve", "System", "Subnetwork Controller")
        cursor.insertRow(row)
        row = (1, "Device", "Controllable Valve", "System", "Tier Group Terminator")
        cursor.insertRow(row)
        del cursor

        # Set system valve as transmission and distribution system tier controller
        # arcpy.SetSubnetworkDefinition_un(sUN, "Pipeline", "Pipe Distribution System", valid_subnetwork_controller="Controllable Valve/System")
        # arcpy.SetSubnetworkDefinition_un(sUN, "Pipeline", "Pipe Transmission System", valid_subnetwork_controller="Controllable Valve/System")
        print("   Updating subnetwork devices...")
        sWhere = "(tier_name = 'Pipe Distribution System' OR tier_name = 'Pipe Transmission System')"
        sWhere += " AND asset_group = 'Controllable Valve' AND asset_type = 'System'"
        Fields = ["tier_name", "asset_group", "asset_type", "valid_subnetwork_controller"]
        with arcpy.da.UpdateCursor("B_Subnetwork_Devices", Fields, sWhere) as cursor:
            for row in cursor:
                row[3] = 1
                cursor.updateRow(row)

        # Set regulator as transmission pressure tier controller
        # arcpy.SetSubnetworkDefinition_un(sUN, "Pipeline", "Pipe Transmission Pressure", valid_subnetwork_controller="Regulator/Regulator")
        sWhere = "tier_name = 'Pipe Transmission Pressure' AND asset_group = 'Regulator'"
        with arcpy.da.UpdateCursor("B_Subnetwork_Devices", Fields, sWhere) as cursor:
            for row in cursor:
                row[3] = 1
                cursor.updateRow(row)

        # Add features to subnetwork tiers
        print("   Adding subnetwork features...")
        Fields = ["tier_name", "asset_group", "asset_type", "aggregated_line"]
        cursor = arcpy.da.InsertCursor("B_Subnetwork_Lines", Fields)
        row = ("Pipe Distribution System", "Bond Wire", "Bond Wire", 0)
        cursor.insertRow(row)
        row = ("Pipe Distribution Pressure", "Bond Wire", "Bond Wire", 0)
        cursor.insertRow(row)
        row = ("Pipe Distribution Isolation", "Bond Wire", "Bond Wire", 0)
        cursor.insertRow(row)
        row = ("Pipe Transmission System", "Service Pipe", "Coated Steel", 0)
        cursor.insertRow(row)
        row = ("Pipe Transmission Pressure", "Service Pipe", "Coated Steel", 0)
        cursor.insertRow(row)
        del cursor
        Fields = ["tier_name", "asset_group", "asset_type", "valid_subnetwork_controller"]
        cursor = arcpy.da.InsertCursor("B_Subnetwork_Devices", Fields)
        row = ("Pipe Transmission System", "Meter", "Customer", 0)
        cursor.insertRow(row)
        row = ("Pipe Transmission Pressure", "Meter", "Customer", 0)
        cursor.insertRow(row)
        row = ("Pipe Transmission System", "Short Stop", "Short Stop", 0)
        cursor.insertRow(row)
        row = ("Pipe Transmission Pressure", "Short Stop", "Short Stop", 0)
        cursor.insertRow(row)
        del cursor

        # Define distribution condition barriers for CP bond wire
        print("   Adding condition barriers...")
        Fields = ["tier_name", "filter_name", "operator", "type", "value", "combine"]
        cursor = arcpy.da.InsertCursor("B_Subnetwork_ConditionBarriers", Fields)
        row = ("Pipe Distribution Pressure", "LINEASSETTYPE", "IS_EQUAL_TO", "SPECIFIC_VALUE", "281", "OR")
        cursor.insertRow(row)
        row = ("Pipe Distribution Isolation", "LINEASSETTYPE", "IS_EQUAL_TO", "SPECIFIC_VALUE", "281", "OR")
        cursor.insertRow(row)
        del cursor

        # Update subnetwork summaries
        print("   Updating subnetwork summary attributes...")
        sWhere = "summary_attribute = 'NUMBERMETERS'"
        Fields = ["summary_attribute", "filter_value", "function"]
        with arcpy.da.UpdateCursor("B_Subnetwork_Summaries", Fields, sWhere) as cursor:
            for row in cursor:
                row[1] = 42 # "Customer"
                cursor.updateRow(row)
        sWhere = "summary_attribute = 'NUMBERVALVES'"
        with arcpy.da.UpdateCursor("B_Subnetwork_Summaries", Fields, sWhere) as cursor:
            for row in cursor:
                row[1] = 483 # "Emergency"
                cursor.updateRow(row)
        sWhere = "summary_attribute = 'MAOPRECORD'"
        with arcpy.da.UpdateCursor("B_Subnetwork_Summaries", Fields, sWhere) as cursor:
            for row in cursor:
                row[2] = "MAX"
                cursor.updateRow(row)

        # Add new asset group
        if self._bNewAssetGroup:
            self.Define_SubnetworkDevices()

        # Clean up
        print("Finishing up...")
        sBackupGDB = self._sWorkingDir + "/AssetPackage_AZGas_NoData.gdb"
        if arcpy.Exists(sBackupGDB):
            arcpy.Delete_management(sBackupGDB)
        arcpy.Copy_management(sTargetGDB, sBackupGDB)
        print("Done.")
        dtEnd = datetime.now()
        print("Elapsed time: " + str(dtEnd - dtStart))
        return

    def Define_SubnetworkDevices(self):
        # Create bogus device asset group for points of custody that do not actually have devices
        sTargetGDB = self._sWorkingDir + "/" + self._sTargetGDBName

        # Add asset group
        iAssetGroup = 101
        sAssetGroup = self._sNewAssetGroupName
        print("   Adding subnetwork device asset group...")
        arcpy.AddSubtype_management("PipelineDevice", iAssetGroup, sAssetGroup)

        # Add asset types
        print("   Adding subnetwork device asset types...")
        sDomainName = "Asset Type Pipeline Device " + sAssetGroup
        sDesc = "The Asset Types for the PipelineDevice's Asset Group " + sAssetGroup
        Values = [0, 1001, 1002]
        Labels = ["Unknown", "System", "Pressure"]
        arcpy.CreateDomain_management(sTargetGDB, sDomainName, sDesc, "SHORT", "CODED")
        for i in range(len(Values)):
            arcpy.AddCodedValueToDomain_management(sTargetGDB, sDomainName, Values[i], Labels[i])
        arcpy.AssignDomainToField_management("PipelineDevice", "ASSETTYPE", sDomainName, iAssetGroup)

        # Add rules
        print("   Adding rules...")
        Fields = ["rule_type", "from_domain_network", "from_feature_class", "from_asset_group", 
                  "from_asset_type", "from_terminal", "to_domain_network", "to_feature_class", 
                  "to_asset_group", "to_asset_type", "to_terminal"]
        cursor = arcpy.da.InsertCursor("B_Rules", Fields)
        row = ("Junction Junction Connectivity", 1, "Device", sAssetGroup, "*",
               "*", 1, "Junction", "Connection Point", "*", None)
        cursor.insertRow(row)
        row = ("Containment", 1, "Assembly", "Town Border Station", "*",
               None, 1, "Device", sAssetGroup, "*", "*")
        cursor.insertRow(row)
        del cursor

        # Add terminal configuration
        print("   Adding terminal configuration...")
        Fields = ["domain_network", "asset_group", "asset_type", "terminal_configuration_name"]
        cursor = arcpy.da.InsertCursor("B_TerminalConfiguration_Assignment", Fields)
        row = (1, sAssetGroup, "System", "Pipe Bidirectional Dual Terminal")
        cursor.insertRow(row)
        row = (1, sAssetGroup, "Pressure", "Pressure Reducing")
        cursor.insertRow(row)
        del cursor

        # Add subnetwork controller category
        print("   Adding subnetwork controller category...")
        Fields = ["domain_network", "feature_class", "asset_group", "asset_type", "category_name"]
        cursor = arcpy.da.InsertCursor("B_NetworkCategory_Assignment", Fields)
        row = (1, "Device", sAssetGroup, "System", "Subnetwork Controller")
        cursor.insertRow(row)
        row = (1, "Device", sAssetGroup, "Pressure", "Subnetwork Controller")
        cursor.insertRow(row)
        del cursor

        # Add to subnetworks
        Fields = ["tier_name", "asset_group", "asset_type", "valid_subnetwork_controller"]
        cursor = arcpy.da.InsertCursor("B_Subnetwork_Devices", Fields)
        row = ("Pipe Transmission System", sAssetGroup, "System", 1)
        cursor.insertRow(row)
        row = ("Pipe Transmission System", sAssetGroup, "Pressure", 0)
        cursor.insertRow(row)
        row = ("Pipe Transmission Pressure", sAssetGroup, "System", 0)
        cursor.insertRow(row)
        row = ("Pipe Transmission Pressure", sAssetGroup, "Pressure", 1)
        cursor.insertRow(row)
        row = ("Pipe Distribution System", sAssetGroup, "System", 1)
        cursor.insertRow(row)
        row = ("Pipe Distribution System", sAssetGroup, "Pressure", 0)
        cursor.insertRow(row)
        row = ("Pipe Distribution Pressure", sAssetGroup, "System", 0)
        cursor.insertRow(row)
        row = ("Pipe Distribution Pressure", sAssetGroup, "Pressure", 1)
        cursor.insertRow(row)
        del cursor
        return

    def Migrate_Data(self, bDeleteOnly = False):
        dtStart = datetime.now()
        sTargetGDB = self._sWorkingDir + "/" + self._sTargetGDBName

        # Disable editor tracking
        print("   Disabling editor tracking...")
        for sName in self.__EditTrackingDatasets:
            sDS = sTargetGDB + "/" + sName
            self.SetEditorTracking(sDS, False)

        # Truncate tables
        print("   Truncating tables...")
        TruncateList = []
        TruncateList.extend(self.__EditTrackingDatasets)
        TruncateList.extend(["C_Associations", "C_SubnetworkControllers"])
        for sName in TruncateList:
            sDS = sTargetGDB + "/" + sName
            arcpy.TruncateTable_management(sDS)
        if bDeleteOnly:
            return

        # Execute Migrate_line.py, Migrate_assembly.py, Migrate_device.py, Migrate_junction.py
        print("Migrating lines...")
        from Migrate_line import MigrateLine
        ml = MigrateLine()
        ml.DoIt()
        print("Migrating assemblies...")
        from Migrate_assembly import MigrateAssembly
        ma = MigrateAssembly()
        ma.DoIt()
        print("Migrating devices...")
        from Migrate_device import MigrateDevice
        md = MigrateDevice()
        md.DoIt()
        print("Migrating junctions...")
        from Migrate_junction import MigrateJunction
        mj = MigrateJunction()
        mj.DoIt()

        # Enable editor tracking
        print("   Enabling editor tracking...")
        for sName in self.__EditTrackingDatasets:
            sDS = sTargetGDB + "/" + sName
            self.SetEditorTracking(sDS, True)

        print("Done.")
        dtEnd = datetime.now()
        print("Elapsed time: " + str(dtEnd - dtStart))
        return

    def Create_UN(self):
        # Stage utility network and apply AZGas asset package
        # **** DELETE EXISTING DATABASE FIRST ****
        print("Creating utility network...")
        dtStart = datetime.now()
        # arcpy.env.parallelProcessingFactor = 0

        # Create enterprise GDB and UES user
        print("   Creating GDB...")
        sInst = self._sInstance
        sPass = self._sPortalPass
        sKey = self._sWorkingDir + "/" + self._sKeyCodeFile
        arcpy.management.CreateEnterpriseGeodatabase("SQL_Server", sInst, "AZGas", "DATABASE_AUTH", 
                                                     "sa", sPass, "SDE_SCHEMA", "sde", sPass, None, sKey)
        print("   Creating UES user...")
        sConn = self._sWorkingDir + "/" + self._sAdminConnection
        arcpy.management.CreateDatabaseUser(sConn, "DATABASE_USER", "UES", sPass, None, None)

        # Stage utility network
        print("   Staging utility network...")
        sAssetGDB = self._sWorkingDir + "/" + self._sTargetGDBName
        sTargetGDB = self._sWorkingDir + "/" + self._sDBConnection
        sServiceFC = sAssetGDB + "/ServiceTerritory"
        if not self.__bSignedIn:
            arcpy.SignInToPortal(self._sPortalUrl, self._sPortalUser, self._sPortalPass)
            self.__bSignedIn = True
        arcpy.pt.StageUtilityNetwork(sTargetGDB, sServiceFC, self._sUNDSName, self._sUNName)

        # Apply asset package [WITHOUT DATA]
        print("   Applying asset package...")
        sDS = sTargetGDB + "/" + self._sUNDSName
        sUN = sDS + "/" + self._sUNName
        domain_networks = ["Structure", "Pipeline"]
        arcpy.pt.AssetPackageToUtilityNetwork(sAssetGDB, domain_networks, sUN, False, False)

        print("Done.")
        dtEnd = datetime.now()
        print("Elapsed time: " + str(dtEnd - dtStart))
        return

    def Load_Data(self, bLoadAssociations = True):
        # **** Execute only if not registered as versioned ****
        print("Loading data...")
        dtStart = datetime.now()
        sAssetGDB = self._sWorkingDir + "/" + self._sTargetGDBName
        sTargetGDB = self._sWorkingDir + "/" + self._sDBConnection

        # Disable editor tracking
        print("   Disabling editor tracking...")
        for sName in self.__EditTrackingDatasets:
            sDS = sTargetGDB + "/" + sName
            self.SetEditorTracking(sDS, False)

        # Append data from asset package
        print("   Appending data...")
        bError = False
        arcpy.env.preserveGlobalIds = True
        for sName in self.__EditTrackingDatasets:
            sInput = sAssetGDB + "/" + sName
            sTarget = sTargetGDB + "/" + sName
            result = arcpy.GetCount_management(sInput)
            iInputCount = int(result[0])
            if iInputCount == 0:
                continue
            print("      " + sName)
            arcpy.Append_management(sInput, sTarget, "NO_TEST")
            result = arcpy.GetCount_management(sTarget)
            iTargetCount = int(result[0])
            if iTargetCount != iInputCount:
                print("      ERROR: Record count mismatch: " + sName)
                bError = True

        # Import subnetwork controllers
        self.Import_Controllers()

        # Import associations
        if bLoadAssociations and not bError:
            self.Import_Associations()
        else:
            ("   Skipping association import.")

        print("Done.")
        dtEnd = datetime.now()
        print("Elapsed time: " + str(dtEnd - dtStart))
        return

    def Import_Controllers(self):
        # Write subnetwork controller file
        print("   Importing subnetwork controllers...")
        dtStart = datetime.now()
        if not self.__bSignedIn:
            arcpy.SignInToPortal(self._sPortalUrl, self._sPortalUser, self._sPortalPass)
            self.__bSignedIn = True
        sAssetGDB = self._sWorkingDir + "/" + self._sTargetGDBName
        sTargetGDB = self._sWorkingDir + "/" + self._sDBConnection
        sDS = sTargetGDB + "/" + self._sUNDSName
        sUN = sDS + "/" + self._sUNName
        sControllerTab = sAssetGDB + "/C_SubnetworkControllers"
        result = arcpy.GetCount_management(sControllerTab)
        iRecCount = int(result[0])
        if iRecCount == 0:
            return
        in_fields = ["CONTROLLER_NAME", "GLOBAL_ID", "DOMAIN_NETWORK", "ASSET_GROUP",
                     "ASSET_TYPE", "TERMINAL", "TIER_NAME", "SUBNETWORK_NAME"
                     ]
        sCSVFile = self._sWorkingDir + "/Controllers.csv"
        fOut = open(sCSVFile, "w")
        sRow = "SUBNETWORKCONTROLLERNAME,FEATUREGLOBALID,FEATURECLASSNAME,FEATUREASSETGROUP,"
        sRow += "FEATUREASSETTYPE,FEATURETERMINAL,TIERNAME,SUBNETWORKNAME,DESCRIPTION,NOTES\n"
        fOut.write(sRow)
        fOut.flush()
        with arcpy.da.SearchCursor(sControllerTab, in_fields) as cursor:
            for row in cursor:
                sRow = ""
                sRow += row[0] + "," + row[1] + ",PipelineDevice," + row[3] + ","
                sRow += row[4] + "," + row[5] + "," + row[6] + "," + row[7] + ",,\n"
                fOut.write(sRow)
                fOut.flush()
                # Fewer than 1000 records
        fOut.close()

        # Import subnetwork controllers
        arcpy.ImportSubnetworkControllers_un(sUN, sCSVFile)
        print("   Done.")
        dtEnd = datetime.now()
        print("   Elapsed time: " + str(dtEnd - dtStart))
        return

    def Import_Associations(self, bStep = True):
        # Write associations file
        dtStart = datetime.now()
        print("   Importing associations...")
        if not self.__bSignedIn:
            arcpy.SignInToPortal(self._sPortalUrl, self._sPortalUser, self._sPortalPass)
            self.__bSignedIn = True
        sAssetGDB = self._sWorkingDir + "/" + self._sTargetGDBName
        sTargetGDB = self._sWorkingDir + "/" + self._sDBConnection
        sDS = sTargetGDB + "/" + self._sUNDSName
        sUN = sDS + "/" + self._sUNName
        sAssociationTab = sAssetGDB + "/C_Associations"
        result = arcpy.GetCount_management(sAssociationTab)
        iRecCount = int(result[0])
        if iRecCount == 0:
            return
        in_fields = ["ASSOCIATION_TYPE", "FROM_DOMAIN_NETWORK", "FROM_FEATURE_CLASS", "FROM_ASSET_GROUP",
                      "FROM_ASSET_TYPE", "FROM_GLOBAL_ID", "FROM_TERMINAL", "TO_DOMAIN_NETWORK", 
                      "TO_FEATURE_CLASS", "TO_ASSET_GROUP", "TO_ASSET_TYPE", "TO_GLOBAL_ID", 
                      "TO_TERMINAL", "CONTENT_VISIBLE"
                      ]
        sCSVFile = self._sWorkingDir + "/Associations.csv"
        bFirst = True
        iCount = 0
        with arcpy.da.SearchCursor(sAssociationTab, in_fields) as cursor:
            for row in cursor:
                if bFirst:
                    fOut = open(sCSVFile, "w")
                    sRow = "ASSOCIATIONTYPE,FROMFEATURECLASS,FROMASSETGROUP,FROMASSETTYPE,"
                    sRow += "FROMGLOBALID,FROMTERMINAL,TOFEATURECLASS,TOASSETGROUP,"
                    sRow += "TOASSETTYPE,TOGLOBALID,TOTERMINAL,ISCONTENTVISIBLE\n"
                    fOut.write(sRow)
                    fOut.flush()
                    bFirst = False
                sRow = ""
                sRow += row[0] + ",Pipeline" + row[2] + "," + row[3] + ","
                sTerm = row[6]
                if sTerm is None:
                    sTerm = ""
                sRow += row[4] + "," + row[5] + "," + sTerm + ",Pipeline" + row[8] + ","
                sTerm = row[12]
                if sTerm is None:
                    sTerm = ""
                sRow += row[9] + "," + row[10] + "," + row[11] + "," + sTerm + ","
                if row[13] == 1:
                    sRow += "True"
                elif row[13] == 0:
                    sRow += "False"
                sRow += "\n"
                fOut.write(sRow)
                fOut.flush()
                iCount += 1
                if bStep and not (iCount % 1000 == 0 or iCount == iRecCount):
                    continue
                elif not bStep and iCount < iRecCount:
                    continue
                fOut.close()
                bFirst = True

                # Import associations
                print("   Importing associations..." + str(iCount))
                arcpy.ImportAssociations_un(sUN, "ALL", sCSVFile)

        print("   Done.")
        dtEnd = datetime.now()
        print("   Elapsed time: " + str(dtEnd - dtStart))
        return

    def Register_Versioned(self, bRegister):
        print("Register as versioned: " + str(bRegister) + "...")
        dtStart = datetime.now()
        sTargetGDB = self._sWorkingDir + "/" + self._sDBConnection
        sUNDS = sTargetGDB + "/" + self._sUNDSName
        if bRegister:
            print("   Enabling editor tracking...")
            for sName in self.__EditTrackingDatasets:
                sDS = sTargetGDB + "/" + sName
                self.SetEditorTracking(sDS, True)
            print("   Registering " + self._sUNDSName + " as versioned...")
            arcpy.RegisterAsVersioned_management(sUNDS)
        else:
            print("   Unregistering " + self._sUNDSName + " as versioned...")
            arcpy.UnregisterAsVersioned_management(sUNDS)
            print("   Disabling editor tracking...")
            for sName in self.__EditTrackingDatasets:
                sDS = sTargetGDB + "/" + sName
                self.SetEditorTracking(sDS, False)
        print("Done.")
        dtEnd = datetime.now()
        print("Elapsed time: " + str(dtEnd - dtStart))
        return

    def Enable_Topology(self, bErrorsOnly):
        dtStart = datetime.now()
        if not self.__bSignedIn:
            arcpy.SignInToPortal(self._sPortalUrl, self._sPortalUser, self._sPortalPass)
            self.__bSignedIn = True
        sTargetGDB = self._sWorkingDir + "/" + self._sDBConnection
        sDS = sTargetGDB + "/" + self._sUNDSName
        sUN = sDS + "/" + self._sUNName
        if bErrorsOnly:

            # Check network topology
            # **** PRO 2.3.0/2.3.1 BUG: "ONLY_ERRORS" option will fail ****
            print("Checking network topology...")
            try:
                if self._bTopoBug:
                    print("   [BUG WORKAROUND]")
                    self.Register_Versioned(True)
                    self.Enable_Topology(False)
                    self.Disable_Topology()
                    self.Register_Versioned(False)
                else:
                    arcpy.EnableNetworkTopology_un(sUN, "", "ONLY_ERRORS")
            except Exception as ErrorDesc:
                sMsg = "Topology check failed:"
                sMsg += "\n" + str(ErrorDesc)
                #sMsg += "\n" + traceback.format_exc()
                print(sMsg)

            # Get the UN ID
            sID = self.GetUNID()
            if sID is None:
                return
            sUNPrefix = "UN_" + sID + "_"

            # Report errors [Only works properly when non-versioned]
            print("   Topology errors:")
            ex = arcpy.ArcSDESQLExecute(sTargetGDB)
            iCount = 0
            for sName in ["PointErrors", "LineErrors"]:
                sTabName = sUNPrefix + sName
                sSQL = "SELECT errortype, COUNT(errortype) AS count FROM " + sTabName
                sSQL += " GROUP BY errortype ORDER BY errortype"
                result = ex.execute(sSQL)
                if result is True:
                    continue # No errors!
                print("   " + sName + ":")
                print("      ERRORTYPE - COUNT")
                for row in result:
                    print("      " + str(row[0]) + " - " + str(row[1]))
                    iCount += row[1]
            print("   Total errors: " + str(iCount))
        else:
            print("Enabling network topology...")
            try:
                arcpy.EnableNetworkTopology_un(sUN)
            except Exception as ErrorDesc:
                sMsg = "Enable topology failed:"
                sMsg += "\n" + str(ErrorDesc)
                sMsg += "\n" + traceback.format_exc()
                print(sMsg)

        print("Done.")
        dtEnd = datetime.now()
        print("Elapsed time: " + str(dtEnd - dtStart))
        return

    def Disable_Topology(self):
        dtStart = datetime.now()
        if not self.__bSignedIn:
            arcpy.SignInToPortal(self._sPortalUrl, self._sPortalUser, self._sPortalPass)
            self.__bSignedIn = True
        sTargetGDB = self._sWorkingDir + "/" + self._sDBConnection
        sDS = sTargetGDB + "/" + self._sUNDSName
        sUN = sDS + "/" + self._sUNName
        print("Disabling network topology...")
        arcpy.DisableNetworkTopology_un(sUN)
        print("Done.")
        dtEnd = datetime.now()
        print("Elapsed time: " + str(dtEnd - dtStart))
        return

    def Fix_Errors(self):
        # Fix geometry and stacked point errors
        print("Fixing geometry and point errors...")
        dtStart = datetime.now()
        from Fix_GeometryErrors import GeometryHelper
        gh = GeometryHelper()
        gh.FixGeometryErrors(False)
        from Fix_PointErrors import PointHelper
        ph = PointHelper()
        ph.FixPointErrors(False)
        print("Done.")
        dtEnd = datetime.now()
        print("Elapsed time: " + str(dtEnd - dtStart))
        return

    def Export_AZGas(self, bExportData):
        # Export asset package
        print("Exporting AZGas asset package...")
        dtStart = datetime.now()
        sSourceGDB = self._sWorkingDir + "/" + self._sDBConnection
        sDS = sSourceGDB + "/" + self._sUNDSName
        sUN = sDS + "/" + self._sUNName
        domain_networks = ["Pipeline"]
        sOutName = "AssetPackage_AZGas_Export"
        sAssetGDB = self._sWorkingDir + "/" + sOutName
        if arcpy.Exists(sAssetGDB):
            arcpy.Delete_management(sAssetGDB)
        arcpy.pt.UtilityNetworkToAssetPackage(sUN, domain_networks, self._sWorkingDir, sOutName, bExportData)
        print("Done.")
        dtEnd = datetime.now()
        print("Elapsed time: " + str(dtEnd - dtStart))

print("Ready.")
#if __name__ == "__main__":
#    mh = MigrateHelper()
