# Migrate_junction.py [Python 3.6]

# Feature classes migrated [-> PipelineJunction]: 
#   NonControllableFitting
#   MainJobSeparator
#   GasGeomNetwork_Junctions

print("Migrate_junction.py")
print("Initalizing...")


import arcpy
from datetime import datetime
import traceback
import uuid
from Migrate_helper import MigrateHelper

class MigrateJunction(object):
    # Constants
    _bSkipProposed = True

    # Private fields
    __helper = None
    __sSourceGDB = ""
    __sTargetGDB = ""
    __sTargetFC = ""
    __cursor = None

    def __init__(self):
        self.__helper = MigrateHelper()
        self.__sSourceGDB = MigrateHelper._sSourceGDB
        self.__sTargetGDB = MigrateHelper._sWorkingDir + "/" + MigrateHelper._sTargetGDBName
        self.__sTargetFC = self.__sTargetGDB + "/PipelineJunction"
        return

    def DoIt(self, bDeleteOnly = False):
        print("Running...")
        dtStart = datetime.now()

        # Populate non-controllable fittings
        with arcpy.da.Editor(self.__sTargetGDB) as edit:
            self.__MigrateNonControllableFitting()

        # Populate main job separators
        with arcpy.da.Editor(self.__sTargetGDB) as edit:
            self.__MigrateMainJobSeparator()

        # Populate unspecified junctions
        with arcpy.da.Editor(self.__sTargetGDB) as edit:
            self.__MigrateJunctions()

        dtEnd = datetime.now()
        print("Done.")
        print("Elapsed time: " + str(dtEnd - dtStart))
        return

    def __MigrateNonControllableFitting(self):
        sFCName = "NonControllableFitting"
        print("Migrating " + sFCName + "...")

        # Create InsertCursor
        out_fields = ["ASSETGROUP", "ASSETTYPE", "CREATIONDATE", "CREATOR", 
                      "LASTUPDATE", "UPDATEDBY", "DESIGNTYPE", "INSULATORDEVICE", 
                      "INSERVICEDATE", "LIFECYCLESTATUS", "DIAMETER", "ROTATION", 
                      "WORKORDERID", "STORES_STOCK_NUMBER", "SEQUENTIAL_FILE_NUMBER", "CHART_DATE", 
                      "SHAPE@", "GLOBALID"
                      ]
        self.__cursor = arcpy.da.InsertCursor(self.__sTargetFC, out_fields)

        # Get projected shapes
        print("   Getting projected shapes...")
        ProjectedShapes = self.__helper.GetProjectedShapes(sFCName)

        # Create SearchCursor
        print("   Processing rows...")
        sSourceFC = self.__sSourceGDB + "/" + sFCName
        in_fields = ["SUBTYPECD", "MATERIAL", "DATECREATED", "CREATIONUSER", 
                     "DATEMODIFIED", "LASTUSER", "FITTINGDIAMETER", "INSULATEDINDICATOR", 
                     "INSERVICEDATE", "WORKFUNCTION", "SYMBOLROTATION", "WORKORDERID", 
                     "STORES_STOCK_NUMBER", "SEQUENTIAL_FILE_NUMBER", "CHART_DATE", "GLOBALID"
                     ]
        iCount = 0
        with arcpy.da.SearchCursor(sSourceFC, in_fields) as cursorIn:
            for in_row in cursorIn:
                sGlobalID = in_row[15]
                if sGlobalID not in ProjectedShapes:
                    continue
                shape = ProjectedShapes[sGlobalID]
                self.__ProcessFittingRow(in_row, shape)
                iCount += 1
                if iCount % 1000 == 0:
                    print("   NonControllableFitting row " + str(iCount) + "...")
        del self.__cursor
        return

    def __ProcessFittingRow(self, in_row, shape):

        # Subtype
        iAssetGroup = 0
        iAssetType = 0
        iSubtype = in_row[0]
        sMat = in_row[1]
        bReducer = False
        if iSubtype == 1: # "Coupling"
            iAssetGroup = 1 # "Coupling"
            if sMat == "MTL":
                iAssetType = 1 # "Metal Coupling"
            elif sMat == "PL":
                iAssetType = 2 # "Plastic Coupling"
        elif iSubtype == 2: # "Ell"
            iAssetGroup = 2 # "Elbow"
            if sMat == "MTL":
                iAssetType = 41 # "Metal Elbow"
            elif sMat == "PL":
                iAssetType = 42 # "Plastic Elbow"
        elif iSubtype == 3: # "End Cap"
            iAssetGroup = 3 # "End Cap"
            if sMat == "MTL":
                iAssetType = 81 # "Metal End Cap"
            elif sMat == "PL":
                iAssetType = 82 # "Metal End Cap"
        elif iSubtype == 6: # "Reducer"
            iAssetGroup = 6 # "Reducer"
            if sMat == "MTL":
                iAssetType = 201 # "Metal Reducer"
            elif sMat == "PL":
                iAssetType = 202 # "Metal Reducer"
            bReducer = True
        elif iSubtype == 7: # "Tee"
            iAssetGroup = 7 # "Tee"
            if sMat == "MTL":
                iAssetType = 241 # "Metal 3-Way"
            elif sMat == "PL":
                iAssetType = 243 # "Plastic 3-Way"
        elif iSubtype == 8: # "Transition"
            iAssetGroup = 8 # "Transition"
            iAssetType = 281 # "Transition"
        elif iSubtype == 9: # "Weld End Insulator"
            iAssetGroup = 11 # "Electro Stop"
            iAssetType = 401 # "Electro Stop"
        elif iSubtype == 10: # "Branch Saddle"
            iAssetGroup = 13 # "Saddle"
            if sMat == "MTL":
                iAssetType = 481 # "Metal Saddle"
            elif sMat == "PL":
                iAssetType = 482 # "Plastic Saddle"

        dtCreated = in_row[2]
        sCreationUser = in_row[3]
        dtModified = in_row[4]
        sLastUser = in_row[5]

        # Fitting diameter or reducer design type
        iDesignType = None
        dDiameter = None
        sDiameter = in_row[6]
        if bReducer:
            iDesignType = self.__GetReducerType(sDiameter)
        else:
            if sDiameter.find("x") > -1:
                sDiameter = sDiameter.split("x")[0]
            dDiameter = self.__helper.StringToFloat(sDiameter)

        # Domain: UPDM_INSULATORDEVICE
        iInsulated = None
        sInsulated = in_row[7]
        if sInsulated == "No":
            iInsulated = 1
        elif sInsulated == "Yes":
            iInsulated = 2;

        dtInService = in_row[8]

        # Domain: UPDM_LifeCycleStatus
        iLifeCycle = 3 # "In Service"
        iWorkFunction = in_row[9]
        if iWorkFunction == 1:
            if self._bSkipProposed:
                return
            iLifeCycle = 1 # "Proposed"

        dSymbolRotation = in_row[10]
        sWorkOrderID = in_row[11]
        sSSN = in_row[12]
        sSFN = in_row[13]
        dtChart = in_row[14]
        sGlobalID = in_row[15]

        out_row = [iAssetGroup, iAssetType, dtCreated, sCreationUser, 
                    dtModified, sLastUser, iDesignType, iInsulated,
                    dtInService, iLifeCycle, dDiameter, dSymbolRotation, 
                    sWorkOrderID, sSSN, sSFN, dtChart, 
                    shape, sGlobalID
                    ]
        self.__cursor.insertRow(out_row)
        return

    def __MigrateMainJobSeparator(self):
        sFCName = "MainJobSeparator"
        print("Migrating " + sFCName + "...")

        # Create InsertCursor
        out_fields = ["ASSETGROUP", "ASSETTYPE", "CREATIONDATE", "CREATOR", 
                      "LASTUPDATE", "UPDATEDBY", "INSERVICEDATE", "LIFECYCLESTATUS", 
                      "WORKORDERID", "SEQUENTIAL_FILE_NUMBER", "CHART_DATE", "SHAPE@",
                      "GLOBALID"
                      ]
        self.__cursor = arcpy.da.InsertCursor(self.__sTargetFC, out_fields)

        # Spatially join with mains
        print("   Joining mains...")
        sUTMGDB = self.__helper._sWorkingDir + "/" + self.__helper._sUTM84GDBName
        sInLayer = self.__helper.GetProjectedLayer(sFCName)
        sJoinFC = self.__sTargetGDB + "/PipelineLine"
        sJoinLayer = "lyr_" + str(uuid.uuid4())
        sWhere = "assetgroup IN (2, 3)"
        sSourceFC = sUTMGDB + "/MainJobSeparatorPipe"
        if arcpy.Exists(sSourceFC):
            arcpy.Delete_management(sSourceFC)
        arcpy.MakeFeatureLayer_management(sJoinFC, sJoinLayer, sWhere)
        arcpy.SpatialJoin_analysis(sInLayer, sJoinLayer, sSourceFC)
        arcpy.AddGlobalIDs_management(sSourceFC)

        print("   Processing rows...")
        in_fields = ["ASSETTYPE", "DATECREATED", "CREATIONUSER", "DATEMODIFIED", 
                     "LASTUSER", "INSERVICEDATE", "WORKFUNCTION", "WORKORDERID", 
                     "SEQUENTIAL_FILE_NUMBER", "CHART_DATE", "SHAPE@", "GLOBALID"
                     ]
        iCount = 0
        with arcpy.da.SearchCursor(sSourceFC, in_fields) as cursorIn:
            for in_row in cursorIn:
                self.__ProcessSeparatorRow(in_row)
                iCount += 1
                if iCount % 1000 == 0:
                    print("   MainJobSeparator row " + str(iCount) + "...")
        del self.__cursor
        return

    def __ProcessSeparatorRow(self, in_row):
        iMaterial = in_row[0]
        if iMaterial is None:
            return # This should never happen
        if iMaterial == 3: # "Coated Steel"
            iAssetGroup = 9 # "Weld"
            iAssetType = 321 # "Weld"
        else:
            iAssetGroup = 12 # "Plastic Fusion"
            iAssetType = 441 # "Plastic Fusion"
        dtCreated = in_row[1]
        sCreationUser = in_row[2]
        dtModified = in_row[3]
        sLastUser = in_row[4]
        dtInService = in_row[5]

        # Domain: UPDM_LifeCycleStatus
        iLifeCycle = 3 # "In Service"
        iWorkFunction = in_row[6]
        if iWorkFunction == 1:
            if self._bSkipProposed:
                return
            iLifeCycle = 1 # "Proposed"

        sWorkOrderID = in_row[7]
        sSFN = in_row[8]
        dtChart = in_row[9]
        shape = in_row[10]
        sGlobalID = in_row[11]

        out_row = [iAssetGroup, iAssetType, dtCreated, sCreationUser, 
                    dtModified, sLastUser, dtInService, iLifeCycle, 
                    sWorkOrderID, sSFN, dtChart, shape,
                    sGlobalID
                    ]
        self.__cursor.insertRow(out_row)
        return

    def __MigrateJunctions(self):
        sFCName = "GasGeomNetwork_Junctions_bak"
        print("Migrating " + sFCName + "...")

        # Create InsertCursor
        out_fields = ["ASSETGROUP", "ASSETTYPE", "SHAPE@", "GLOBALID"]
        self.__cursor = arcpy.da.InsertCursor(self.__sTargetFC, out_fields)

        # Create SearchCursor
        sSourceLayer = self.__helper.GetProjectedLayer(sFCName)
        in_fields = ["SHAPE@"]
        iCount = 0
        with arcpy.da.SearchCursor(sSourceLayer, in_fields) as cursorIn:
            for in_row in cursorIn:
                shape = in_row[0]
                self.__ProcessJunctionRow(shape)
                iCount += 1
                if iCount % 1000 == 0:
                    print("   GasGeomNetwork_Junctions_bak row " + str(iCount) + "...")
        del self.__cursor
        return

    def __ProcessJunctionRow(self, shape):
        # Generate global ID
        uid = uuid.uuid4()
        sGID = "{" + str(uid).upper() + "}"
        iAssetGroup = 21 # Connection Point
        iAssetType = 801 # Connection Point
        out_row = [iAssetGroup, iAssetType, shape, sGID]
        self.__cursor.insertRow(out_row)
        return

    def __GetReducerType(self, sDia):

        # Domain:  UPDM_ReducerFittingSize
        iType = None
        if sDia == "1x0.5":
            iType = 1 # 1" x 1/2"
        elif sDia == "1x0.75":
            iType = 3 # 1" x 3/4"
        elif sDia == "1.25x1":
            iType = 8 # 1.25" x 1"
        elif sDia == "2x0.5":
            iType = 101 # Add: 2" x 1/2"
        elif sDia == "2x0.75":
            iType = 102 # Add: 2" x 3/4"
        elif sDia == "2x1":
            iType = 12 # 2" x 1"
        elif sDia == "2x1.25":
            iType = 13 # 2" x 1 1/4"
        elif sDia == "3x1":
            iType = 103 # Add: 3" x 1"
        elif sDia == "3x2":
            iType = 16 # 3" x 2"
        elif sDia == "4x1":
            iType = 104 # Add: 4" x 1"
        elif sDia == "4x2":
            iType = 17 # 4" x 2"
        elif sDia == "4x3":
            iType = 18 # 4" x 3"
        elif sDia == "6x2":
            iType = 105 # Add: 6" x 2"
        elif sDia == "6x3":
            iType = 19 # 6" x 3"
        elif sDia == "6x4":
            iType = 20 # 6" x 4"
        elif sDia == "8x4":
            iType = 21 # 8" x 4"
        elif sDia == "8x6":
            iType = 22 # 8" x 6"
        elif sDia == "10x4":
            iType = 106 # Add: 10" x 4"
        elif sDia == "10x8":
            iType = 24 # 10" x 8"
        elif sDia == "16x6":
            iType = 107 # Add: 16" x 6"
        elif sDia == "16x8":
            iType = 31 # 16" x 8"
        elif sDia == "UNK":
            iType = 0 # Unknown
        return iType

print("Ready.")
if __name__ == "__main__":
    mj = MigrateJunction()
    mj.DoIt()
