﻿# Migrate_line.py [Python 3.6]

# Feature classes migrated [-> PipelineLine]: 
#   DistributionMain 
#   SubTransmissionMain
#   TransmissionMain
#   AbandonedDistributionMain
#   AbandonedSubTransmissionMain
#   AbandonedTransmissionMain
#   Service
#   AbandonedService
#   CPBondWire

print("Migrate_line.py")
print("Initalizing...")

import arcpy
from datetime import datetime
import traceback
from Migrate_helper import MigrateHelper

class MigrateLine(object):
    # Constants
    _bSkipProposed = True
    _bSkipAbandoned = True

    # Private fields
    __helper = None
    __sSourceGDB = ""
    __sTargetGDB = ""
    __sTargetFC = ""
    __bGetMaterial = False
    __bAsBuilt = False
    __bTransmission = False
    __iAssetGroup = 0
    __cursor = None

    def __init__(self):
        self.__helper = MigrateHelper()
        self.__sSourceGDB = MigrateHelper._sSourceGDB
        self.__sTargetGDB = MigrateHelper._sWorkingDir + "/" + MigrateHelper._sTargetGDBName
        self.__sTargetFC = self.__sTargetGDB + "/PipelineLine"
        self.__sSumFC = self.__sTargetGDB + "/PipelineServiceSummary"
        return

    def DoIt(self):
        print("Running...")
        dtStart = datetime.now()

        # Populate mains
        sFC_Dist = "DistributionMain"
        sFC_HP = "SubTransmissionMain"
        sFC_Trans = "TransmissionMain"
        for sFC in [sFC_Dist, sFC_HP, sFC_Trans]:
            with arcpy.da.Editor(self.__sTargetGDB) as editAsBuilt:
                self.__MigrateMainFC(sFC, False)
            if self._bSkipAbandoned:
                continue
            with arcpy.da.Editor(self.__sTargetGDB) as editAbandon:
                self.__MigrateMainFC(sFC, True)

        # Populate services
        with arcpy.da.Editor(self.__sTargetGDB) as editAsBuilt:
            self.__MigrateService(False)
        if not self._bSkipAbandoned:
            with arcpy.da.Editor(self.__sTargetGDB) as editAbandon:
                self.__MigrateService(True)

        # Populate CP bond wire
        with arcpy.da.Editor(self.__sTargetGDB) as edit:
            self.__MigrateCPBondWire()

        # Repair line geometry
        print("   Repairing geometry...")
        arcpy.RepairGeometry_management(self.__sTargetFC)

        dtEnd = datetime.now()
        print("Done.")
        print("Elapsed time: " + str(dtEnd - dtStart))
        return

    def __MigrateMainFC(self, sFC, bAbandoned):

        # Subtypes
        self.__bTransmission = False
        if sFC == "DistributionMain":
            self.__bGetMaterial = True
            self.__iAssetGroup = 2 # "Distribution Pipe"
        elif sFC == "SubTransmissionMain":
            self.__bGetMaterial = False
            self.__iAssetGroup = 2 # "Distribution Pipe"
        else:
            self.__bGetMaterial = False
            self.__iAssetGroup = 3 # "Transmission Pipe"
            self.__bTransmission = True

        sFCName = sFC
        if bAbandoned:
            sFCName = "Abandoned" + sFC
            self.__bAsBuilt = False;
        else:
            self.__bAsBuilt = True
        print("   Migrating " + sFCName + "...")
        sSourceFC = self.__sSourceGDB + "/" + sFCName

        # Get projected shapes
        print("   Getting projected shapes...")
        ProjectedShapes = self.__helper.GetProjectedShapes(sFCName)

        # Create InsertCursor
        print("   Processing rows...")
        out_fields = ["ASSETGROUP", "ASSETTYPE", "CREATIONDATE", "CREATOR", 
                      "LASTUPDATE", "UPDATEDBY", "INSERVICEDATE", "LIFECYCLESTATUS", 
                      "NOMINALDIAMETER", "MAOPRECORD", "MEASUREDLENGTH", "LENGTHSOURCE", 
                      "WALLTHICKNESS", "ACTUALINTERNALDIAMETER", "PIPETESTDATE", "PIPETESTPRESSURE",
                      "DEPTH", "WORKORDERID", "STORES_STOCK_NUMBER", "SYSTEM_NAME",
                      "SYSTEM_SECTION", "SEQUENTIAL_FILE_NUMBER", "INSTALLED_BY", "STATUS_DATE", 
                      "LOCATABLE", "LS_MAINLINE", "CP_TOWN_NAME", "CP_CIRCUIT_NUMBER",
                      "CP_COLOR", "SHAPE@", "GLOBALID" 
                      ]
        if bAbandoned:
            out_fields.extend(["DATERETIRED"])
        if self.__bTransmission:
            out_fields.extend(["ASSETID"])
        self.__cursor = arcpy.da.InsertCursor(self.__sTargetFC, out_fields)

        # Create SearchCursor
        in_fields = ["DATECREATED", "CREATIONUSER", "DATEMODIFIED", "LASTUSER", 
                     "INSERVICEDATE", "WORKFUNCTION", "NOMINALDIAMETER", "SOPENTERED", 
                     "MEASUREDLENGTH", "LENGTHSOURCE", "WALLTHICKNESS", "ACTUALINTERNALDIAMETER", 
                     "PIPETESTDATE", "PIPETESTPRESSURE", "DEPTH", "WORKORDERID", 
                     "STORES_STOCK_NUMBER", "SYSTEM_NAME", "SYSTEM_SECTION", "SEQUENTIAL_FILE_NUMBER", 
                     "INSTALLED_BY", "STATUS_DATE", "LOCATABLE", "LS_MAINLINE",
                     "CP_TOWN_NAME", "CP_CIRCUIT_NUMBER", "CP_COLOR", "GLOBALID"
                     ]
        if self.__bGetMaterial:
            in_fields.extend(["MATERIAL", "PLASTICTYPE"])
        if bAbandoned:
            in_fields.extend(["RETIREMENT_DATE"])
        if self.__bTransmission:
            in_fields.extend(["OBJECTID"])

        iCount = 0
        with arcpy.da.SearchCursor(sSourceFC, in_fields) as c:
            for in_row in c:
                sGlobalID = in_row[27]
                if sGlobalID not in ProjectedShapes:
                    continue
                shape = ProjectedShapes[sGlobalID]
                self.__ProcessMainRow(in_row, shape)
                iCount += 1
                if iCount % 1000 == 0:
                    print("   " + sFCName + " row " + str(iCount) + "...")
        del self.__cursor
        return

    def __ProcessMainRow(self, in_row, shape):
        dtCreated = in_row[0]
        sCreationUser = in_row[1]
        dtModified = in_row[2]
        sLastUser = in_row[3]
        dtInService = in_row[4]

        # Domain: UPDM_LifeCycleStatus
        iLifeCycle = 0 # "Unknown"
        if self.__bAsBuilt:
            iLifeCycle = 3 # "In Service"
            iWorkFunction = in_row[5]
            if iWorkFunction == 1:
                if self._bSkipProposed:
                    return 
                iLifeCycle = 1 # "Proposed"
        else:
            iLifeCycle = 7 # "Abandoned"

        sNominalDiameter = in_row[6]
        dNominalDiameter = self.__helper.StringToFloat(sNominalDiameter)
        dMAOP = None
        if self.__bAsBuilt:
            dMAOP = in_row[7]
        dMeasuredLength = in_row[8]

        # Domain: UPDM_LengthSource
        sLS = in_row[9]
        iLengthSource = None
        if sLS == "MS":
            iLengthSource = 7 # "Mapping System"

        sWallThickness = in_row[10]
        dWallThickness = self.__helper.StringToFloat(sWallThickness)
        dActualInternalDiameter = in_row[11]
        dtPipeTest = in_row[12]
        sPipeTestPressure = in_row[13]
        dPipeTestPressure = self.__helper.StringToFloat(sPipeTestPressure)
        iDepth = in_row[14]
        sWorkOrderID = in_row[15]
        sStoresStockNumber = in_row[16]
        sSystemName = None
        iSystemSection = None
        if self.__bAsBuilt:
            sSystemName = in_row[17]
            iSystemSection = in_row[18]
        sSequentialFileNumber = in_row[19]
        sInstalledBy = in_row[20]
        dtStatus = in_row[21]

        # Domain: UPDM_YesNoIndicator
        iLocatable = None
        sLoc = in_row[22]
        if sLoc == "Yes":
            iLocatable = 1
        elif sLoc == "No":
            iLocatable = 2
        iMainline = None
        sML = in_row[23]
        if sML == "Yes":
            iMainline = 1
        elif sML == "No":
            iMainline = 2

        sCPTown = None
        sCPCircuit = None
        sCPColor = None
        if self.__bAsBuilt:
            sCPTown = in_row[24]
            sCPCircuit = in_row[25]
            sCPColor = in_row[26]
        sGlobalID = in_row[27]
        iEnd = 28

        # Domain: Asset Type Pipeline Line Distribution [or Transmission] Pipe
        iAssetType = None
        if self.__bGetMaterial:
            iMaterial = in_row[28]
            if iMaterial == 3: # "Coated Steel"
                iAssetType = 3 # "Coated Steel"
            elif iMaterial == 5: # "Plastic"
                iAssetType = None
                sPlasticType = in_row[29]
                if sPlasticType == "POLY":
                    iAssetType = 9 # "Plastic PE"
                elif sPlasticType == "PVC":
                    iAssetType = 10 # "Plastic PVC"
            iEnd += 2
        else:
            iAssetType = 3 # "Coated Steel"

        dtRetired = None
        if not self.__bAsBuilt:
            dtRetired = in_row[iEnd]
            iEnd += 1

        sFID = None
        if self.__bTransmission:
            iOID = in_row[iEnd]
            sFID = str(iOID)

        # Create feature
        out_row = [self.__iAssetGroup, iAssetType, dtCreated, sCreationUser, 
                   dtModified, sLastUser, dtInService, iLifeCycle,
                   dNominalDiameter, dMAOP, dMeasuredLength, iLengthSource, 
                   dWallThickness, dActualInternalDiameter, dtPipeTest, dPipeTestPressure, 
                   iDepth, sWorkOrderID, sStoresStockNumber, sSystemName,
                   iSystemSection, sSequentialFileNumber, sInstalledBy, dtStatus, 
                   iLocatable, iMainline, sCPTown, sCPCircuit,
                   sCPColor, shape, sGlobalID]
        if not self.__bAsBuilt:
            out_row.extend([dtRetired])
        if self.__bTransmission:
            out_row.extend([sFID])
        self.__cursor.insertRow(out_row)
        return

    def __MigrateService(self, bAbandoned):
        
        sFCName = "Service"
        self.__iAssetGroup = 1 # "Service Line"
        if bAbandoned:
            sFCName = "Abandoned" + sFCName
            self.__bAsBuilt = False;
        else:
            self.__bAsBuilt = True
        print("   Migrating " + sFCName + "...")
        sSourceFC = self.__sSourceGDB + "/" + sFCName

        # Get projected shapes
        print("   Getting projected shapes...")
        ProjectedShapes = self.__helper.GetProjectedShapes(sFCName)

        # Create InsertCursor
        # **** TO DO!!! INCLUDE STUB ATTRIBUTE ****
        print("   Processing rows...")
        out_fields = ["ASSETGROUP", "ASSETTYPE", "CREATIONDATE", "CREATOR", 
                      "LASTUPDATE", "UPDATEDBY", "INSERVICEDATE", "LIFECYCLESTATUS", 
                      "NOMINALDIAMETER", "MEASUREDLENGTH", "LENGTHSOURCE", "WALLTHICKNESS", 
                      "ACTUALINTERNALDIAMETER", "WORKORDERID", "STORES_STOCK_NUMBER", "INSTALLED_BY", 
                      "ISOLATED_MISMATCH", "STATUS_DATE", "LS_MAINLINE", "CP_TOWN_NAME", 
                      "CP_CIRCUIT_NUMBER", "CP_COLOR", "COMMENTS", "SERVICE_CARD", 
                      "SLEEVED", "SHAPE@", "GLOBALID", "STUB"
                      ]
        if bAbandoned:
            out_fields.extend(["DATERETIRED"])
        self.__cursor = arcpy.da.InsertCursor(self.__sTargetFC, out_fields)

        # Create SearchCursor
        in_fields = ["MATERIAL", "PLASTICTYPE", "DATECREATED", "CREATIONUSER", 
                     "DATEMODIFIED", "LASTUSER", "INSERVICEDATE", "WORKFUNCTION", 
                     "NOMINALDIAMETER", "MEASUREDLENGTH", "LENGTHSOURCE", "ACTUALINTERNALDIAMETER", 
                     "WALLTHICKNESS", "WORKORDERID", "STORES_STOCK_NUMBER", "INSTALLED_BY", 
                     "ISOLATED_MISMATCH", "STATUS_DATE", "LS_MAINLINE", "CP_TOWN_NAME", 
                     "CP_CIRCUIT_NUMBER", "CP_COLOR", "COMMENTS", "SERVICE_CARD", 
                     "SLEEVED", "GLOBALID", "SUBTYPECD"
                     ]
        if bAbandoned:
            in_fields.extend(["RETIREMENT_DATE"])

        iCount = 0
        with arcpy.da.SearchCursor(sSourceFC, in_fields) as c:
            for in_row in c:
                sGlobalID = in_row[25]
                if sGlobalID not in ProjectedShapes:
                    continue
                shape = ProjectedShapes[sGlobalID]
                self.__ProcessServiceRow(in_row, shape)
                iCount += 1
                if iCount % 1000 == 0:
                    print("   " + sFCName + " row " + str(iCount) + "...")
        del self.__cursor
        return

    def __ProcessServiceRow(self, in_row, shape):

        # Domain: Asset Type Pipeline Line Service Pipe
        iAssetType = None
        sMaterial = in_row[0]
        if sMaterial == "CS":
            iAssetType = 3 # "Coated Steel"
        elif sMaterial == "PL":
            iAssetType = 9 # "Plastic PE"
            sPlasticType = in_row[1]
            if sPlasticType == "PVC":
                iAssetType = 10 # "Plastic PVC"

        dtCreated = in_row[2]
        sCreationUser = in_row[3]
        dtModified = in_row[4]
        sLastUser = in_row[5]
        dtInService = in_row[6]

        # Domain: UPDM_LifeCycleStatus
        iLifeCycle = 0 # "Unknown"
        if self.__bAsBuilt:
            iLifeCycle = 3 # "In Service"
            iWorkFunction = in_row[7]
            if iWorkFunction == 1:
                if self._bSkipProposed:
                    return 
                iLifeCycle = 1 # "Proposed"
        else:
            iLifeCycle = 7 # "Abandoned"

        sNominalDiameter = in_row[8]
        dNominalDiameter = self.__helper.StringToFloat(sNominalDiameter)
        dMeasuredLength = in_row[9]

        # Domain: UPDM_LengthSource
        sLS = in_row[10]
        iLengthSource = None
        if sLS == "MS":
            iLengthSource = 7 # "Mapping System"

        dActualInternalDiameter = in_row[11]
        sWallThickness = in_row[12]
        dWallThickness = self.__helper.StringToFloat(sWallThickness)
        sWorkOrderID = in_row[13]
        sStoresStockNumber = in_row[14]
        sInstalledBy = in_row[15]

        # Domain: UPDM_YesNoIndicator
        iMismatch = None
        sMM = in_row[16]
        if sMM == "Yes":
            iMismatch = 1
        elif sMM == "No":
            iMismatch = 2

        dtStatus = in_row[17]

        iMainline = None
        sML = in_row[18]
        if sML == "Yes":
            iMainline = 1
        elif sML == "No":
            iMainline = 2

        sCPTown = None
        sCPCircuit = None
        sCPColor = None
        if self.__bAsBuilt:
            sCPTown = in_row[19]
            sCPCircuit = in_row[20]
            sCPColor = in_row[21]

        sComments = in_row[22]
        sServiceCard = in_row[23]

        iSleeved = None
        sSlv = in_row[24]
        if sSlv == "Yes":
            iSleeved = 1
        elif sSlv == "No":
            iSleeved = 2
        sGlobalID = in_row[25]

        iStub = None
        iSubType = in_row[26]
        if iSubType == 3:
            iStub = 1
        elif iSubType == 1:
            iStub = 2
        iEnd = 27

        dtRetired = None
        if not self.__bAsBuilt:
            dtRetired = in_row[iEnd]
        out_row = [self.__iAssetGroup, iAssetType, dtCreated, sCreationUser,
                   dtModified, sLastUser, dtInService, iLifeCycle,
                   dNominalDiameter, dMeasuredLength, iLengthSource, dWallThickness, 
                   dActualInternalDiameter, sWorkOrderID, sStoresStockNumber, sInstalledBy, 
                   iMismatch, dtStatus, iMainline, sCPTown,
                   sCPCircuit, sCPColor, sComments, sServiceCard, 
                   iSleeved, shape, sGlobalID, iStub]
        if not self.__bAsBuilt:
            out_row.extend([dtRetired])
        self.__cursor.insertRow(out_row)
        return

    def __MigrateCPBondWire(self):

        # Project to temporary FC
        sFCName = "CPBondWire"
        self.__iAssetGroup = 8 # "Bond Wire"
        print("   Migrating " + sFCName + "...")
        sSourceFC = self.__sSourceGDB + "/" + sFCName

        # Get projected shapes
        print("   Getting projected shapes...")
        ProjectedShapes = self.__helper.GetProjectedShapes(sFCName)

        # Create InsertCursor
        print("   Processing rows...")
        out_fields = ["ASSETGROUP", "ASSETTYPE", "CREATIONDATE", "CREATOR", 
                      "LASTUPDATE", "UPDATEDBY", "CP_TOWN_NAME", "CP_CIRCUIT_NUMBER", 
                      "CP_COLOR",  "SHAPE@", "GLOBALID" 
                      ]
        self.__cursor = arcpy.da.InsertCursor(self.__sTargetFC, out_fields)

        # Create SearchCursor
        in_fields = ["DATECREATED", "CREATIONUSER", "DATEMODIFIED", "LASTUSER", 
                     "CP_TOWN_NAME", "CP_CIRCUIT_NUMBER", "CP_COLOR", "GLOBALID"
                     ]
        with arcpy.da.SearchCursor(sSourceFC, in_fields) as cursorIn:
            for in_row in cursorIn:
                sGlobalID = in_row[7]
                if sGlobalID not in ProjectedShapes:
                    continue
                shape = ProjectedShapes[sGlobalID]
                self.__ProcessCPRow(in_row, shape)
                # Fewer than 1000 records
        del self.__cursor
        return

    def __ProcessCPRow(self, in_row, shape):
        iAssetType = 281 # "Bond Wire"
        dtCreated = in_row[0]
        sCreationUser = in_row[1]
        dtModified = in_row[2]
        sLastUser = in_row[3]
        sCPTown = in_row[4]
        sCPCircuit = in_row[5]
        sCPColor = in_row[6]
        sGlobalID = in_row[7]
        out_row = [self.__iAssetGroup, iAssetType, dtCreated, sCreationUser, 
                    dtModified, sLastUser, sCPTown, sCPCircuit, 
                    sCPColor, shape, sGlobalID]
        self.__cursor.insertRow(out_row)
        return

print("Ready.")
if __name__ == "__main__":
    ml = MigrateLine()
    ml.DoIt()

