################################################################################################
#  Create UPDM 2018 Core Assets
#   01CreateUPDMBaseAdd-On_2018.py
#
#  Created on: 2017-12-27
#  Revised on: 2018-2-7
#
#  Description: Purpose of this python script is to append the asset portion of the Utility and
#  Pipeline Data Model (UPDM) which is not included in the Pipeline Industry Asset package.
#  This base add-on portion of UPDM 2018 Edition is used to provide a more comprehensive list of attributes
#  and domains to describe the assets.
#
#  It is required that this module of UPDM be run prior to either the Integrity or Inspections modules.
#
################################################################################################
import arcpy
from timeit import default_timer as timer
import logging
import sys
import os

file_logger = logging.FileHandler(filename=os.path.join(os.path.dirname(__file__), 'log_UPDM2018BaseAddOn.txt'), mode='w')
stream_logger = logging.StreamHandler(stream=sys.stdout)

logging.basicConfig(level=logging.DEBUG,
                    format="%(asctime)s\t%(message)s",
                    handlers=[file_logger, stream_logger])
logging.debug('Initialize')
try:
    #########################
    # input Parameters
    #########################
    logging.debug("Setting input parameters...")
    # Enterprise Geodatabase Connection File
    targetWorkspace = "C:/apps/Data/UPDM_Migration/AssetPackage_AZGas.gdb"  # r"C:\Users\tom\Documents\ArcGIS\Projects\UPDM2018\UPDM2018.sde"
    # SQL Server Database Name
    #SQLDatabaseName = "UPDM2018"
    # SQL Server Database Gas Domain Data Owner
    #SQLDatabaseOwner = "GIS"
    # SQL Server Data Prefix
    SQLPrefix = "/" # "\\" + SQLDatabaseName + "." + SQLDatabaseOwner + "."
    # Utility Network Name
    utilityNetworkName = "UtilityNetwork" # "PipeSystem"
    # Utility Network Gas Domain Name
    DomainName = "Pipeline"
    #Assign Service Territory polygon featureclass
    #serviceTerritoryFeatureClass = targetWorkspace + SQLPrefix + "ServiceArea"
    # Aggregated featureDataset Name
    featureDataset = SQLPrefix + utilityNetworkName
    # Aggregated featureDataset Path
    featureDatasetPath = targetWorkspace + featureDataset
    # Aggregated Utility Network Path
    utilityNetworkPath = targetWorkspace + featureDataset + SQLPrefix + utilityNetworkName

    ################################
    # Auto-defined script variables (Do not modify)
    #################################
    UPDMAssembly = targetWorkspace + featureDataset + SQLPrefix + DomainName + "Assembly"
    UPDMDevice = targetWorkspace + featureDataset + SQLPrefix + DomainName + "Device"
    UPDMJunction = targetWorkspace + featureDataset + SQLPrefix + DomainName + "Junction"
    UPDMLine = targetWorkspace + featureDataset + SQLPrefix + DomainName + "Line"
    UPDMSubNetLine = targetWorkspace + featureDataset + SQLPrefix + DomainName + "SubnetLine"
    
    structureJunction = targetWorkspace + featureDataset + SQLPrefix + "StructureJunction"
    structureEdge = targetWorkspace + featureDataset + SQLPrefix + "StructureLine"
    structureBoundary = targetWorkspace + featureDataset + SQLPrefix + "StructureBoundary"

    UPDMServiceSummary = targetWorkspace + featureDataset + SQLPrefix + DomainName + "ServiceSummary"
    UPDMServiceSummaryREL = targetWorkspace + featureDataset + SQLPrefix + DomainName + "ServiceSum_Service"


    #Auto-define new variables, because some tools don't like paths
    roleUPDMAssembly = DomainName + "Assembly"
    roleUPDMLine = DomainName + "Line"
    roleUPDMDevice = DomainName + "Device"
    roleUPDMJunction = DomainName + "Junction"
    serviceTerritoryName = "ServiceArea"
    roleStructureJunction = "StructureJunction"
    roleStructureEdge = "StructureLine"
    roleStructureBoundary = "StructureBoundary"


    # Environment valiables
    arcpy.env.workspace = targetWorkspace
    arcpy.env.preserveGlobalIds = True

    startTime = timer()    

    ########################################
    #Add Domains into the Geodatabase
    ########################################
    logging.debug("Creating Domains in the Geodatabase")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_MeasuredLength", "Valid Range of values for gas pipe lengths", "DOUBLE", "RANGE", "GEOMETRY_RATIO", "DEFAULT")
    arcpy.SetValueForRangeDomain_management(targetWorkspace, "UPDM_MeasuredLength", 0, 1000000000)

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_PressureRange", "Valid Range of values for gas pipe pressure", "DOUBLE", "RANGE", "GEOMETRY_RATIO", "DEFAULT")
    arcpy.SetValueForRangeDomain_management(targetWorkspace, "UPDM_PressureRange", 0, 10000)

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_SymbolPosition", "Valid Range of degrees for rotation", "SHORT", "RANGE", "GEOMETRY_RATIO", "DEFAULT")
    arcpy.SetValueForRangeDomain_management(targetWorkspace, "UPDM_SymbolPosition", -359, 359)

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_Accessible", "Boolean value for designating whether a device is accessable.", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_Accessible", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_Accessible", 1, "Yes")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_Accessible", 2, "No")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "List of ASTM F-2897 Material for Componenta Specification", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "A", "PE2406")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "B", "PE2708")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "C", "PE3408")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "D", "PE3608")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "E", "PE3708")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "F", "PE3710")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "G", "PE4608")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "H", "PE4710")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "J", "PVC")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "K", "PA11")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "L", "PA12")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "M", "Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "N", "Stainless Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "O", "Cast Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "P", "Copper")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "Q", "Brass")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "R", "Malleable Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "S", "Ductile Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "T", "Reinforced Epoxy Resin")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "U", "Nylon")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "V", "Glass Filler Nylon")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "0", "Grade A")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "1", "Grade B")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "2", "Grade C")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "3", "Grade X42")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "4", "Grade X46")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "5", "Grade X52")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "6", "Grade X56")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "7", "Grade X60")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "8", "Grade X65")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "9", "Grade X70")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMComponentMaterial", "X", "Other")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "List of ASTM F-2897 types of Coupling Manufacture Models", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "20", "Other Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "21", "Socket Fusion Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "22", "Socket Fusion with EFV Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "23", "Electrofusion Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "24", "Electrofusion with EFV Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "25", "Mechanical Compression or nut follower Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "26", "Mechanical Compression or nut follower Coupling with EFV")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "27", "Mechanical stab Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "28", "Mechanical stab with EFV Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "29", "Mechanical interference fit Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "2A", "Mechanical interference fit with EFV Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "2B", "Welded Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "2C", "Threaded Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "2D", "Flanged Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "30", "Other Adapter Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "31", "Compression by Male pipe thread Adapter Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "32", "Compression by femail pipe thread Adapter Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "33", "Compression by butt fusion Adapter Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "34", "Compression by butt welded Adapter Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "35", "Compression by solvent welded Adapter Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "36", "Stab by male pipe thread Adapter Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "37", "Stab by female pipe thread Adapter Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "38", "Stab by solvent welded Adapter Coupling")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMCouplingModel", "39", "Compression by stab Adapter Coupling")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMEFVModel", "List of ASTM F-2897 types of EFV Manufacture Models", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMEFVModel", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMEFVModel", "EF", "Excess Flow Valve")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMEFVModel", "V0", "Other Valve") 

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "List of ASTM F-2897 types of Elbow Manufacture Models", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "50", "Other Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "51", "Butt fusion 90 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "52", "Socket fusion 90 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "53", "Electrofusion 90 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "54", "Mechanical compression or nut follower Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "55", "Mechanical stab 90 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "56", "Mechanical interference fit 90 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "57", "Welded 90 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "58", "Threaded 90 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "59", "Fabricated 90 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "5A", "Butt fusion 45 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "5B", "Socket fusion 45 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "5C", "Electrofusion 45 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "5D", "Mechanical compression or nut follower 45 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "5E", "Mechanical stab 45 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "5F", "Mechanical interference fit 45 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "5G", "Welded 45 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "5H", "Threaded 45 Elbow")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMElbowModel", "5J", "Fabricated 45 Elbow")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMEndCapModel", "List of ASTM F-2897 types of End Cap Manufacture Models", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMEndCapModel", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMEndCapModel", "40", "Other End Cap")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMEndCapModel", "41", "Butt fusion End Cap")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMEndCapModel", "42", "Socket fusion End Cap")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMEndCapModel", "43", "Electrofusion End Cap")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMEndCapModel", "44", "Mechanical compression or nut follower End Cap")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMEndCapModel", "45", "Mechanical stab End Cap")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMEndCapModel", "46", "Mechanical interference fit End Cap")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMEndCapModel", "47", "Welded End Cap")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMEndCapModel", "48", "Threaded End Cap")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMEndCapModel", "49", "Fabricated End Cap")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMMeterModel", "List of ASTM F-2897 Meter Set Manufacture Model", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMMeterModel", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMMeterModel", "M0", "Other Meter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMMeterModel", "M1", "Diaphragm Meter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMMeterModel", "M2", "Rotary Meter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMMeterModel", "M3", "Meter Set Assembly")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMMeterModel", "M4", "Meter bar")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMMeterModel", "M5", "Meter Swival")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMMeterModel", "M6", "Meter Nut")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMMeterModel", "M7", "Ultrasonic Meter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMMeterModel", "M8", "Turbine Meter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMMeterModel", "M9", "Remote shut off Meter")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "List of ASTM F-2897 Manufacturers of Pipe, fittings and devices", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "0", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "35", "Industrial Pipe Fittings")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "73", "Plasson")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "78", "Torre Gas, S.L.")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "AD", "A-D Technologies")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "AG", "Agru America Inc")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "AY", "A.Y. McDonald Mfg Co.")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "CE", "Centennial Plastics, Inc")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "CH", "Charter Plastics Inc")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "CI", "Continental Industries")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "CP", "Georg Fischer Central Plastics LLC")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "DR", "Dresser Pipeline Solutions")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "EL", "Elster Perfection")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "EN", "Endot Industries Inc")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "FR", "IPEX Inc")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "GB", "Gas Breaker Inc")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "HF", "Valencia Pipe Company")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "IN", "Innoge")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "IU", "IPEX USA LLC")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "JM", "J-M Manufacturing Co.")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "KT", "Kerotest Manufacturing Corp")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "MT", "M.T. Deason Co. inc")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "NG", "Nupi Americas")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "PE", "Performance Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "PL", "Pipeline Plastics, LLC")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "PP", "DuraLine")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "PT", "Polytubes LP")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "PV", "PolyValve LLC")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "RP", "Rahn Plastics")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "RW", "R.W. Lyall")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "SB", "Smith-Blair Inc")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "TD", "T.D. Williamson")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "TP", "Teel Pipe LLC")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "UI", "Uponor Infra Limited")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "UP", "United Poly Systems")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "VP", "Versaprofiles Inc")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMManufacturer", "WL", "WL Plastics")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "List of ASTM F-2897 Material for Pipe Grade", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "A", "PE2406")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "B", "PE2708")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "C", "PE3408")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "D", "PE3608")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "E", "PE3708")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "F", "PE3710")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "G", "PE4608")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "H", "PE4710")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "J", "PVC")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "K", "PA11")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "L", "PA12")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "O", "Cast Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "R", "Malleable Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "S", "Ductile Iron")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "0", "Grade A")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "1", "Grade B")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "2", "Grade C")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "3", "Grade X42")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "4", "Grade X46")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "5", "Grade X52")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "6", "Grade X56")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "7", "Grade X60")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "8", "Grade X65")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "9", "Grade X70")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "M", "Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "N", "Stainless Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "T", "Reinforced Expoy Resin")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "U", "Nylon")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "V", "Glass Filler Nylon")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeMaterial", "X", "Other")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "List of ASTM F-2897 valid Pipe Specification", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "0", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "10", "Other Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "11", "Straight Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "12", "Coiled Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "13", "Casing Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "1A", "Seamless Line Pipe, API 5L, PSL1, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "1B", "Seamless Line Pipe, API 5L, PSL1, Dual Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "1C", "Seamless Line Pipe, API 5L, PSL2, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "1D", "Seamless Line Pipe, API 5L, PSL2, Dual Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "1E", "Electric Resistance Weld, API 5L, PSL1, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "1F", "Electric Resistance Weld, API 5L, PSL1, Dual Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "1G", "Electric Resistance Weld, API 5L, PSL2, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "1H", "Electric Resistance Weld, API 5L, PSL2, Dual Coat Pipe")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "1J", "Seamless and Welded ASTM A53/53M Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMPipeType", "1K", "Seamless Carbon Steel ASTM A106 Pipe")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "List of ASTM F-2897 Material for Riser Pipe Grade", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "O", "Cast Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "R", "Malleable Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "S", "Ductile Iron")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "0", "Grade A")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "1", "Grade B")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "2", "Grade C")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "3", "Grade X42")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "4", "Grade X46")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "5", "Grade X52")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "6", "Grade X56")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "7", "Grade X60")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "8", "Grade X65")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "9", "Grade X70")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "M", "Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "N", "Stainless Steel")   
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeMaterial", "X", "Other")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeType", "List of ASTM F-2897 valid Riser Pipe Specification", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeType", "0", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeType", "10", "Other Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeType", "13", "Casing Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeType", "1A", "Seamless Line Pipe, API 5L, PSL1, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeType", "1B", "Seamless Line Pipe, API 5L, PSL1, Dual Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeType", "1C", "Seamless Line Pipe, API 5L, PSL2, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeType", "1D", "Seamless Line Pipe, API 5L, PSL2, Dual Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeType", "1E", "Electric Resistance Weld, API 5L, PSL1, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeType", "1F", "Electric Resistance Weld, API 5L, PSL1, Dual Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeType", "1G", "Electric Resistance Weld, API 5L, PSL2, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeType", "1H", "Electric Resistance Weld, API 5L, PSL2, Dual Coat Pipe")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeType", "1J", "Seamless and Welded ASTM A53/53M Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRiserPipeType", "1K", "Seamless Carbon Steel ASTM A106 Pipe")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "List of ASTM F-2897 Material for Station Pipe Grade", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "O", "Cast Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "R", "Malleable Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "S", "Ductile Iron")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "0", "Grade A")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "1", "Grade B")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "2", "Grade C")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "3", "Grade X42")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "4", "Grade X46")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "5", "Grade X52")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "6", "Grade X56")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "7", "Grade X60")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "8", "Grade X65")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "9", "Grade X70")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "M", "Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "N", "Stainless Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "T", "Reinforced Expoy Resin")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "U", "Nylon")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "V", "Glass Filler Nylon")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeMaterial", "X", "Other")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMStationPipeType", "List of ASTM F-2897 valid Station Pipe Specification", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeType", "0", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeType", "10", "Other Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeType", "13", "Casing Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeType", "1A", "Seamless Line Pipe, API 5L, PSL1, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeType", "1B", "Seamless Line Pipe, API 5L, PSL1, Dual Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeType", "1C", "Seamless Line Pipe, API 5L, PSL2, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeType", "1D", "Seamless Line Pipe, API 5L, PSL2, Dual Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeType", "1E", "Electric Resistance Weld, API 5L, PSL1, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeType", "1F", "Electric Resistance Weld, API 5L, PSL1, Dual Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeType", "1G", "Electric Resistance Weld, API 5L, PSL2, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeType", "1H", "Electric Resistance Weld, API 5L, PSL2, Dual Coat Pipe")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeType", "1J", "Seamless and Welded ASTM A53/53M Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStationPipeType", "1K", "Seamless Carbon Steel ASTM A106 Pipe")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMTransPipeType", "List of ASTM F-2897 valid Transmission Pipe Specification", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeType", "0", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeType", "10", "Other Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeType", "13", "Casing Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeType", "1A", "Seamless Line Pipe, API 5L, PSL1, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeType", "1B", "Seamless Line Pipe, API 5L, PSL1, Dual Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeType", "1C", "Seamless Line Pipe, API 5L, PSL2, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeType", "1D", "Seamless Line Pipe, API 5L, PSL2, Dual Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeType", "1E", "Electric Resistance Weld, API 5L, PSL1, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeType", "1F", "Electric Resistance Weld, API 5L, PSL1, Dual Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeType", "1G", "Electric Resistance Weld, API 5L, PSL2, Single Coat Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeType", "1H", "Electric Resistance Weld, API 5L, PSL2, Dual Coat Pipe")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeType", "1J", "Seamless and Welded ASTM A53/53M Pipe")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeType", "1K", "Seamless Carbon Steel ASTM A106 Pipe")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMReducerModel", "List of ASTM F-2897 Reducer Manufacture Model Types", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMReducerModel", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMReducerModel", "70", "Other Reducer")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMReducerModel", "71", "Butt fusion Reducer")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMReducerModel", "72", "Socket fusion Reducer")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMReducerModel", "73", "Electrofusion Reducer")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMReducerModel", "74", "Mechanical compression or nut follower Reducer")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMReducerModel", "75", "Mechanical stab Reducer")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMReducerModel", "76", "Mechanical Interference fit Reducer")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMReducerModel", "77", "Welded Reducer")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMReducerModel", "78", "Threaded Reducer")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMReducerModel", "79", "Fabricated Reducer")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMRegulatorModel", "List of ASTM F-2897 Regulator Manufacture Model Types", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRegulatorModel", "0", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRegulatorModel", "RX", "Other Regulator")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRegulatorModel", "RP", "Pilot Regulator")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRegulatorModel", "RS", "Service Regulator")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMRegulatorModel", "RR", "Relief Regulator")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "List of ASTM F-2897 Saddle Manufacture Model Types", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "0", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "B0", "Other branch saddle")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "B1", "Electrofusion branch saddle")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "B2", "Saddle heat fusion branch saddle")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "D0", "Other mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "D1", "Welded by welded mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "D2", "Welded by butt fusion mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "D3", "Welded by thread mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "D4", "Welded by compression or nut follower saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "D5", "Welded by mechanical interference fit mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "DD", "Welded by stab mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "D6", "Thread by Welded mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "D7", "Thread by compression or nut follower mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "DE", "Thread by mechanical interference fit mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "DF", "Thread by stab mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "DG", "Thread by thread mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "DH", "Thread by butt fusion mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "D8", "Mechanical saddle by welded mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "D9", "Mechanical saddle by butt fusion mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "DA", "Mechanical saddle by thread mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "DB", "Mechanical saddle by compression or nut follower mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "DC", "Mechanical saddle by mechanical interference fit mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "DJ", "Mechanical saddle by stab mechanical saddle service tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "E0", "Other service saddle")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "E1", "Single strap service saddle")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMSaddleModel", "E2", "Double strap service saddle")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMStrainerModel", "List of ASTM F-2897 Strainer Manufacture Model Types", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStrainerModel", "0", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStrainerModel", "F0", "Other Filter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStrainerModel", "F1", "Pilot Filter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStrainerModel", "F2", "Service and mains Filter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMStrainerModel", "F3", "Strainer Filter")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "List of ASTM F-2897 Tee Manufacture Model Types", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "60", "Other 3-Way Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "61", "Butt fusion 3-Way Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "62", "Socket fusion 3-Way Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "63", "Electrofusion 3-Way Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "64", "Mechanical compression or nut follower 3-Way Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "65", "Mechanical stab 3-Way Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "66", "Mechanical interference fit 3-Way Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "67", "Welded 3-Way Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "68", "Threaded 3-Way Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "69", "Fabricated 3-Way Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "80", "Other Tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "81", "Saddle heat fusion by butt fusion outlet tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "82", "Saddle heat fusion by butt fusion outlet with EFV tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "83", "Saddle heat fusion by socket outlet tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "84", "Saddle heat fusion by socket outlet with EFV tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "85", "Saddle heat fusion by mechanical compression outlet tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "86", "Saddle heat fusion by mechanical compression outlet with EFV tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "87", "Saddle heat fusion by stab outlet tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "88", "Saddle heat fusion by stab outlet with EFV tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "89", "Electrofusion by butt fusion outlet tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8A", "Electrofusion by butt fusion outlet with EFV tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8B", "Electrofusion by socket fusion tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8C", "Electrofusion by socket fusion with EFV tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8D", "Electrofusion by mechanical compression outlet tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8E", "Electrofusion by mechanical compression outlet with EFV tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8F", "Electrofusion by stab outlet tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8G", "Electrofusion by stab outlet with EFV tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8H", "Mechanical by butt fusion tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8J", "Mechanical by butt fusion with EFV tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8K", "Mechanical by Socket outlet tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8L", "Mechanical by socket outlet with EFV tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8M", "Mechanical by mechanical compression outlet tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8N", "Mechanical by mechanical compression outlet with EFV tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8P", "Mechanical by stab outlet tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8Q", "Mechanical by stab outlet with EFV tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8R", "Mechanical by mechanical interference fit tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "8S", "Mechanical by mechanical interference fit with EFV tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "90", "Other high volume tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "91", "Electrofusion by butt fusion high volume tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "92", "Saddle heat fusion by butt fusion high volume tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "93", "Mechanical by compression outlet high volume tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "94", "Electrofusion by socket outlet high volume tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "95", "Saddle heat fusion by socket outlet high volume tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "96", "Mechanical by stab outlet high volume tapping Tee")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTeeModel", "97", "Mechanical by mechanical interference fit high volume tapping Tee")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMTransitionModel", "List of ASTM F-2897 Transition Manufacture Model Types", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransitionModel", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransitionModel", "T0", "Other Transition")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransitionModel", "T1", "Welded end Transition")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransitionModel", "T2", "Thread end Transition")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransitionModel", "T3", "Flanged end Transition")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMFValve", "List of ASTM F2897 manufacturer specified types of valves", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMFValve", "0", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMFValve", "V0", "Other Valve")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMFValve", "V1", "Ball Valve")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMFValve", "V2", "Butterfly Valve")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMFValve", "V3", "Check Valve")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMFValve", "V4", "Relief Valve")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMFValve", "V5", "Gas Valve")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMFValve", "V6", "Needle Valve")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMFValve", "V7", "Plug Valve")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "List of ASTM F-2897 Material for Transmission Pipe Grade", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "O", "Cast Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "R", "Malleable Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "S", "Ductile Iron")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "0", "Grade A")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "1", "Grade B")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "2", "Grade C")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "3", "Grade X42")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "4", "Grade X46")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "5", "Grade X52")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "6", "Grade X56")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "7", "Grade X60")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "8", "Grade X65")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "9", "Grade X70")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "M", "Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "N", "Stainless Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "T", "Reinforced Expoy Resin")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "U", "Nylon")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "V", "Glass Filler Nylon")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ASTMTransPipeMaterial", "X", "Other")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_CommodityType", "List of types of commodity which can be transported by the pipe system", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CommodityType", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CommodityType", 1, "Natural Gas")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CommodityType", 2, "Synthetic Gas")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CommodityType", 3, "Hydrogen Gas")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CommodityType", 4, "Propane Gas")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CommodityType", 5, "Landfill Gas")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CommodityType", 6, "Petroleum")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CommodityType", 7, "Gasoline")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CommodityType", 8, "Diesel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CommodityType", 9, "Water")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CommodityType", 10, "Wastewater")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CommodityType", 11, "Steam")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_CoolingSysManufacturer", "List of valid manufacturers of oil and gas Cooling Systems", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CoolingSysManufacturer", "UN", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_CoolingSysModel", "List of valid models of oil and gas Cooling Systems", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CoolingSysModel", "UN", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_CPAnodeMaterialType", "List of valid types of Anode material", "Text", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeMaterialType", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeMaterialType", "9A", "Carbon-Graphite")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeMaterialType", "9B", "Cast Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeMaterialType", "9C", "Galvanic")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeMaterialType", "9D", "Graphite")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeMaterialType", "9E", "Magnesium")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeMaterialType", "9F", "Magnesium - High Grade B")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeMaterialType", "9G", "Magnesium - High Potential")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeMaterialType", "9H", "Mixed Metal Oxide")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeMaterialType", "9I", "Scrap Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeMaterialType", "9J", "Zinc")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_CPAnodeWeight", "List of valid types of sizes by weight for Anodes", "DOUBLE", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeWeight", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeWeight", 0.75, "3/4 lb")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeWeight", 1, "1 lb")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeWeight", 1.5, "1 1/2lb")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeWeight", 3, "3 lb")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeWeight", 5, "5 lb")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeWeight", 9, "9 lb")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeWeight", 17, "17 lb")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeWeight", 20, "20 lb")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeWeight", 32, "32 lb")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPAnodeWeight", 48, "48 lb")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_CPASTMF2897Anodes", "List of valid types Anode types as defined by ASTM standard F2897", "Text", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPASTMF2897Anodes", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPASTMF2897Anodes", "A0", "Other Anode")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPASTMF2897Anodes", "A1", "Cast Iron Anode")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPASTMF2897Anodes", "A2", "Graphite Anode")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPASTMF2897Anodes", "A3", "Magnesium Anode")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPASTMF2897Anodes", "A4", "Zinc Anode")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_CPBondJunctionType", "List of valid types of Bond Junctions", "Text", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPBondJunctionType", "Unknown", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPBondJunctionType", "Above Ground With Foreign Line", "Above Ground With Foreign Line")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPBondJunctionType", "Above Ground Bond", "Above Ground Bond")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPBondJunctionType", "Above Ground With Anode", "Above Ground With Anode")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPBondJunctionType", "On Post", "Above Ground")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPBondJunctionType", "Curb Box Bond", "Curb Box Bond")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPBondJunctionType", "Curb Box With Foreign Line", "Curb Box With Foreign Line")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPBondJunctionType", "Curb Box", "Curb Box")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPBondJunctionType", "Polarization", "Polarization")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_CPColorCode", "List of valid color used for bond wires", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPColorCode", "Unknown", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPColorCode", "Aqua", "Aqua")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPColorCode", "Blue", "Blue")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPColorCode", "Forest", "Forest")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPColorCode", "Green", "Green")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPColorCode", "Lavender", "Lavender")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPColorCode", "Orange", "Orange")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPColorCode", "Pink", "Pink")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPColorCode", "Purple", "Purple")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPColorCode", "Red", "Red")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPColorCode", "Yellow", "Yellow")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_CPRectifierCableGuage", "List of valid sizes used for rectifier cables", "SHORT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPRectifierCableGuage", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPRectifierCableGuage", 4, "4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPRectifierCableGuage", 8, "8")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_CPRectifierManufacture", "List of valid manufacturers of Rectifiers", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPRectifierManufacture", "U", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPRectifierManufacture", "GD", "Goodall")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPRectifierManufacture", "KT", "Kerotest")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPRectifierManufacture", "RW", "Rockwell")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPRectifierManufacture", "RL", "R.W. Lyall")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_CPTestPointNumWires", "List of valid number of wires which can connect to a test point", "SHORT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPTestPointNumWires", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPTestPointNumWires", 1, "1")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CPTestPointNumWires", 2, "2")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", "List of valid diameter sizes for customer meters", "DOUBLE", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 0.25, '1/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 0.38, '3/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 0.50, '1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 0.63, '5/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 0.75, '3/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 1.0, '1"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 1.125, '1 1/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 1.25, '1 1/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 1.375, '1 3/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 1.5, '1 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 1.875, '1 7/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 2, '2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 2.5, '2 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 3, '3"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 4, '4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 5, '5"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 5.5, '5 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 6, '6"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 8, '8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 10, '10"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 12, '12"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 14, '14"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 16, '16"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 18, '18"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 20, '20"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustomerMeterDiameter", 24, '24"')
                     
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", "List of valid diameter sizes for Custody Transfer meters", "DOUBLE", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 4, '4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 5, '5"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 5.5, '5 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 6, '6"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 8, '8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 10, '10"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 12, '12"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 14, '14"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 16, '16"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 18, '18"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 20, '20"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 24, '24"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 26, '26"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 30, '30"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 36, '36"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 40, '40"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_CustodyTransMeterDiameter", 48, '48"')

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_DEVICESTATUS", "List of valid values for the Network Attribute Device Status", "SHORT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_DEVICESTATUS", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_DEVICESTATUS", 1, "Open")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_DEVICESTATUS", 2, "Closed")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_DehydrateManufacturer", "List of valid manufacturers of oil and gas dehydration equipment", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_DehydrateManufacturer", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_DehydrateModel", "List of valid models of oil and gas dehydration equipment", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_DehydrateModel", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_DripManufacturer", "List of valid manufacturers of gas drips", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_DripManufacturer", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_DripModel", "List of valid models of gas drips", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_DripModel", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_EFVDiameter", "List of valid diameter sizes for excess flow valves", "DOUBLE", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 0.25, '1/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 0.38, '3/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 0.50, '1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 0.63, '5/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 0.75, '3/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 1.0, '1"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 1.125, '1 1/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 1.25, '1 1/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 1.375, '1 3/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 1.5, '1 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 1.875, '1 7/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 2, '2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 3, '3"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 4, '4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 5, '5"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_EFVDiameter", 6, '6"')
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_FittingDiameter", "List of valid diameter sizes for Gas Junction Single Diameter Fittings", "DOUBLE", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 0.25, '1/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 0.38, '3/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 0.50, '1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 0.63, '5/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 0.75, '3/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 1.0, '1"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 1.125, '1 1/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 1.25, '1 1/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 1.375, '1 3/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 1.5, '1 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 1.875, '1 7/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 2, '2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 2.5, '2 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 3, '3"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 4, '4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 5, '5"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 5.5, '5 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 6, '6"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 8, '8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 10, '10"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 12, '12"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 14, '14"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 16, '16"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 18, '18"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 20, '20"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 24, '24"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 26, '26"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 30, '30"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 36, '36"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 40, '40"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_FittingDiameter", 48, '48"')    

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_INSULATORDEVICE", "List of valid values for the Network Attribute Insulating Device", "SHORT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_INSULATORDEVICE", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_INSULATORDEVICE", 1, "No")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_INSULATORDEVICE", 2, "Yes")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LengthSource", "List of valid types of sources for measured length", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LengthSource", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LengthSource", 1, "Field Measurement")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LengthSource", 2, "GPS Sub- 30 Meter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LengthSource", 3, "GPS Sub- 10 Meter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LengthSource", 4, "GPS Sub-Meter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LengthSource", 5, "GPS Sub-Foot")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LengthSource", 6, "GPS centimeter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LengthSource", 7, "Mapping System")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LengthSource", 8, "User-Derived")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LengthSource", 9, "Begin-End Measures")   

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LifeCycleStatus", "List of the valid status's for an asset", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LifeCycleStatus", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LifeCycleStatus", 1, "Proposed")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LifeCycleStatus", 2, "Proposed Retired")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LifeCycleStatus", 3, "In Service")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LifeCycleStatus", 4, "Out of Service")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LifeCycleStatus", 5, "Retired")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LifeCycleStatus", 6, "Removed")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LifeCycleStatus", 7, "Abandoned")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LifeCycleStatus", 8, "Under Construction")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LineCoatingType", "List of valid types of coating applied to gas pipe", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Asbestos Felt", "Asbestos Felt")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Asphalt", "Asphalt")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "3M Tape", "3M Tape")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Blue Gard", "Blue Gard")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Coal Tar", "Coal Tar")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Coal Tar Enamel", "Coal Tar Enamel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Cold Tape", "Cold Tape")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Composite", "Composite")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Concrete", "Concrete")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Enamel", "Enamel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Epoxy", "Epoxy")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Extruded Asphalt", "Extruded Asphalt")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Extruded Polyethylene", "Extruded Polyethylene")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Felt Wrap", "Felt Wrap")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Flakeline", "Flakeline")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Flex Clad", "Flex Clad")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Fusion Bonded Epoxy", "Fusion Bonded Epoxy")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Galvanized", "Galvanized")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Glass Filled Polyester", "Glass Filled Polyester")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Grease", "Grease")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Hot Tape", "Hot Tape")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Hysol", "Hysol")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Mill Wrap", "Mill Wrap")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Neoprene", "Neoprene")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Paint", "Paint")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Paper", "Paper")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Primer", "Primer")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Rock Jacket", "Rock Jacket")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Scotch Wrap", "Scotch Wrap")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Tar", "Tar")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Wax", "Wax")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Yellow Jacket", "Yellow Jacket")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineCoatingType", "Other", "Other")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LineDesignClass", "List of U.S. DOT classifications for design criteria of gas pipe", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDesignClass", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDesignClass", 1, "Class 1") 
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDesignClass", 2, "Class 2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDesignClass", 3, "Class 3")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDesignClass", 4, "Class 4")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", "List of Valid Nominal Diameter Size for Customer Pipe", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 0.25, "1/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 0.38, "3/8")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 0.50, "1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 0.75, "3/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 1.0, "1")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 1.25, "1 1/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 1.50, "1 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 2.0, "2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 2.5, "2 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 3.0, "3")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 3.5, "3 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 4.0, "4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 4.5, "4 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 5.0, "5")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterCustomPipe", 6.0, "6")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", "List of Valid Nominal Diameter Size for Distribution Pipe", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 0.25, "1/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 0.38, "3/8")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 0.50, "1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 0.75, "3/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 1.0, "1")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 1.25, "1 1/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 1.50, "1 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 2.0, "2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 2.5, "2 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 3.0, "3")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 3.5, "3 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 4.0, "4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 4.5, "4 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 5.0, "5")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 6.0, "6")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 7.0, "7")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 8.0, "8")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 9.0, "9")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 10.0, "10")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 12.0, "12")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 14.0, "14")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterDistribution", 16.0, "16")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", "List of Valid Nominal Diameter Size for Gathering Pipe", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 0.25, "1/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 0.38, "3/8")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 0.50, "1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 0.75, "3/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 1.0, "1")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 1.25, "1 1/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 1.50, "1 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 2.0, "2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 2.5, "2 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 3.0, "3")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 3.5, "3 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 4.0, "4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 4.5, "4 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 5.0, "5")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 6.0, "6")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 7.0, "7")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 8.0, "8")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 9.0, "9")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 10.0, "10")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 12.0, "12")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 14.0, "14")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterGathering", 16.0, "16")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", "List of Valid Nominal Diameter Size for Riser Pipe", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 0.25, "1/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 0.38, "3/8")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 0.50, "1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 0.75, "3/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 1.0, "1")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 1.25, "1 1/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 1.50, "1 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 2.0, "2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 2.5, "2 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 3.0, "3")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 3.5, "3 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 4.0, "4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 4.5, "4 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 5.0, "5")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterRiserPipe", 6.0, "6")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LineDiameterServices", "List of Valid Nominal Diameter Size for Service Lines", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 0.25, "1/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 0.38, "3/8")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 0.50, "1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 0.75, "3/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 1.0, "1")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 1.25, "1 1/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 1.50, "1 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 2.0, "2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 2.5, "2 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 3.0, "3")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 3.5, "3 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 4.0, "4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 4.5, "4 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 5.0, "5")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterServices", 6.0, "6")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", "List of Valid Nominal Diameter Size for Transmission Pipe", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 12, "12")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 14, "14")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 16, "16")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 18, "18")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 20, "20")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 22, "22")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 24, "24")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 26, "26")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 28, "28")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 30, "30")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 32, "32")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 34, "34")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 36, "36")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 40, "40")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 42, "42")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 44, "44")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 48, "48")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 52, "52")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 56, "56")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 60, "60")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 64, "64")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 68, "68")  
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 72, "72")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 76, "76")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterTransmission", 80, "80") 

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", "List of Valid Nominal Diameter Size for Unknown Pipe", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 0.25, "1/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 0.38, "3/8")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 0.50, "1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 0.75, "3/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 1.0, "1")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 1.25, "1 1/4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 1.50, "1 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 2.0, "2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 2.5, "2 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 3.0, "3")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 3.5, "3 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 4.0, "4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 4.5, "4 1/2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 5.0, "5")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 6.0, "6")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 7.0, "7")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 8.0, "8")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 9.0, "9")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 10.0, "10")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 12.0, "12")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 14.0, "14")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 16.0, "16")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 18.0, "18")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 20.0, "20")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 22.0, "22")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 24.0, "24")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 26.0, "26")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 28.0, "28")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 30.0, "30")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 32.0, "32")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 34.0, "34")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 36.0, "36")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 40.0, "40")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 42.0, "42")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 44.0, "44")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 48.0, "48")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 52.0, "52")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 56.0, "56")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 60.0, "60")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 64.0, "64")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 68.0, "68")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 72.0, "72")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 76.0, "76")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 80.0, "80")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterUnknown", 88.0, "88")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", "List of Valid Nominal Diameter Size for Station Pipe", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 4.0, "4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 6.0, "6")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 7.0, "7")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 8.0, "8")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 9.0, "9")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 10.0, "10")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 12.0, "12")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 14.0, "14")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 16.0, "16")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 18.0, "18")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 20.0, "20")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 22.0, "22")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 24.0, "24")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 26.0, "26")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 28.0, "28")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 30.0, "30")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 32.0, "32")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 34.0, "34")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 36.0, "36")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 40.0, "40")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 42.0, "42")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 44.0, "44")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 48.0, "48")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 52.0, "52")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 56.0, "56")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 60.0, "60")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 64.0, "64")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 68.0, "68")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 72.0, "72")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 76.0, "76")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 80.0, "80")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineDiameterStationPipe", 88.0, "88")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LineHeaterManufacturer", "List of valid manufacturers of oil and gas Line Heaters", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineHeaterManufacturer", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LineHeaterModel", "List of valid models of oil and gas storage Line Heaters", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineHeaterModel", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LineInstallMethod", "List of valid types of methods for installing gas lines", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineInstallMethod", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineInstallMethod", 1, "Trench")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineInstallMethod", 2, "Joint Trench")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineInstallMethod", 3, "Insertion")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineInstallMethod", 4, "Direct Bore")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineInstallMethod", 5, "Split and Pull")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LineInstallMethod", 6, "Assemble")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LinePipeTestType", "List of valid types of substances for pressure testing gas lines", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LinePipeTestType", "Unknown", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LinePipeTestType", "Air", "Air")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LinePipeTestType", "Gas", "Gas")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LinePipeTestType", "Water", "Water")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_LongSeamType", "List of valid types of longitudinal seams", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LongSeamType", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LongSeamType", 1, "Seamless")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LongSeamType", 2, "Electric Weld")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LongSeamType", 3, "Lap Weld")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LongSeamType", 4, "Butt Weld")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LongSeamType", 5, "Single Submerged Arc Weld")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LongSeamType", 6, "Single Submerged Arc Weld Spiraled")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LongSeamType", 7, "Double Submerged Arc Weld")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LongSeamType", 8, "Double Submerged Arc Weld Spiraled")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_LongSeamType", 9, "Not Applicable")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_MeterType", "List of valid types of meters", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_MeterType", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_MeterType", 21, "Annubar")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_MeterType", 22, "Diaphragm")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_MeterType", 23, "Oriface")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_MeterType", 24, "Rotary")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_MeterType", 25, "Thermal Well")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_MeterType", 26, "Turbine")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_MeterType", 27, "Ultrasonic")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_OdorizerManufacturer", "List of valid manufacturers of gas Odorizer systems", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_OdorizerManufacturer", "U", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_OdorizerModel", "List of valid models of gas odorizer system models", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_OdorizerModel", "UN", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_OverPressureProtectType", "List of types of over pressure protection for gas facilities", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_OverPressureProtectType", "Unknown", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_OverPressureProtectType", "Full Capacity Relief Valve", "Full Capacity Relief Valve")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_OverPressureProtectType", "Monitor and Operator", "Monitor and Operator")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_OverPressureProtectType", "Monitor Only", "Monitor Only")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_OverPressureProtectType", "Operator and Monitor", "Operator and Monitor")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_OverPressureProtectType", "Operator Only", "Operator Only")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_PipeOutsideDiameter", "List of outside pipe diameter", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeOutsideDiameter", 0, "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_PresentStatus", "Boolean list indicating whether a flow device is operational", "LONG", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PresentStatus", 0, "Offline")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PresentStatus", 1, "Online")    

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_PiggingStructureMaterial", "List of types of material for pigging structures", "TEXT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PiggingStructureMaterial", "Unknown", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PiggingStructureMaterial", "Steel", "Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PiggingStructureMaterial", "Concrete", "Concrete")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_PipeCasingFillType", "List of valid type of fill material within a pipe casing", "SHORT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeCasingFillType", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeCasingFillType", 1, "Sand")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_PipeFittingMaterial", "List of valid types of material for fittings and strainers", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeFittingMaterial", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeFittingMaterial", "Plastic", "Plastic")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeFittingMaterial", "Cast Iron", "Cast Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeFittingMaterial", "Steel", "Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeFittingMaterial", "Copper", "Copper")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeFittingMaterial", "Fiberglass", "Fiberglass")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeFittingMaterial", "Wrought Iron", "Wrought Iron")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_PipeMaterial", "List of valid pipe materials", "LONG", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeMaterial", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeMaterial", 1, "Bare Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeMaterial", 2, "Cast Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeMaterial", 3, "Coated Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeMaterial", 4, "Copper")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeMaterial", 5, "Plastic PE")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeMaterial", 6, "Wrought Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeMaterial", 7, "Ductile Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeMaterial", 8, "Plastic PVC")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeMaterial", 9, "Composite")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeMaterial", 10, "Plastic ABS")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeMaterial", 11, "Reconditioned Cast Iron")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_PipeWallThickness", "List of valid wall thickness of Pipe", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_PipeWallThickness", 0, "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", "List of valid sizes for Gas Junction Reducer Fittings", "SHORT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 1, '1" x 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 2, '1" x 5/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 3, '1" x 3/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 4, '1.25" x 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 5, '1.25" x 5/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 6, '1.25" x 3/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 7, '1.25" x 7/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 8, '1.25" x 1"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 9, '1 1/2" x 3/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 10, '1 1/2" x 1"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 11, '1 1/2" x 1 1/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 12, '2" x 1"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 13, '2" x 1 1/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 14, '2" x 1 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 15, '3" x 1 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 16, '3" x 2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 17, '4" x 2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 18, '4" x 3"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 19, '6" x 3"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 20, '6" x 4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 21, '8" x 4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 22, '8" x 6"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 23, '10" x 6"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 24, '10" x 8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 25, '12" x 6"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 26, '12" x 8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 27, '12" x 10"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 28, '14" x 8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 29, '14" x 10"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 30, '14" x 12"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 31, '16" x 8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 32, '16" x 10"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 33, '16" x 12"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 34, '16" x 14"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 35, '18" x 10"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 36, '18" x 12"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 37, '18" x 14"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 38, '18" x 16"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 39, '20" x 10"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 40, '20" x 12"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 41, '20" x 14"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 42, '20" x 16"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 43, '20" x 18"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 44, '24" x 12"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 45, '24" x 14"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 46, '24" x 16"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 47, '24" x 18"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 48, '24" x 20"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 49, '26" x 14"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 50, '26" x 16"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 51, '26" x 18"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 52, '26" x 20"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 53, '26" x 24"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 54, '30" x 16"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 55, '30" x 18"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 56, '30" x 20"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 57, '30" x 24"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 58, '30" x 26"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 59, '36" x 18"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 60, '36" x 20"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 61, '36" x 24"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 62, '36" x 26"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 63, '40" x 20"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 64, '40" x 24"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 65, '40" x 26"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 66, '40" x 30"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 67, '40" x 36"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 68, '48" x 24"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 69, '48" x 26"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 70, '48" x 30"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ReducerFittingSize", 71, '48" x 36"')

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", "List of NPS outside diameter for Riser Steel Pipe Material", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 0.540, "0.540")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 0.675, "0.675")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 0.840, "0.840")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 1.050, "1.050")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 1.315, "1.315")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 1.660, "1.660")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 1.900, "1.900")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 2.375, "2.375")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 2.875, "2.875")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 3.500, "3.500")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 4.000, "4.000")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 4.500, "4.500")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 5.000, "5.000")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 5.563, "5.563")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 6.625, "6.625")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 7.625, "7.625")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 8.625, "8.625")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 9.625, "9.625")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 10.75, "10.75")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 12.75, "12.75")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 14, "14")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 16, "16")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 18, "18")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 20, "20")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 22, "22")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 24, "24")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 26, "26")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 28, "28")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 30, "30")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 32, "32")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 34, "34")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 36, "36")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 40, "40")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 42, "42")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 44, "44")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 46, "46")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 48, "48")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 52, "52")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 56, "56")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 60, "60")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 64, "64")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 68, "68")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 72, "72")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 76, "76")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 80, "80")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeOutsideDiameter", 88, "88")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", "List of NPS valid wall thickness of Riser Steel Pipe", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.068, "0.068")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.083, "0.083")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.088, "0.088")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.091, "0.091")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.109, "0.109")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.113, "0.113")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.120, "0.120")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.133, "0.133")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.134, "0.134")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.140, "0.140")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.145, "0.145")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.148, "0.148")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.154, "0.154")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.165, "0.165")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.180, "0.180")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.203, "0.203")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.216, "0.216")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.226, "0.226")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.237, "0.237")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.247, "0.247")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.250, "0.250")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.258, "0.258")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.277, "0.277")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.280, "0.280")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.281, "0.281")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.301, "0.301")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.307, "0.307")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.312, "0.312")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.322, "0.322")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.330, "0.330")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.337, "0.337")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.342, "0.342")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.355, "0.355")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.365, "0.365")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.375, "0.375")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.406, "0.406")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.432, "0.432")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.437, "0.437")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.500, "0.500")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.562, "0.562")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.593, "0.593")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.625, "0.625")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.656, "0.656")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.750, "0.750")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.812, "0.812")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RiserPipeWallThickness", 0.968, "0.968")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_RegulatorCasingMaterial", "List of valid types of material which can be used for regulator casing material", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RegulatorCasingMaterial", "Unkown", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RegulatorCasingMaterial", "Aluminum", "Aluminum")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RegulatorCasingMaterial", "Cast Iron", "Cast Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RegulatorCasingMaterial", "Ductile Iron", "Ductile Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RegulatorCasingMaterial", "Steel", "Steel")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_RegulatoryType", "List of valid types of material which can be used for regulator casing material", "SHORT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RegulatoryType", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RegulatoryType", 1, "Distribution")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RegulatoryType", 2, "Transmission Operated Distribution")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RegulatoryType", 3, "Transmission")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RegulatoryType", 4, "Offshore Transmission")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RegulatoryType", 5, "Gathering")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_RegulatoryType", 6, "Offshore Gathering")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_SpatialAccuracy", "Assessment of the spatial accuracy of the XY coordinates", "SHORT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SpatialAccuracy", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SpatialAccuracy", 1, 'Greater than 300 Meters')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SpatialAccuracy", 2, '101-300 Meters')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SpatialAccuracy", 3, '31-100 Meters')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SpatialAccuracy", 4, '11-30 Meters')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SpatialAccuracy", 5, '1-10 Meters')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SpatialAccuracy", 6, '0.3-1 Meter')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SpatialAccuracy", 7, '11-29 Centimeters')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SpatialAccuracy", 8, '1-10 Centimeters')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SpatialAccuracy", 9, '1-9 Millimeters')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SpatialAccuracy", 10, 'Less than 1 Millimeter')

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_SeamOrientation", "Clock based description of the orientation of the seam", "DOUBLE", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamOrientation", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamOrientation", 1, "1:00")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamOrientation", 2, "2:00")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamOrientation", 3, "3:00")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamOrientation", 4, "4:00")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamOrientation", 5, "5:00")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamOrientation", 6, "6:00")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamOrientation", 7, "7:00")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamOrientation", 8, "8:00")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamOrientation", 9, "9:00")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamOrientation", 10, "10:00")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamOrientation", 11, "11:00")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamOrientation", 12, "12:00")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_SeamType", "List of valid types of longitudinal seams", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamType", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamType", 1, "Seamless")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamType", 2, "Electric Weld")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamType", 3, "Lap Weld")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamType", 4, "Butt Weld")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamType", 5, "Single Submerged Arc Weld")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamType", 6, "Single Submerged Arc Weld Spiraled")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamType", 7, "Double Submerged Arc Weld")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamType", 8, "Double Submerged Arc Weld Spiraled")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SeamType", 9, "Not Applicable")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", "List of valid nominal diameter sizes for station pipe", "DOUBLE", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 4, '4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 6, '6"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 8, '8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 10, '10"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 12, '12"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 14, '14"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 16, '16"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 18, '18"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 20, '20"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 22, '22"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 24, '24"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 26, '26"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 28, '28"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 30, '30"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 36, '36"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 40, '40"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeDiameter", 48, '48"')

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", "List of NPS outside diameter for Station Steel Pipe Material", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 0.540, "0.540")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 0.675, "0.675")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 0.840, "0.840")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 1.050, "1.050")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 1.315, "1.315")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 1.660, "1.660")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 1.900, "1.900")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 2.375, "2.375")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 2.875, "2.875")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 3.500, "3.500")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 4.000, "4.000")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 4.500, "4.500")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 5.000, "5.000")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 5.563, "5.563")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 6.625, "6.625")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 7.625, "7.625")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 8.625, "8.625")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 9.625, "9.625")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 10.75, "10.75")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 12.75, "12.75")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 14, "14")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 16, "16")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 18, "18")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 20, "20")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 22, "22")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 24, "24")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 26, "26")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 28, "28")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 30, "30")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 32, "32")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 34, "34")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 36, "36")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 40, "40")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 42, "42")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 44, "44")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 46, "46")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 48, "48")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 52, "52")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 56, "56")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 60, "60")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 64, "64")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 68, "68")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 72, "72")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 76, "76")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 80, "80")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeOutsideDiameter", 88, "88")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", "List of NPS valid wall thickness of Station Steel Pipe", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0, "Unknown")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.068, "0.068")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.083, "0.083")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.088, "0.088")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.091, "0.091")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.109, "0.109")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.113, "0.113")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.120, "0.120")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.133, "0.133")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.134, "0.134")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.140, "0.140")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.145, "0.145")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.148, "0.148")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.154, "0.154")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.165, "0.165")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.180, "0.180")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.203, "0.203")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.216, "0.216")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.226, "0.226")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.237, "0.237")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.247, "0.247")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.250, "0.250")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.258, "0.258")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.277, "0.277")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.280, "0.280")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.281, "0.281")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.301, "0.301")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.307, "0.307")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.312, "0.312")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.322, "0.322")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.330, "0.330")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.337, "0.337")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.342, "0.342")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.355, "0.355")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.365, "0.365")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.375, "0.375")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.406, "0.406")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.432, "0.432")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.437, "0.437")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.500, "0.500")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.562, "0.562")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.593, "0.593")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.625, "0.625")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.656, "0.656")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.750, "0.750")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.812, "0.812")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StationPipeWallThickness", 0.968, "0.968")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrJunctAnchorManuf", "List of valid manufacturers of oil and gas anchors or river weights", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctAnchorManuf", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrJunctAnchorManComp", "List of valid manufacturers models of oil and gas anchors or river weights", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctAnchorManComp", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrJunctAnchorMaterial", "List of valid materials of oil and gas anchors or river weights", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctAnchorMaterial", "0", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctAnchorMaterial", "1", "Concrete")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrJunctMufflerManuf", "List of valid manufacturers of gas compressor station mufflers", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctMufflerManuf", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrJunctMufflerManComp", "List of valid manufacturer models of gas compressor station muffler", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctMufflerManComp", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrJunctMufflerMaterial", "List of valid materials of oil and gas mufflers", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctMufflerMaterial", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppManuf", "List of valid manufacturers of oil and gas pipe supports", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppManuf", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppManComp", "List of valid manufacturers models of oil and gas pipe supports", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppManComp", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "List of valid materials of oil and gas pipe supports", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "O", "Cast Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "R", "Malleable Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "S", "Ductile Iron")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "0", "Grade A")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "1", "Grade B")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "2", "Grade C")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "3", "Grade X42")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "4", "Grade X46")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "5", "Grade X52")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "6", "Grade X56")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "7", "Grade X60")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "8", "Grade X65")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "9", "Grade X70")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "M", "Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "N", "Stainless Steel")   
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeSuppMaterial", "X", "Other")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangManuf", "List of valid manufacturers of oil and gas pipe hangers", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangManuf", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangManComp", "List of valid manufacturers models of oil and gas pipe hangers", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangManComp", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "List of valid materials of oil and gas pipe hangers", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "UN", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "O", "Cast Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "R", "Malleable Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "S", "Ductile Iron")    
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "0", "Grade A")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "1", "Grade B")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "2", "Grade C")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "3", "Grade X42")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "4", "Grade X46")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "5", "Grade X52")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "6", "Grade X56")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "7", "Grade X60")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "8", "Grade X65")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "9", "Grade X70")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "M", "Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "N", "Stainless Steel")   
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctPipeHangMaterial", "X", "Other")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrJunctEndConnectManuf", "List of valid manufacturers of oil and gas pipe casing end connections", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctEndConnectManuf", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrJunctEndConnectManComp", "List of valid manufacturers models of oil and gas pipe casing end connections", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctEndConnectManComp", "0", "Unknown")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrJunctEndConnectMaterial", "List of valid materials of oil and gas end connections", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrJunctEndConnectMaterial", "0", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", "List of NPS outside diameter for Transmission Steel Pipe Material", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 0, "Unknown") 
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 4, "4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 6, "6")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 8, "8")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 10, "10")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 12, "12")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 14, "14")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 16, "16")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 18, "18")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 20, "20")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 22, "22")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 24, "24")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 26, "26")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 28, "28")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 30, "30")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 32, "32")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 34, "34")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 36, "36")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 40, "40")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 42, "42")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 44, "44")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 46, "46")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_StrLinePCasingDiameter", 48, "48")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_SurfaceType", "List of valid types of cover over the buried gas asset", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SurfaceType", "0", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SurfaceType", "Asphalt", "Asphalt")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SurfaceType", "Agriculture Crop", "Agriculture Crop")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SurfaceType", "Bare", "Bare")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SurfaceType", "Concrete", "Concrete")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SurfaceType", "Grass", "Grass")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SurfaceType", "Gravel", "Gravel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_SurfaceType", "Above Ground", "Above Ground")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_TankManufacturer", "List of valid manufacturers of oil and gas storage tanks", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TankManufacturer", "0", "Unknown")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_TankModel", "List of valid models of oil and gas storage tanks", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TankModel", "0", "Unknown")    

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", "List of NPS outside diameter for Transmission Steel Pipe Material", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 0, "Unknown") 
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 0.540, "0.540")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 0.675, "0.675")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 0.840, "0.840")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 1.050, "1.050")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 1.315, "1.315")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 1.660, "1.660")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 1.900, "1.900")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 2.375, "2.375")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 2.875, "2.875")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 3.500, "3.500")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 4.000, "4.000")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 4.500, "4.500")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 5.000, "5.000")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 5.563, "5.563")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 6.625, "6.625")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 7.625, "7.625")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 8.625, "8.625")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 9.625, "9.625")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 10.75, "10.75")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 12.75, "12.75")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 14, "14")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 16, "16")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 18, "18")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 20, "20")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 22, "22")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 24, "24")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 26, "26")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 28, "28")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 30, "30")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 32, "32")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 34, "34")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 36, "36")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 40, "40")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 42, "42")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 44, "44")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 46, "46")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 48, "48")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 52, "52")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 56, "56")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 60, "60")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 64, "64")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 68, "68")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 72, "72")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 76, "76")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 80, "80")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeOutsideDiameter", 88, "88")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", "List of NPS valid wall thickness of Transmission Steel Pipe", "DOUBLE", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0, "Unknown") 
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.068, "0.068")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.083, "0.083")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.088, "0.088")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.091, "0.091")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.109, "0.109")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.113, "0.113")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.120, "0.120")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.133, "0.133")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.134, "0.134")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.140, "0.140")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.145, "0.145")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.148, "0.148")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.154, "0.154")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.165, "0.165")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.180, "0.180")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.203, "0.203")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.216, "0.216")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.226, "0.226")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.237, "0.237")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.247, "0.247")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.250, "0.250")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.258, "0.258")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.277, "0.277")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.280, "0.280")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.281, "0.281")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.301, "0.301")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.307, "0.307")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.312, "0.312")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.322, "0.322")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.330, "0.330")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.337, "0.337")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.342, "0.342")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.355, "0.355")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.365, "0.365")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.375, "0.375")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.406, "0.406")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.432, "0.432")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.437, "0.437")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.500, "0.500")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.562, "0.562")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.593, "0.593")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.625, "0.625")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.656, "0.656")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.750, "0.750")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.812, "0.812")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_TransPipeWallThickness", 0.968, "0.968")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_UnitsForDepth", "List of valid units of measure for pipe depth below ground surface", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForDepth", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForDepth", 1, "Inch")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForDepth", 2, "Foot")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForDepth", 3, "Centimeter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForDepth", 4, "Meter")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_UnitsForDiameter", "List of valid units of measure for pipe diameter", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForDiameter", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForDiameter", 1, "Inch")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForDiameter", 2, "Foot")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForDiameter", 3, "Millimeter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForDiameter", 4, "Meter")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_UnitsForFlow", "List of valid units of measure of the volumetric flow", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForFlow", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForFlow", 1, "Cubic Feet/sec")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForFlow", 2, "Cubic Inches/sec")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForFlow", 3, "Cubic Meters/sec")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForFlow", 4, "Cubic Centimeters/sec")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_UnitsForHeight", "List of valid units of measure for structure height above ground surface", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForHeight", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForHeight", 1, "Inch")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForHeight", 2, "Foot")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForHeight", 3, "Centimeter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForHeight", 4, "Meter")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_UnitsForHeat", "List of valid units of measure for heat generation", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForHeat", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForHeat", 1, "BTU/Hour") 

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_UnitsForLength", "List of valid types of measures of length", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForLength", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForLength", 1, "Foot")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForLength", 2, "Mile")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForLength", 3, "Meter")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForLength", 4, "Kilometer")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_UnitsForPressure", "List of valid units of measure of pressure", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForPressure", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForPressure", 1, "Pounds/Square Inch")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForPressure", 2, "Inch Water Column")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_UnitsForSurfaceArea", "List of valid units of measure of surface area", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForSurfaceArea", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForSurfaceArea", 1, "Square Feet")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForSurfaceArea", 2, "Square Inches")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForSurfaceArea", 3, "Square Meters")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForSurfaceArea", 4, "Square Centimeters")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_UnitsForVolume", "List of valid units of measure of pipe volume", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForVolume", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForVolume", 1, "Cubic Feet")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForVolume", 2, "Cubic Inches")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForVolume", 3, "Cubic Meters")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitsForVolume", 4, "Cubic Centimeters")
    
    arcpy.CreateDomain_management(targetWorkspace, "UPDM_UnitRateForVolume", "List of valid units of measure of the rate of flow pipe volume", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitRateForVolume", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitRateForVolume", 1, "Cubic Feet/Second")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitRateForVolume", 2, "Cubic Inches/Second")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitRateForVolume", 3, "Cubic Meters/Second")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UnitRateForVolume", 4, "Cubic Centimeters/Second")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", "List of valid diameter sizes for utility meters", "DOUBLE", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 0.25, '1/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 0.38, '3/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 0.50, '1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 0.63, '5/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 0.75, '3/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 1.0, '1"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 1.125, '1 1/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 1.25, '1 1/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 1.375, '1 3/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 1.5, '1 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 1.875, '1 7/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 2, '2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 2.5, '2 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 3, '3"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 4, '4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 5, '5"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 5.5, '5 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 6, '6"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 8, '8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 10, '10"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 12, '12"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 14, '14"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 16, '16"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 18, '18"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 20, '20"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 24, '24"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 26, '26"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 30, '30"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 36, '36"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 40, '40"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_UtilityMeterDiameter", 48, '48"')

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ValvePosition", "Boolean value to determine whether the valve is open or closed.", "SHORT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValvePosition", 0, "Closed")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValvePosition", 1, "Open")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ValveDiameter", "List of valid diameter sizes for valves", "DOUBLE", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 0.25, '1/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 0.38, '3/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 0.50, '1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 0.63, '5/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 0.75, '3/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 1.0, '1"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 1.125, '1 1/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 1.25, '1 1/4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 1.375, '1 3/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 1.5, '1 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 1.875, '1 7/8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 2, '2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 2.5, '2 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 3, '3"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 4, '4"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 5, '5"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 5.5, '5 1/2"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 6, '6"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 8, '8"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 10, '10"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 12, '12"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 14, '14"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 16, '16"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 18, '18"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 20, '20"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 24, '24"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 26, '26"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 30, '30"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 36, '36"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 40, '40"')
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveDiameter", 48, '48"')

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ValveMaterial", "List of types of materials for valves", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveMaterial", "0", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveMaterial", "Brass", "Brass")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveMaterial", "Cast Iron", "Cast Iron")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveMaterial", "Fiberglass", "Fiberglass")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveMaterial", "Plastic", "Plastic")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveMaterial", "Steel", "Steel")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveMaterial", "Wrought Iron", "Wrought Iron")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ValveUseType", "List of uses for gas valves", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveUseType", "Unknown", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveUseType", "Bypass", "Bypass")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveUseType", "Isolation", "Isolation")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveUseType", "Purge", "Purge")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveUseType", "Blow-Off", "Blow-Off")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ValveControllableType", "List of types a controllable valves", "SHORT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveControllableType", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveControllableType", 1, "Ball")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveControllableType", 2, "Butterfly")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveControllableType", 3, "Gate")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveControllableType", 4, "Needle")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveControllableType", 5, "Pop")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveControllableType", 6, "Plug")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveControllableType", 7, "Stop")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveControllableType", 8, "Stop Cock")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveControllableType", 9, "V-Ball")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveControllableType", 10, "Other")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_ValveFlowType", "List of types of flow valves", "SHORT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveFlowType", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_ValveFlowType", 11, "Check")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_WellheadConnectionType", "List of valid types of connections for wellheads", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadConnectionType", "Unknown", "Unknown")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_WellheadDiameter", "List of valid nominal diameters for wellheads", "DOUBLE", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadDiameter", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadDiameter", 2.0, "2")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadDiameter", 2.5, "2.5")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadDiameter", 3.0, "3")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadDiameter", 4.0, "4")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadDiameter", 5.0, "5")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadDiameter", 6.0, "6")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_WellheadPressureRating", "List of valid pressure ratings for wellheads", "DOUBLE", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadPressureRating", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadPressureRating", 2000, "2000")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadPressureRating", 3000, "3000")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadPressureRating", 5000, "5000")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadPressureRating", 10000, "10000")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadPressureRating", 15000, "15000")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_WellheadMaterialClass", "List of valid classifications of the wellhead material", "TEXT", "CODED", "DEFAULT", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadMaterialClass", "Unknown", "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadMaterialClass", "AA", "AA")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadMaterialClass", "BB", "BB")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadMaterialClass", "CC", "CC")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadMaterialClass", "DD", "DD")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadMaterialClass", "EE", "EE")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadMaterialClass", "FF", "FF")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_WellheadMaterialClass", "HH", "HH")

    arcpy.CreateDomain_management(targetWorkspace, "UPDM_YesNoIndicator", "Boolean value for fields with a Yes or No input.", "SHORT", "CODED", "DUPLICATE", "DEFAULT")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_YesNoIndicator", 0, "Unknown")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_YesNoIndicator", 1, "Yes")
    arcpy.AddCodedValueToDomain_management(targetWorkspace, "UPDM_YesNoIndicator", 2, "No")
    
    logging.debug("Finished Creating the domains")
 
    ######################################
    # Defining GAS ASSEMBLY
    ######################################

    logging.debug("Setting up the UPDM ASSEMBLY class...")

    #Adding Attributes to UPDMAssembly featureclass
    logging.debug("     Adding 37 Attributes to Assembly") 
    arcpy.management.AddField(UPDMAssembly, "PIPELINEOPERATOR", "TEXT", None, None, 50, "Pipeline Operator", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "PRESSUREUNITS", "SHORT", None, None, None, "Pressure Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "BYPASSINDICATOR", "SHORT", None, None, None, "Bypass indicator", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "DESIGNCAPACITY", "DOUBLE", 16, 5, None, "Design Capacity", "NULLABLE", "NON_REQUIRED", None)    
    arcpy.management.AddField(UPDMAssembly, "DESIGNCAPACITYUNITS", "SHORT", None, None, None, "Design Capacity Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "DIAMETER", "DOUBLE", 16, 5, None, "Pipe Diameter", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "DIAMETERUNITS", "SHORT", None, None, None, "Diameter Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "PIPEGRADE", "TEXT", None, None, 2, "Pipe Grade", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "WALLTHICKNESS", "DOUBLE", 10, 6, None, "Pipe Wall Thickness", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "MANUFACTURER", "TEXT", None, None, 5, "Manufacturer", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("       Set 1 done")    
    arcpy.management.AddField(UPDMAssembly, "PIPEMATERIAL", "TEXT", None, None, 30, "Pipe Material", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "STRUCTURELENGTH", "DOUBLE", 16, 5, None, "Structure Length", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "LENGTHUNITS", "SHORT", None, None, None, "Length Units", "NULLABLE", "NON_REQUIRED", None)    
    arcpy.management.AddField(UPDMAssembly, "MATERIALCLASS", "TEXT", None, None, 20, "Material Class", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "CONNECTIONTYPE", "TEXT", None, None, 30, "Connection Type", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "OUTLETVALVETYPE", "TEXT", None, None, 30, "Outlet Valve Type", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "PRESSURERATINGUNITS", "SHORT", None, None, None, "Pressure Rating Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "ODORIZERINDICATOR", "SHORT", None, None, None, "Odorizer Indicator", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "SHUTOFFVALVEDIAMETER", "DOUBLE", 10, 6, None, "Shutoff Valve Diameter", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "SHUTOFFVALVETYPE", "TEXT", None, None, 25, "Shutoff Valve Type", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("       Set 2 done")    
    arcpy.management.AddField(UPDMAssembly, "TAPDIAMETER", "DOUBLE", 10, 6, None, "Tap Diameter", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "TAPTYPE", "TEXT", None, None, 20, "Tap Type", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "TOTALPOWER", "DOUBLE", 10, 6, None, "Total Power", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "NUMBEROFENGINES", "SHORT", None, None, None, "Number of Engines", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "NUMBEROFTURBINES", "SHORT", None, None, None, "Number of Turbines", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "NUMBEROFMOTORS", "SHORT", None, None, None, "Number of Motors", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "NUMBEROFUNITS", "SHORT", None, None, None, "Number of Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "COMMENTS", "TEXT", None, None, 255, "Comments", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "SPATIALACCURACY", "SHORT", None, None, None, "Spatial Accuracy", "NULLABLE", "NON_REQUIRED", None)   
    arcpy.management.AddField(UPDMAssembly, "GPSX", "DOUBLE", 16, 8, None, "GPS X", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("       Set 3 done")     
    arcpy.management.AddField(UPDMAssembly, "GPSY", "DOUBLE", 16, 8, None, "GPS Y", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "GPSZ", "DOUBLE", 16, 8, None, "GPS Z", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "LEGACYID", "TEXT", None, None, 38, "LegacyID", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "WORKORDERID", "TEXT", None, None, 38, "Work OrderID", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "PROJECTNUMBER", "TEXT", None, None, 38, "Project Number", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "ASSETUNITCODE", "TEXT", None, None, 38, "Asset Unit Code", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMAssembly, "SERVICESUMMARYID", "GUID", None, None, None, "Service Summary ID", "NULLABLE", "NON_REQUIRED", None)

    logging.debug("     Successfully Added 37 Attributes to Assembly")

    #Adding Domains to UPDMAssembly featureclass
    logging.debug("     Adding 27 Domains to Assembly")
  
    arcpy.management.AssignDomainToField(UPDMAssembly, "SPATIALACCURACY", "UPDM_SpatialAccuracy", "'2: Meter Setting';'6: Wellhead'")
    arcpy.management.AssignDomainToField(UPDMAssembly, "ROTATION", "UPDM_SymbolPosition", None)
    arcpy.management.AssignDomainToField(UPDMAssembly, "ROTATION", "UPDM_SymbolPosition", "0;1;2;3;4;5;6;7;8")
    arcpy.management.AssignDomainToField(UPDMAssembly, "VAULT", "UPDM_YesNoIndicator", None)
    arcpy.management.AssignDomainToField(UPDMAssembly, "VAULT", "UPDM_YesNoIndicator", "0;1;2;3;4;5;6;7;8")
    arcpy.management.AssignDomainToField(UPDMAssembly, "REGULATORYTYPE", "UPDM_RegulatoryType", None)
    arcpy.management.AssignDomainToField(UPDMAssembly, "REGULATORYTYPE", "UPDM_RegulatoryType", "0;1;2;3;4;5;6;7;8")
    arcpy.management.AssignDomainToField(UPDMAssembly, "LIFECYCLESTATUS", "UPDM_LifeCycleStatus", None)
    arcpy.management.AssignDomainToField(UPDMAssembly, "LIFECYCLESTATUS", "UPDM_LifeCycleStatus", "0;1;2;3;4;5;6;7;8")
    arcpy.management.AssignDomainToField(UPDMAssembly, "PRESSUREUNITS", "UPDM_UnitsForPressure", "0;1;2;3;4;5;6;7;8")
    logging.debug("       Set 1 done")     
    arcpy.management.AssignDomainToField(UPDMAssembly, "DIAMETERUNITS", "UPDM_UnitsForDiameter", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMAssembly, "ODORIZERINDICATOR", "UPDM_YesNoIndicator", "0;1;2;3;4;5;6")
    arcpy.management.AssignDomainToField(UPDMAssembly, "BYPASSINDICATOR", "UPDM_YesNoIndicator", "0;1;2;3;4;5;9")
    arcpy.management.AssignDomainToField(UPDMAssembly, "DIAMETER", "UPDM_StationPipeDiameter", "0;3;5;7")    
    arcpy.management.AssignDomainToField(UPDMAssembly, "WALLTHICKNESS", "UPDM_PipeWallThickness", "7: Pigging Structure")
    arcpy.management.AssignDomainToField(UPDMAssembly, "PIPEGRADE", "UPDM_ASTMTransPipeMaterial", "7: Pigging Structure")
    arcpy.management.AssignDomainToField(UPDMAssembly, "PIPEMATERIAL", "UPDM_PiggingStructureMaterial", "7: Pigging Structure")
    arcpy.management.AssignDomainToField(UPDMAssembly, "LENGTHUNITS", "UPDM_UnitsForLength", "7: Pigging Structure")   
    arcpy.management.AssignDomainToField(UPDMAssembly, "TAPDIAMETER", "Pipeline Valve Diameter", "4: Rural Tap")
    arcpy.management.AssignDomainToField(UPDMAssembly, "SHUTOFFVALVEDIAMETER", "Pipeline Valve Diameter", "4: Rural Tap")
    logging.debug("       Set 2 done")     
    arcpy.management.AssignDomainToField(UPDMAssembly, "SHUTOFFVALVETYPE", "UPDM_ASTMFValve", "4: Rural Tap")
    arcpy.management.AssignDomainToField(UPDMAssembly, "MATERIALCLASS", "UPDM_WellheadMaterialClass", "6: Wellhead")
    arcpy.management.AssignDomainToField(UPDMAssembly, "DIAMETER", "UPDM_WellheadDiameter", "6: Wellhead")
    arcpy.management.AssignDomainToField(UPDMAssembly, "WALLTHICKNESS", "UPDM_PipeWallThickness", "6: Wellhead")   
    arcpy.management.AssignDomainToField(UPDMAssembly, "CONNECTIONTYPE", "UPDM_WellheadConnectionType", "6: Wellhead")
    arcpy.management.AssignDomainToField(UPDMAssembly, "PRESSURERATING", "UPDM_WellheadPressureRating", "6: Wellhead")
    arcpy.management.AssignDomainToField(UPDMAssembly, "PRESSURERATINGUNITS", "UPDM_UnitsForPressure", "6: Wellhead")

    logging.debug("     Successfully Added 27 Domains to Assembly")

    #Assigning Default values to UPDMAssembly attributes
    logging.debug("     Assigning 22 Default Values to Assembly")
    arcpy.management.AssignDefaultToField(UPDMAssembly, "LIFECYCLESTATUS", 3, "0;1;2;3;4;5;6;7", False) 
    arcpy.management.AssignDefaultToField(UPDMAssembly, "REGULATORYTYPE", 1, "0;2;3;8", False)    
    arcpy.management.AssignDefaultToField(UPDMAssembly, "REGULATORYTYPE", 5, "6", False)
    arcpy.management.AssignDefaultToField(UPDMAssembly, "REGULATORYTYPE", 3, "7", False)
    arcpy.management.AssignDefaultToField(UPDMAssembly, "VAULT", 0, "0;1;2;3;4;5;8", False)
    arcpy.management.AssignDefaultToField(UPDMAssembly, "VAULT", 2, "6;7", False)
    arcpy.management.AssignDefaultToField(UPDMAssembly, "SPATIALACCURACY", 0, "'2: MeterSetting';'6: Wellhead'", False)
    arcpy.management.AssignDefaultToField(UPDMAssembly, "PRESSUREUNITS", 1, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMAssembly, "DIAMETERUNITS", 1, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMAssembly, "ODORIZERINDICATOR", 0, "0;1;2;3;4;5;6", False)
    logging.debug("       Set 1 done")      
    arcpy.management.AssignDefaultToField(UPDMAssembly, "BYPASSINDICATOR", 0, "0;1;2;3;4;5", False)
    arcpy.management.AssignDefaultToField(UPDMAssembly, "DIAMETER", 0, "'3: RegulatorStation';'5: Town Border Station'", False)    
    arcpy.management.AssignDefaultToField(UPDMAssembly, "TAPDIAMETER", 0, "'4: Rural Tap'", False)
    arcpy.management.AssignDefaultToField(UPDMAssembly, "WALLTHICKNESS", 0, "'6: Wellhead'", False)
    arcpy.management.AssignDefaultToField(UPDMAssembly, "SHUTOFFVALVEDIAMETER", 0, "'4: Rural Tap'", False)
    arcpy.management.AssignDefaultToField(UPDMAssembly, "SHUTOFFVALVETYPE", "0", "'4: Rural Tap'", False) 
    arcpy.management.AssignDefaultToField(UPDMAssembly, "CONNECTIONTYPE", "Unknown", "'6: Wellhead'", False)
    arcpy.management.AssignDefaultToField(UPDMAssembly, "PRESSURERATING", 0, "'6: Wellhead'", False)
    arcpy.management.AssignDefaultToField(UPDMAssembly, "MATERIALCLASS", "Unknown", "'6: Wellhead'", False)
    arcpy.management.AssignDefaultToField(UPDMAssembly, "DIAMETER", 0, "'6: Wellhead'", False)
    logging.debug("       Set 2 done")      
    arcpy.management.AssignDefaultToField(UPDMAssembly, "PRESSURERATINGUNITS", 1, "'6: Wellhead'", False)
    arcpy.management.AssignDefaultToField(UPDMAssembly, "PROTTYPE", 0, "3;4;5", False)

    logging.debug("     Successfully Assigned 22 Default Values to Assembly")

    ######################################
    # Defining GAS DEVICE
    ######################################
    logging.debug("Setting up the DEVICE class...")

    #Adding Attributes to UPDMDevice featureclass
    logging.debug("     Adding 50 Attributes to Device")
    arcpy.management.AddField(UPDMDevice, "ISCRITICAL", "SHORT", None, None, None, "Critical Valve (yes/no)", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "MAOPUNITS", "SHORT", None, None, None, "MAOP Unit of Measure", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "DIAMETERUNITS", "SHORT", None, None, None, "Diameter Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "LENGTHUNITS", "SHORT", None, None, None, "Length Units", "NULLABLE", "NON_REQUIRED", None)    
    arcpy.management.AddField(UPDMDevice, "RECORDINGMETHOD", "TEXT", None, None, 25, "Recording Method", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "FLOWCAPACITY", "DOUBLE", 16, 5, None, "Flow Capacity", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "FLOWCAPACITYUNITS", "SHORT", None, None, None, "Flow Capacity Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "CLOCKWISETOCLOSE", "SHORT", None, None, None, "Clockwise to Close", "NULLABLE", "NON_REQUIRED", None)                          
    arcpy.management.AddField(UPDMDevice, "DEPTHUNITS", "SHORT", None, None, None, "Depth Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "REGULATORBODYTYPE", "TEXT", None, None, 20, "Regulator", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("       Set 1 done")     
    arcpy.management.AddField(UPDMDevice, "CASINGMATERIAL", "TEXT", None, None, 20, "Casing Material", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "OUTLETDIAMETER", "DOUBLE", 16, 5, None, " Outlet Diameter", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "DIAPHRAGMDIAMETER", "DOUBLE", 16, 5, None, "Diaphragm Diameter", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "PRESSUREUNITS", "SHORT", None, None, None, "Set Pressure Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "PREMISE", "TEXT", None, None, 10, "Premise", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "INDOORMETER", "SHORT", None, None, None, "Indoor Meter", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "OPERATINGCLASSIFICATION", "TEXT", None, None, 40, "Operating Classification", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "MAXDIFFERENTIALPRESSURE", "DOUBLE", 16, 5, None, "Max Differential Pressure", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "DRAINVALVEINDICATOR", "SHORT", None, None, None, "Drain Valve Indicator", "NULLABLE", "NON_REQUIRED", None)                           
    arcpy.management.AddField(UPDMDevice, "COMPANYPAIDTELEPHONE", "SHORT", None, None, None, "Company Paid Telephone", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("       Set 2 done")      
    arcpy.management.AddField(UPDMDevice, "TELEPHONENUMBER", "TEXT", None, None, 15, "Telephone Number", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "DURATION", "LONG", None, None, None, "Duration", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "PERMANENTINDICATOR", "SHORT", None, None, None, "Permanent Indicator", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "TEMPRATING", "TEXT", None, None, 20, "Temperature Rating", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "USAGETYPE", "TEXT", None, None, 3, "Usage Type", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "HEIGHT", "DOUBLE", 16, 5, None, "Height", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "HEIGHTUNITS", "SHORT", None, None, None, "Height Units", "NULLABLE", "NON_REQUIRED", None)                            
    arcpy.management.AddField(UPDMDevice, "METERED", "SHORT", None, None, None, "Metered", "NULLABLE", "NON_REQUIRED", None)     
    arcpy.management.AddField(UPDMDevice, "USAGE", "DOUBLE", 16, 5, None, "Usage", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "USAGEUNITS", "SHORT", None, None, None, "Usage Units", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("       Set 3 done")    
    arcpy.management.AddField(UPDMDevice, "USAGERATE", "DOUBLE", 16, 5, None, "Usage Rate", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "USAGERATEUNITS", "SHORT", None, None, None, "Usage Rate Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "QUANTITY", "DOUBLE", 8, 2, None, "Quantity", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "WEIGHT", "DOUBLE", 16, 6, None, "Weight", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "AMPERAGE", "DOUBLE", 16, 6, None, "Amperage", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "COOLINGMETHOD", "TEXT", None, None, 20, "Cooling Method", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "INTERNALMETER", "SHORT", None, None, None, "Internal Meter", "NULLABLE", "NON_REQUIRED", None)                             
    arcpy.management.AddField(UPDMDevice, "VOLTAGE", "LONG", None, None, None, "Voltage", "NULLABLE", "NON_REQUIRED", None)    
    arcpy.management.AddField(UPDMDevice, "TESTPOINTNAME", "TEXT", None, None, 20, "Test Point Name", "NULLABLE", "NON_REQUIRED", None)    
    arcpy.management.AddField(UPDMDevice, "WIRECOUNT", "SHORT", None, None, None, "Wire Count", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("       Set 4 done")    
    arcpy.management.AddField(UPDMDevice, "SPATIALACCURACY", "SHORT", None, None, None, "Spatial Accuracy", "NULLABLE", "NON_REQUIRED", None)                            
    arcpy.management.AddField(UPDMDevice, "GPSX", "DOUBLE", 16, 8, None, "GPS X", "NULLABLE", "NON_REQUIRED", None)  
    arcpy.management.AddField(UPDMDevice, "GPSY", "DOUBLE", 16, 8, None, "GPS Y", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "GPSZ", "DOUBLE", 16, 8, None, "GPS Z", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "LEGACYID", "TEXT", None, None, 38, "LegacyID", "NULLABLE", "NON_REQUIRED", None) 
    arcpy.management.AddField(UPDMDevice, "WORKORDERID", "TEXT", None, None, 38, "Work OrderID", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "PROJECTNUMBER", "TEXT", None, None, 38, "Project Number", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "SERVICESUMMARYID", "GUID", None, None, None, "Service Summary ID", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "ASSETUNITCODE", "TEXT", None, None, 38, "Asset Unit Code", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMDevice, "COMMENTS", "TEXT", None, None, 255, "Comments", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("       Set 5 done")
    logging.debug("     Successfully Added 50 Attributes to Device")

    #Adding Domains to UPDMDevice featureclass
    logging.debug("     Adding 46 Domains to Device")

    arcpy.management.AssignDomainToField(UPDMDevice, "MANUFACTURER", "UPDM_ASTMManufacturer", "0;1;2;3;4;5;6;8;9;10;11;12;13;15;17;19;20")
    arcpy.management.AssignDomainToField(UPDMDevice, "MANUFACTURER", "UPDM_CPRectifierManufacture", "16: Rectifier")
    arcpy.management.AssignDomainToField(UPDMDevice, "MANUFACTURECOMPONENT", "UPDM_ASTMEFVModel", "1: Excess Flow Valve")
    arcpy.management.AssignDomainToField(UPDMDevice, "MANUFACTURECOMPONENT", "UPDM_ASTMMeterModel", "2: Meter")
    arcpy.management.AssignDomainToField(UPDMDevice, "MANUFACTURECOMPONENT", "UPDM_ASTMFValve", "3: Controllable Valve")
    arcpy.management.AssignDomainToField(UPDMDevice, "MANUFACTURECOMPONENT", "UPDM_ASTMFValve", "4: Flow Valve")
    arcpy.management.AssignDomainToField(UPDMDevice, "MANUFACTURECOMPONENT", "UPDM_ASTMRegulatorModel", "19: Regulator")
    arcpy.management.AssignDomainToField(UPDMDevice, "MANUFACTURECOMPONENT", "UPDM_ASTMStrainerModel", "6: Strainer")
    arcpy.management.AssignDomainToField(UPDMDevice, "MANUFACTURECOMPONENT", "UPDM_CPASTMF2897Anodes", "15: Anode")
    arcpy.management.AssignDomainToField(UPDMDevice, "MATERIAL", "UPDM_ASTMComponentMaterial", "1;3;4;5;6;12;13")
    logging.debug("       Set 1 done")    
    arcpy.management.AssignDomainToField(UPDMDevice, "MATERIAL", "UPDM_CPAnodeMaterialType", "15: Anode")
    arcpy.management.AssignDomainToField(UPDMDevice, "ROTATION", "UPDM_SymbolPosition", None)
    arcpy.management.AssignDomainToField(UPDMDevice, "ROTATION", "UPDM_SymbolPosition", "0;1;2;3;4;5;6;8;9;10;11;12;13;15;16;17;18;19;20")
    arcpy.management.AssignDomainToField(UPDMDevice, "DEVICESTATUS", "UPDM_DEVICESTATUS", None)
    arcpy.management.AssignDomainToField(UPDMDevice, "DEVICESTATUS", "UPDM_DEVICESTATUS", "0;1;2;3;4;5;6;8;9;10;11;12;13;15;16;17;18;19;20")
    arcpy.management.AssignDomainToField(UPDMDevice, "LIFECYCLESTATUS", "UPDM_LifeCycleStatus", None)
    arcpy.management.AssignDomainToField(UPDMDevice, "LIFECYCLESTATUS", "UPDM_LifeCycleStatus", "0;1;2;3;4;5;6;8;9;10;11;12;13;15;16;17;18;19;20")
    arcpy.management.AssignDomainToField(UPDMDevice, "REGULATORYTYPE", "UPDM_RegulatoryType", None)
    arcpy.management.AssignDomainToField(UPDMDevice, "REGULATORYTYPE", "UPDM_RegulatoryType", "0;1;2;3;4;5;6;8;9;10;11;12;13;15;16;17;18;19;20")
    arcpy.management.AssignDomainToField(UPDMDevice, "DIAMETER", "UPDM_EFVDiameter", "1: Excess Flow Valve")
    logging.debug("       Set 2 done")    
    arcpy.management.AssignDomainToField(UPDMDevice, "DIAMETER", "UPDM_CustomerMeterDiameter", "2: Meter")     
    arcpy.management.AssignDomainToField(UPDMDevice, "DIAMETER", "UPDM_ValveDiameter", "0;3;4;5;12;19")     
    arcpy.management.AssignDomainToField(UPDMDevice, "DIAMETERUNITS", "UPDM_UnitsForDiameter", "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20")
    arcpy.management.AssignDomainToField(UPDMDevice, "LENGTHUNITS", "UPDM_UnitsForLength", "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20")                                 
    arcpy.management.AssignDomainToField(UPDMDevice, "INSULATORDEVICE", "UPDM_INSULATORDEVICE", None)    
    arcpy.management.AssignDomainToField(UPDMDevice, "INSULATORDEVICE", "UPDM_INSULATORDEVICE", "0;1;2;3;4;5;6;8;9;10;11;12;13;15;16;17;18;19;20")
    arcpy.management.AssignDomainToField(UPDMDevice, "MAOPDESIGN", "UPDM_PressureRange", None)
    arcpy.management.AssignDomainToField(UPDMDevice, "MAOPDESIGN", "UPDM_PressureRange", "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20")
    arcpy.management.AssignDomainToField(UPDMDevice, "MAOPTEST", "UPDM_PressureRange", None)
    arcpy.management.AssignDomainToField(UPDMDevice, "MAOPTEST", "UPDM_PressureRange", "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20")
    logging.debug("       Set 3 done")     
    arcpy.management.AssignDomainToField(UPDMDevice, "MAOPRECORD", "UPDM_PressureRange", None)    
    arcpy.management.AssignDomainToField(UPDMDevice, "MAOPRECORD", "UPDM_PressureRange", "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20")    
    arcpy.management.AssignDomainToField(UPDMDevice, "MAOPUNITS", "UPDM_UnitsForPressure", "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20")                             
    arcpy.management.AssignDomainToField(UPDMDevice, "ACCESSIBLE", "UPDM_Accessible", None)
    arcpy.management.AssignDomainToField(UPDMDevice, "ACCESSIBLE", "UPDM_Accessible", "0;1;2;3;4;5;6;8;9;10;11;12;13;15;16;17;18;19;20")    
    arcpy.management.AssignDomainToField(UPDMDevice, "DEPTHUNITS", "UPDM_UnitsForDepth", "0;1;2;3;4;5;6;8;9;10;11;12;13")          
    arcpy.management.AssignDomainToField(UPDMDevice, "SPATIALACCURACY", "UPDM_SpatialAccuracy", "0;1;2;3;4;5;6;8;9;10;11;12;13;15;16;17;18;19;20")   
    arcpy.management.AssignDomainToField(UPDMDevice, "CLOCKWISETOCLOSE", "UPDM_YesNoIndicator", "3: Controllable Valve") 
    arcpy.management.AssignDomainToField(UPDMDevice, "INDOORMETER", "UPDM_YesNoIndicator", "2: Meter")   
    arcpy.management.AssignDomainToField(UPDMDevice, "PRESSUREUNITS", "UPDM_UnitsForPressure", "5: Relief Valve")
    logging.debug("       Set 3 done")    
    arcpy.management.AssignDomainToField(UPDMDevice, "DRAINVALVEINDICATOR", "UPDM_YesNoIndicator", "6: Strainer")   
    arcpy.management.AssignDomainToField(UPDMDevice, "COMPANYPAIDTELEPHONE", "UPDM_YesNoIndicator", "8: Pressure Monitoring Device")     
    arcpy.management.AssignDomainToField(UPDMDevice, "PERMANENTINDICATOR", "UPDM_YesNoIndicator", "8: Pressure Monitoring Device")
    arcpy.management.AssignDomainToField(UPDMDevice, "HEIGHTUNITS", "UPDM_UnitsForHeight", "9: Lamp")   
    arcpy.management.AssignDomainToField(UPDMDevice, "METERED", "UPDM_YesNoIndicator", "9: Lamp")    
    arcpy.management.AssignDomainToField(UPDMDevice, "PRESSUREUNITS", "UPDM_UnitsForPressure", "10;18")   
    arcpy.management.AssignDomainToField(UPDMDevice, "WEIGHT", "UPDM_CPAnodeWeight", "15: Anode")    
    arcpy.management.AssignDomainToField(UPDMDevice, "INTERNALMETER", "UPDM_YesNoIndicator", "16: Rectifier")    
    arcpy.management.AssignDomainToField(UPDMDevice, "FLOWCAPACITYUNITS", "UPDM_UnitsForVolume", "0;1")
    arcpy.management.AssignDomainToField(UPDMDevice, "CASINGMATERIAL", "UPDM_RegulatorCasingMaterial", "19: Regulator")
    logging.debug("       Set 4 done")     
    arcpy.management.AssignDomainToField(UPDMDevice, "USAGEUNITS", "UPDM_UnitsForVolume", "9: Lamp")   
    arcpy.management.AssignDomainToField(UPDMDevice, "USAGERATEUNITS", "UPDM_UnitRateForVolume", "19: Regulator")
    arcpy.management.AssignDomainToField(UPDMDevice, "WIRECOUNT", "UPDM_CPTestPointNumWires", "17: Test Point")
    arcpy.management.AssignDomainToField(UPDMDevice, "DESIGNTYPE", "UPDM_MeterType", "2")
    arcpy.management.AssignDomainToField(UPDMDevice, "DESIGNTYPE", "UPDM_ValveControllableType", "3")    
    arcpy.management.AssignDomainToField(UPDMDevice, "DESIGNTYPE", "UPDM_ValveFlowType", "3")
    logging.debug("     Successfully Added 46 Domains to Device")

    #Add Default Values to Gas Device Featureclass
    logging.debug("     Adding 37 Default Values to Device")
    
    arcpy.management.AssignDefaultToField(UPDMDevice, "MAOPUNITS", 1, "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20", False)   
    arcpy.management.AssignDefaultToField(UPDMDevice, "DIAMETERUNITS", 1, "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "LENGTHUNITS", 1, "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20", False)                              
    arcpy.management.AssignDefaultToField(UPDMDevice, "DEPTHUNITS", 2, "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20", False)    
    arcpy.management.AssignDefaultToField(UPDMDevice, "DESIGNTYPE", 0, "2: Meter", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "DESIGNTYPE", 0, "3: Controllable Valve", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "REGULATORYTYPE", 0, "0;3;4;5;6;8;12;13;15;16;17;20", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "REGULATORYTYPE", 1, "1;2;9;19", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "REGULATORYTYPE", 3, "10;18", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "REGULATORYTYPE", 5, "11", False)
    logging.debug("       Set 1 done")     
    arcpy.management.AssignDefaultToField(UPDMDevice, "SPATIALACCURACY", 0, "0;1;2;3;4;5;6;8;9;10;11;12;15;16;17;18;19;20", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "PRESSUREUNITS", 1, "5;10;18", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "DRAINVALVEINDICATOR", 0, "6: Strainer", False)     
    arcpy.management.AssignDefaultToField(UPDMDevice, "COMPANYPAIDTELEPHONE", 0, "8: Pressure Monitoring Device", False)    
    arcpy.management.AssignDefaultToField(UPDMDevice, "PERMANENTINDICATOR", 0, "8: Pressure Monitoring Device", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "HEIGHTUNITS", 2, "9: Lamp", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "INSULATORDEVICE", 0, "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "INSULATORDEVICE", 2, "15;16;17", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "ACCESSIBLE", 1, "0;1;2;3;4;5;6;8;9;10;11;12;13;15;16;17;18;19;20", False)    
    arcpy.management.AssignDefaultToField(UPDMDevice, "NORMALPOSITION", 1, "1;3;12", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "DEVICESTATUS", 0, "9", False)
    logging.debug("       Set 2 done")    
    arcpy.management.AssignDefaultToField(UPDMDevice, "DEVICESTATUS", 1, "1;2;3;4;5;6;8;10;11;12;13;15;16;17;18;19;20", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "MANUFACTURER", "0", "0;1;2;3;4;5;6;8;9;10;11;12;13;16;18;19;20", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "MANUFACTURECOMPONENT", "0", "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "MANUFACTURECOMPONENT", "UN", "15: Anode", False) 
    arcpy.management.AssignDefaultToField(UPDMDevice, "ROTATION", 0, "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20", False)   
    arcpy.management.AssignDefaultToField(UPDMDevice, "DIAMETER", 0, "0;1;2;3", False)   
    arcpy.management.AssignDefaultToField(UPDMDevice, "DIAMETERUNITS", 1, "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "LENGTHUNITS", 1, "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20", False)                              
    arcpy.management.AssignDefaultToField(UPDMDevice, "DEPTHUNITS", 2, "0;1;2;3;4;5;6;8;9;10;11;12;13;18;19;20", False) 
    arcpy.management.AssignDefaultToField(UPDMDevice, "METERED", 0, "9: Lamp", False)
    logging.debug("       Set 3 done")     
    arcpy.management.AssignDefaultToField(UPDMDevice, "WEIGHT", 0, "15: Anode", False) 
    arcpy.management.AssignDefaultToField(UPDMDevice, "INTERNALMETER", 0, "16: Rectifier", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "INDOORMETER", 0, "2: Meter", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "CLOCKWISETOCLOSE", 0, "3: Valve", False)      
    arcpy.management.AssignDefaultToField(UPDMDevice, "FLOWCAPACITYUNITS", 1, "0;1", False)   
    arcpy.management.AssignDefaultToField(UPDMDevice, "USAGEUNITS", 1, "9: Lamp", False)
    arcpy.management.AssignDefaultToField(UPDMDevice, "USAGERATEUNITS", 1, "19: Regulator", False)
    logging.debug("     Successfully Added 37 Default Values to Device")

    ######################################
    # Defining GAS JUNCTION
    ######################################
    logging.debug("Setting up the JUNCTION class...")

    #Adding Attributes to UPDMJunction featureclass
    logging.debug("     Adding 30 Attributes to Junction")
    arcpy.management.AddField(UPDMJunction, "MAOPUNITS", "SHORT", None, None, None, "MAOP Unit of Measure", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "DIAMETERUNITS", "SHORT", None, None, None, "Diameter Units", "NULLABLE", "NON_REQUIRED", None)                            
    arcpy.management.AddField(UPDMJunction, "LENGTHUNITS", "SHORT", None, None, None, "Length Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "HEIGHTUNITS", "SHORT", None, None, None, "Height Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "CAPACITY", "DOUBLE", 16, 5, None, "Capacity", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "CAPACITYUNITS", "SHORT", None, None, None, "Capacity Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "FLOWCAPACITY", "DOUBLE", 16, 5, None, "Flow Capacity", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "FLOWCAPACITYUNITS", "SHORT", None, None, None, "Flow Capacity Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "INSTALLPERSON", "TEXT", None, None, 50, "Install Person", "NULLABLE", "NON_REQUIRED", None)                            
    arcpy.management.AddField(UPDMJunction, "INSTALLMETHOD", "TEXT", None, None, 30, "Install Method", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("       Set 1 Done") 
    arcpy.management.AddField(UPDMJunction, "ROOFTYPE", "TEXT", None, None, 50, "Roof Type", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "LININGTYPE", "TEXT", None, None, 50, "Lining Type", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "HEATINGFUEL", "TEXT", None, None, 20, "Heating Fuel", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "HEATINGCAPACITY", "DOUBLE", 16, 5, None, "Heating Capacity", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "HEATINGCAPACITYUNITS", "SHORT", None, None, None, "Heating Capacity Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "ODORANTTYPE", "TEXT", None, None, 30, "Odorant Type", "NULLABLE", "NON_REQUIRED", None)                          
    arcpy.management.AddField(UPDMJunction, "WALLTHICKNESS", "DOUBLE", 16, 5, None, "Wallthickness", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "WALLTHICKNESSUNITS", "SHORT", None, None, None, "Wallthickness Units", "NULLABLE", "NON_REQUIRED", None)   
    arcpy.management.AddField(UPDMJunction, "PERCENTGRAPH", "DOUBLE", 16, 5, None, "Percent Graph", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "BONDJUNCTIONTYPE", "TEXT", None, None, 35, "Bond Junction Type", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("       Set 2 Done")
    arcpy.management.AddField(UPDMJunction, "CRITICALINDICATOR", "SHORT", None, None, None, "Critical Indicator", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "RESISTANCEINDICATOR", "SHORT", None, None, None, "Resistance Indicator", "NULLABLE", "NON_REQUIRED", None)                                   
    arcpy.management.AddField(UPDMJunction, "SPATIALACCURACY", "SHORT", None, None, None, "Spatial Accuracy", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "GPSX", "DOUBLE", 16, 8, None, "GPS X", "NULLABLE", "NON_REQUIRED", None)   
    arcpy.management.AddField(UPDMJunction, "GPSY", "DOUBLE", 16, 8, None, "GPS Y", "NULLABLE", "NON_REQUIRED", None)    
    arcpy.management.AddField(UPDMJunction, "GPSZ", "DOUBLE", 16, 8, None, "GPS Z", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "LEGACYID", "TEXT", None, None, 38, "LegacyID", "NULLABLE", "NON_REQUIRED", None)   
    arcpy.management.AddField(UPDMJunction, "ASSETUNITCODE", "TEXT", None, None, 38, "Asset Unit Code", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "PROJECTNUMBER", "TEXT", None, None, 38, "Project Number", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMJunction, "WORKORDERID", "TEXT", None, None, 38, "Work OrderID", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("     Successfully Added 30 Attributes to Junction")

    #Adding Domains to UPDMJunctions featureclass
    logging.debug("     Adding 49 Domains to Junction")

    arcpy.management.AssignDomainToField(UPDMJunction, "LIFECYCLESTATUS", "UPDM_LifeCycleStatus", None)
    arcpy.management.AssignDomainToField(UPDMJunction, "LIFECYCLESTATUS", "UPDM_LifeCycleStatus", "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20;21;22")
    arcpy.management.AssignDomainToField(UPDMJunction, "REGULATORYTYPE", "UPDM_RegulatoryType", None)
    arcpy.management.AssignDomainToField(UPDMJunction, "REGULATORYTYPE", "UPDM_RegulatoryType", "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20;21;22")
    arcpy.management.AssignDomainToField(UPDMJunction, "INSULATORDEVICE", "UPDM_YesNoIndicator", None)
    arcpy.management.AssignDomainToField(UPDMJunction, "INSULATORDEVICE", "UPDM_YesNoIndicator", "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20;21;22")
    arcpy.management.AssignDomainToField(UPDMJunction, "ROTATION", "UPDM_SymbolPosition", None)
    arcpy.management.AssignDomainToField(UPDMJunction, "ROTATION", "UPDM_SymbolPosition", "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20;21;22")    
    arcpy.management.AssignDomainToField(UPDMJunction, "DIAMETER", "UPDM_FittingDiameter", "0;1;2;3;4;5;8;9;10;11;12;13")
    arcpy.management.AssignDomainToField(UPDMJunction, "DESIGNTYPE", "UPDM_ReducerFittingSize", "6: Reducer")
    logging.debug("       Set 1 Done")     
    arcpy.management.AssignDomainToField(UPDMJunction, "MAOPDESIGN", "UPDM_PressureRange", None)
    arcpy.management.AssignDomainToField(UPDMJunction, "MAOPDESIGN", "UPDM_PressureRange", "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20")
    arcpy.management.AssignDomainToField(UPDMJunction, "MAOPTEST", "UPDM_PressureRange", None)
    arcpy.management.AssignDomainToField(UPDMJunction, "MAOPTEST", "UPDM_PressureRange", "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20")
    arcpy.management.AssignDomainToField(UPDMJunction, "MAOPRECORD", "UPDM_PressureRange", None)    
    arcpy.management.AssignDomainToField(UPDMJunction, "MAOPRECORD", "UPDM_PressureRange", "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20")
    arcpy.management.AssignDomainToField(UPDMJunction, "MAOPUNITS", "UPDM_UnitsForPressure", "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURER", "UPDM_ASTMManufacturer", "0;1;2;3;4;5;6;7;8;10;11;13") 
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURER", "UPDM_TankManufacturer", "14: Tank")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURER", "UPDM_LineHeaterManufacturer", "15: Line Heater")
    logging.debug("       Set 2 Done")     
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURER", "UPDM_CoolingSysManufacturer", "16: Cooling System")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURER", "UPDM_OdorizerManufacturer", "17: Odorizer")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURER", "UPDM_DehydrateManufacturer", "19: Dehydratation Equipment")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURER", "UPDM_DripManufacturer", "20: Drip")    
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURECOMPONENT", "UPDM_ASTMCouplingModel", "1: Coupling")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURECOMPONENT", "UPDM_ASTMElbowModel", "2: Elbow")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURECOMPONENT", "UPDM_ASTMEndCapModel", "3: End Cap")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURECOMPONENT", "UPDM_ASTMReducerModel", "6: Reducer")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURECOMPONENT", "UPDM_ASTMTeeModel", "7: Tee")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURECOMPONENT", "UPDM_ASTMTransitionModel", "8: Transition")
    logging.debug("       Set 3 Done")     
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURECOMPONENT", "UPDM_ASTMSaddleModel", "13: Saddle")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURECOMPONENT", "UPDM_TankModel", "14: Tank")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURECOMPONENT", "UPDM_LineHeaterModel", "15: Line Heater")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURECOMPONENT", "UPDM_CoolingSysModel", "16: Cooling System")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURECOMPONENT", "UPDM_OdorizerModel", "17: Odorizer")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURECOMPONENT", "UPDM_DehydrateModel", "19: Dehydration Equipment")
    arcpy.management.AssignDomainToField(UPDMJunction, "MANUFACTURECOMPONENT", "UPDM_DripModel", "20: Drip")    
    arcpy.management.AssignDomainToField(UPDMJunction, "MATERIAL", "UPDM_ASTMComponentMaterial", "0;1;2;3;4;5;6;7;8;10;11;13") 
    arcpy.management.AssignDomainToField(UPDMJunction, "DIAMETERUNITS", "UPDM_UnitsForDiameter", "0;1;2;3;4;5;6;7;9;11;12;13")    
    arcpy.management.AssignDomainToField(UPDMJunction, "LENGTHUNITS", "UPDM_UnitsForLength", "0;1;2;3;4;5;6;7;9;11;12;13;15;16;17;18;19")
    logging.debug("       Set 4 Done")     
    arcpy.management.AssignDomainToField(UPDMJunction, "SPATIALACCURACY", "UPDM_SpatialAccuracy", "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20;21;22")
    arcpy.management.AssignDomainToField(UPDMJunction, "BONDJUNCTIONTYPE", "UPDM_CPBondJunctionType", "22: Bond Junction")
    arcpy.management.AssignDomainToField(UPDMJunction, "RESISTANCEINDICATOR", "UPDM_YesNoIndicator", "'22: Bond Junction'")
    arcpy.management.AssignDomainToField(UPDMJunction, "CRITICALINDICATOR", "UPDM_YesNoIndicator", "'22: Bond Junction'")
    arcpy.management.AssignDomainToField(UPDMJunction, "CAPACITYUNITS", "UPDM_UnitsForVolume", "0;14;17;20")
    arcpy.management.AssignDomainToField(UPDMJunction, "HEIGHTUNITS", "UPDM_UnitsForHeight", "0;14")    
    arcpy.management.AssignDomainToField(UPDMJunction, "FLOWCAPACITYUNITS", "UPDM_UnitsForFlow", "0;14")
    arcpy.management.AssignDomainToField(UPDMJunction, "HEATINGCAPACITYUNITS", "UPDM_UnitsForHeat", "0;15")   
    arcpy.management.AssignDomainToField(UPDMJunction, "WALLTHICKNESSUNITS", "UPDM_UnitsForDiameter", "0;1;2;3;4;5;6;7;8;10;11;13;18")
    
    logging.debug("     Successfully Added 49 Domains to Junction")

    #Add Default Values to Gas Device Featureclass
    logging.debug("     Adding 20 Default Values to Junction")
    arcpy.management.AssignDefaultToField(UPDMJunction, "REGULATORYTYPE", 0, "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20;21", False)
    arcpy.management.AssignDefaultToField(UPDMJunction, "ROTATION", 0, "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20;21", False)
    arcpy.management.AssignDefaultToField(UPDMJunction, "MAOPDESIGN", 0, "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20;21;22", False)
    arcpy.management.AssignDefaultToField(UPDMJunction, "MAOPTEST", 0, "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20;21;22", False)
    arcpy.management.AssignDefaultToField(UPDMJunction, "MAOPUNITS", 1, "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20;21;22", False)
    arcpy.management.AssignDefaultToField(UPDMJunction, "MAOPRECORD", 0, "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20;21;22", False)
    arcpy.management.AssignDefaultToField(UPDMJunction, "MAOPUNITS", 1, "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20;21;22", False)    
    arcpy.management.AssignDefaultToField(UPDMJunction, "MANUFACTURER", "0", "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20", False)   
    arcpy.management.AssignDefaultToField(UPDMJunction, "MANUFACTURECOMPONENT", "0", "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20", False)   
    arcpy.management.AssignDefaultToField(UPDMJunction, "DIAMETERUNITS", 1, "0;1;2;3;4;5;6;8;10;11;12;13", False)
    logging.debug("       Set 1 Done")     
    arcpy.management.AssignDefaultToField(UPDMJunction, "LENGTHUNITS", 1, "0;1;2;3;4;5;6;8;10;11;12;13;15;16;17;18;19", False)        
    arcpy.management.AssignDefaultToField(UPDMJunction, "SPATIALACCURACY", 0, "0;1;2;3;4;5;6;7;8;9;10;11;12;13;14;15;16;17;18;19;20;21;22", False)    
    arcpy.management.AssignDefaultToField(UPDMJunction, "BONDJUNCTIONTYPE", "Unknown", "22: Bond Junction", False)
    arcpy.management.AssignDefaultToField(UPDMJunction, "RESISTANCEINDICATOR", 0, "'22: Bond Junction'", False)
    arcpy.management.AssignDefaultToField(UPDMJunction, "CRITICALINDICATOR", 0, "'22: Bond Junction'", False)
    arcpy.management.AssignDefaultToField(UPDMJunction, "CAPACITYUNITS", 1, "0;14;17;20", False)   
    arcpy.management.AssignDefaultToField(UPDMJunction, "HEIGHTUNITS", 1, "0;14", False)
    arcpy.management.AssignDefaultToField(UPDMJunction, "FLOWCAPACITYUNITS", 1, "0;14", False)
    arcpy.management.AssignDefaultToField(UPDMJunction, "HEATINGCAPACITYUNITS", 1, "0;15", False)
    arcpy.management.AssignDefaultToField(UPDMJunction, "WALLTHICKNESSUNITS", 1, "0;1;2;3;4;5;6;7;8;10;11;13;18", False)
    logging.debug("     Successfully Added 20 Default Values to Junction")

    #########################################
    # Defining GAS LINE
    #########################################
    logging.debug("Setting up the LINE class...")


    #Adding Attributes to Gas Line featureclass
    logging.debug("     Adding 42 Attributes to Line")
    arcpy.management.AddField(UPDMLine, "COMMODITY", "SHORT", None, None, None, "Commodity", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "ACTUALINTERNALDIAMETER", "DOUBLE", 16, 5, None, "Actual Internal Diameter", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "COVERTYPE", "TEXT", None, None, 20, "Cover Type", "NULLABLE", "NON_REQUIRED", None)  
    arcpy.management.AddField(UPDMLine, "ORIGINALCOVERDEPTH", "DOUBLE", 16, 5, None, "Original Cover Depth", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "DEPTHUNITS", "SHORT", None, None, None, "Depth Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "PRIMARYCOATINGTYPE", "TEXT", None, None, 30, "Primary Coating Type", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "PRIMARYCOATINGAPPDATE", "DATE", None, None, None, "Primary Coating App Date", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "SECONDARYCOATINGTYPE", "TEXT", None, None, 30, "Secondary Coating Type", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "SECONDARYCOATINGAPPDATE", "DATE", None, None, None, "Secondary Coating App Date", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "TERTIARYCOATINGTYPE", "TEXT", None, None, 30, "Tertiary Coating Type", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("      Set 1 Done")
    arcpy.management.AddField(UPDMLine, "TERTIARYCOATINGAPPDATE", "DATE", None, None, None, "Tertiary Coating App Date", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "SMYS", "DOUBLE", 16, 5, None, "SMYS", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "PERCENTSMYS", "DOUBLE", 16, 5, None, "Percent SMYS", "NULLABLE", "NON_REQUIRED", None)    
    arcpy.management.AddField(UPDMLine, "DATERETIRED", "DATE", None, None, None, "Retired Date", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "CRITICALINDICATOR", "SHORT", None, None, None, "Critical Indicator", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "STUBBED", "SHORT", None, None, None, "Stubbed", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "DESIGNCLASS", "SHORT", None, None, None, "Design Class", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "SPATIALACCURACY", "SHORT", None, None, None, "Spatial Accuracy", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "PIPETESTDATE", "DATE", None, None, None, "Pipe Test Date", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "PIPETESTDURATION", "DOUBLE", 16, 5, None, "Pipe Test Duration", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("      Set 2 Done")
    arcpy.management.AddField(UPDMLine, "PIPETESTPRESSURE", "DOUBLE", 16, 5, None, "Pipe Test Pressure", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "PIPETESTTYPE", "TEXT", None, None, 30, "Pipe Test Type", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "DESIGNPRESSUREENTERED", "DOUBLE", 16, 5, None, "Design Pressure Entered", "NULLABLE", "NON_REQUIRED", None)    
    arcpy.management.AddField(UPDMLine, "DESIGNPRESSURESTANDARD", "DOUBLE", 16, 5, None, "Design Pressure Standard", "NULLABLE", "NON_REQUIRED", None)    
    arcpy.management.AddField(UPDMLine, "DESIGNPRESSUREUNITS", "SHORT", None, None, None, "Design Pressure Unit of Measure", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "DESIGNFACTOR", "DOUBLE", 16, 5, None, "Design Factor", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "COLOR", "TEXT", None, None, 15, "Color", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "DEPTH", "DOUBLE", 10, 5, None, "Depth", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "WIRESIZE", "SHORT", None, None, None, "Wire Size", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "GAUGE", "SHORT", None, None, None, "Gauge", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("      Set 3 Done")
    arcpy.management.AddField(UPDMLine, "WIREMATERIAL", "TEXT", None, None, 20, "Wire Material", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "SERVICESUMMARYID", "GUID", None, None, None, "Service Summary ID", "NULLABLE", "NON_REQUIRED", None)     
    arcpy.management.AddField(UPDMLine, "WORKORDERID", "TEXT", None, None, 38, "Work OrderID", "NULLABLE", "NON_REQUIRED", None) 
    arcpy.management.AddField(UPDMLine, "PROJECTNUMBER", "TEXT", None, None, 38, "Project Number", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "LEGACYID", "TEXT", None, None, 38, "LegacyID", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "ASSETUNITCODE", "TEXT", None, None, 38, "Asset Unit Code", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "PRESSUREUNITS", "SHORT", None, None, None, "Pressure Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "MAOPUNITS", "SHORT", None, None, None, "MAOP Unit of Measure", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "LENGTHUNITS", "SHORT", None, None, None, "Length Units", "NULLABLE", "NON_REQUIRED", None)   
    arcpy.management.AddField(UPDMLine, "DIAMETERUNITS", "SHORT", None, None, None, "Diameter Units", "NULLABLE", "NON_REQUIRED", None)   
    logging.debug("      Set 4 Done")
    arcpy.management.AddField(UPDMLine, "PIPESURFACEAREAUNITS", "SHORT", None, None, None, "Pipe Surface Area Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMLine, "PIPEVOLUMEUNITS", "SHORT", None, None, None, "Pipe Volume Units", "NULLABLE", "NON_REQUIRED", None)          

    logging.debug("     Successfully Added 42 Attributes to Line")
    
    #Add Domains to Gas Line Featureclass
    logging.debug("     Adding 71 Domains to Line")
    arcpy.management.AssignDomainToField(UPDMLine, "LIFECYCLESTATUS", "UPDM_LifeCycleStatus", None)       
    arcpy.management.AssignDomainToField(UPDMLine, "LIFECYCLESTATUS", "UPDM_LifeCycleStatus", "0;1;2;3;4;5;6;7;8;9")
    arcpy.management.AssignDomainToField(UPDMLine, "REGULATORYTYPE", "UPDM_RegulatoryType", None)
    arcpy.management.AssignDomainToField(UPDMLine, "REGULATORYTYPE", "UPDM_RegulatoryType", "0;1;2;3;4;5;6;7;8;9")
    arcpy.management.AssignDomainToField(UPDMLine, "COMMODITY", "UPDM_CommodityType", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "COVERTYPE", "UPDM_SurfaceType", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "LENGTHUNITS", "UPDM_UnitsForLength", "0;1;2;3;4;5;6;7;8;9")
    arcpy.management.AssignDomainToField(UPDMLine, "LENGTHSOURCE", "UPDM_LengthSource", None)
    arcpy.management.AssignDomainToField(UPDMLine, "LENGTHSOURCE", "UPDM_LengthSource", "0;1;2;3;4;5;6;7;8;9")
    arcpy.management.AssignDomainToField(UPDMLine, "NOMINALDIAMETER", "UPDM_LineDiameterUnknown", "0: Unknown")
    arcpy.management.AssignDomainToField(UPDMLine, "NOMINALDIAMETER", "UPDM_LineDiameterServices", "1: Service Pipe")   
    arcpy.management.AssignDomainToField(UPDMLine, "NOMINALDIAMETER", "UPDM_LineDiameterDistribution", "2: Distribution Pipe")
    logging.debug("       Set 1 Done")     
    arcpy.management.AssignDomainToField(UPDMLine, "NOMINALDIAMETER", "UPDM_LineDiameterTransmission", "3: Transmission Pipe")
    arcpy.management.AssignDomainToField(UPDMLine, "NOMINALDIAMETER", "UPDM_LineDiameterGathering", "4: Gathering Pipe")
    arcpy.management.AssignDomainToField(UPDMLine, "NOMINALDIAMETER", "UPDM_LineDiameterStationPipe", "5: Station Pipe")
    arcpy.management.AssignDomainToField(UPDMLine, "NOMINALDIAMETER", "UPDM_LineDiameterCustomPipe", "6: Customer Pipe")
    arcpy.management.AssignDomainToField(UPDMLine, "NOMINALDIAMETER", "UPDM_LineDiameterRiserPipe", "7: Riser Pipe") 
    arcpy.management.AssignDomainToField(UPDMLine, "DIAMETERUNITS", "UPDM_UnitsForDiameter", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "OPERATINGPRESSURE", "UPDM_PressureRange", None)
    arcpy.management.AssignDomainToField(UPDMLine, "OPERATINGPRESSURE", "UPDM_PressureRange", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "PRESSUREUNITS", "UPDM_UnitsForPressure", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "MAOPDESIGN", "UPDM_PressureRange", None)
    arcpy.management.AssignDomainToField(UPDMLine, "MAOPDESIGN", "UPDM_PressureRange", "0;1;2;3;4;5;6;7")
    logging.debug("       Set 2 Done")     
    arcpy.management.AssignDomainToField(UPDMLine, "MAOPTEST", "UPDM_PressureRange", None)
    arcpy.management.AssignDomainToField(UPDMLine, "MAOPTEST", "UPDM_PressureRange", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "MAOPRECORD", "UPDM_PressureRange", None)
    arcpy.management.AssignDomainToField(UPDMLine, "MAOPRECORD", "UPDM_PressureRange", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "MAOPUNITS", "UPDM_UnitsForPressure", "0;1;2;3;4;5;6;7")  
    arcpy.management.AssignDomainToField(UPDMLine, "SPATIALACCURACY", "UPDM_SpatialAccuracy", "0;1;2;3;4;5;6;7;8;9")
    arcpy.management.AssignDomainToField(UPDMLine, "DEPTHUNITS", "UPDM_UnitsForDepth", "0;1;2;3;4;5;6;7;8;9")
    arcpy.management.AssignDomainToField(UPDMLine, "MANUFACTURER", "UPDM_ASTMManufacturer", None)
    arcpy.management.AssignDomainToField(UPDMLine, "MANUFACTURER", "UPDM_ASTMManufacturer", "0;1;2;3;4;5;6;7;8;9")    
    arcpy.management.AssignDomainToField(UPDMLine, "MANUFACTURECOMPONENT", "UPDM_ASTMPipeType", "1;2;4;6")
    logging.debug("       Set 3 Done")     
    arcpy.management.AssignDomainToField(UPDMLine, "MANUFACTURECOMPONENT", "UPDM_ASTMTransPipeType", "3: Transmission Pipe")
    arcpy.management.AssignDomainToField(UPDMLine, "MANUFACTURECOMPONENT", "UPDM_ASTMStationPipeType", "5: Station Pipe")    
    arcpy.management.AssignDomainToField(UPDMLine, "MANUFACTURECOMPONENT", "UPDM_ASTMRiserPipeType", "7: Riser Pipe")
    arcpy.management.AssignDomainToField(UPDMLine, "MATERIAL", "UPDM_ASTMPipeMaterial", "1;2;4;6")    
    arcpy.management.AssignDomainToField(UPDMLine, "MATERIAL", "UPDM_ASTMTransPipeMaterial", "3: Transmission Pipe")   
    arcpy.management.AssignDomainToField(UPDMLine, "MATERIAL", "UPDM_ASTMStationPipeMaterial", "5: Station Pipe")
    arcpy.management.AssignDomainToField(UPDMLine, "MATERIAL", "UPDM_ASTMRiserPipeMaterial", "7: Riser Pipe")
    arcpy.management.AssignDomainToField(UPDMLine, "PINCHABLE", "UPDM_YesNoIndicator", None)
    arcpy.management.AssignDomainToField(UPDMLine, "PINCHABLE", "UPDM_YesNoIndicator", "0;1;2;3;4;5;6;7;8;9")
    arcpy.management.AssignDomainToField(UPDMLine, "TRACERWIRE", "UPDM_YesNoIndicator", None)
    logging.debug("       Set 4 Done")    
    arcpy.management.AssignDomainToField(UPDMLine, "TRACERWIRE", "UPDM_YesNoIndicator", "0;1;2;3;4;5;6;7;8;9")
    arcpy.management.AssignDomainToField(UPDMLine, "WARNINGTAPE", "UPDM_YesNoIndicator", None)     
    arcpy.management.AssignDomainToField(UPDMLine, "WARNINGTAPE", "UPDM_YesNoIndicator", "0;1;2;3;4;5;6;7;8;9")
    arcpy.management.AssignDomainToField(UPDMLine, "PIPESURFACEAREAUNITS", "UPDM_UnitsForSurfaceArea", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "PIPEVOLUMEUNITS", "UPDM_UnitsForVolume", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "WALLTHICKNESS", "UPDM_PipeWallThickness", "0;1;2;4;6")    
    arcpy.management.AssignDomainToField(UPDMLine, "WALLTHICKNESS", "UPDM_TransPipeWallThickness", "3: Transmission Pipe")    
    arcpy.management.AssignDomainToField(UPDMLine, "WALLTHICKNESS", "UPDM_StationPipeWallThickness", "5: Station Pipe")
    arcpy.management.AssignDomainToField(UPDMLine, "WALLTHICKNESS", "UPDM_RiserPipeWallThickness", "7: Riser Pipe")     
    arcpy.management.AssignDomainToField(UPDMLine, "INSTALLATIONMETHOD", "UPDM_LineInstallMethod", "0;1;2;3;4;5;6;7")
    logging.debug("       Set 5 Done")    
    arcpy.management.AssignDomainToField(UPDMLine, "LONGSEAMTYPE", "UPDM_LongSeamType", None)
    arcpy.management.AssignDomainToField(UPDMLine, "LONGSEAMTYPE", "UPDM_LongSeamType", "0;1;2;3;4;5;6;7")      
    arcpy.management.AssignDomainToField(UPDMLine, "SEAMTYPE", "UPDM_SeamType", None)
    arcpy.management.AssignDomainToField(UPDMLine, "SEAMTYPE", "UPDM_SeamType", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "LONGITUDINALSEAMORIENTATION", "UPDM_SeamOrientation", None)
    arcpy.management.AssignDomainToField(UPDMLine, "LONGITUDINALSEAMORIENTATION", "UPDM_SeamOrientation", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "PRIMARYCOATINGTYPE", "UPDM_LineCoatingType", "0;1;2;3;4;5;6;7")      
    arcpy.management.AssignDomainToField(UPDMLine, "SECONDARYCOATINGTYPE", "UPDM_LineCoatingType", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "TERTIARYCOATINGTYPE", "UPDM_LineCoatingType", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "OUTSIDEDIAMETER", "UPDM_PipeOutsideDiameter", "0;1;2;4;6")
    logging.debug("       Set 6 Done")     
    arcpy.management.AssignDomainToField(UPDMLine, "OUTSIDEDIAMETER", "UPDM_TransPipeOutsideDiameter", "3: Transmission Pipe")
    arcpy.management.AssignDomainToField(UPDMLine, "OUTSIDEDIAMETER", "UPDM_StationPipeOutsideDiameter", "5: Station Pipe")
    arcpy.management.AssignDomainToField(UPDMLine, "OUTSIDEDIAMETER", "UPDM_RiserPipeOutsideDiameter", "7: Riser Pipe")   
    arcpy.management.AssignDomainToField(UPDMLine, "DESIGNCLASS", "UPDM_LineDesignClass", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "DESIGNPRESSUREUNITS", "UPDM_UnitsForPressure", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "PIGGABLE", "UPDM_YesNoIndicator", "0;1;2;3;4;5;6")
    arcpy.management.AssignDomainToField(UPDMLine, "STUBBED", "UPDM_YesNoIndicator", "0;1;2;3;4;5;6;7")
    arcpy.management.AssignDomainToField(UPDMLine, "PIPETESTTYPE", "UPDM_LinePipeTestType", "0;1;2;3;4;5;6;7")          
    arcpy.management.AssignDomainToField(UPDMLine, "COLOR", "UPDM_CPColorCode", "8: Bond Wire")
    arcpy.management.AssignDomainToField(UPDMLine, "GAUGE", "UPDM_CPRectifierCableGuage", "9: Rectifier Cable")
    logging.debug("       Set 7 Done")     
    arcpy.management.AssignDomainToField(UPDMLine, "CRITICALINDICATOR", "UPDM_YesNoIndicator", "0;1;2;3;4;5;6;7;8;9")

    logging.debug("     Successfully Added 71 Domains to Line")

    #Add Default Values to Gas Line Featureclass
    logging.debug("     Adding 53 Default Values to Line")
    
    arcpy.management.AssignDefaultToField(UPDMLine, "LIFECYCLESTATUS", 3, "0;1;2;3;4;5;6;7;8;9", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "REGULATORYTYPE", 0, "0;1;2;3;4;5;6;7;8;9", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "COMMODITY", 1, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "COVERTYPE", 0, "0;1;2;3;4;5;6;7", False)    
    arcpy.management.AssignDefaultToField(UPDMLine, "LENGTHUNITS", 1, "0;1;2;3;4;5;6;7;8;9", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "LENGTHSOURCE", 0, "0;1;2;3;4;5;6;7", False)   
    arcpy.management.AssignDefaultToField(UPDMLine, "NOMINALDIAMETER", 0, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "DIAMETERUNITS", 1, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "PRESSUREUNITS", 1, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "SMYS", 0, "0;1;2;3;4;5;6;7", False)
    logging.debug("       Set 2 Done")    
    arcpy.management.AssignDefaultToField(UPDMLine, "PERCENTSMYS", 0, "0;1;2;3;4;5;6;7", False)    
    arcpy.management.AssignDefaultToField(UPDMLine, "OPERATINGPRESSURE", 0, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "MAOPDESIGN", 0, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "MAOPTEST", 0, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "MAOPRECORD", 0, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "MAOPUNITS", 1, "0;1;2;3;4;5;6;7", False)   
    arcpy.management.AssignDefaultToField(UPDMLine, "MANUFACTURER", "0", "0;1;2;3;4;5;6;7;8;9", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "MANUFACTURECOMPONENT", "0", "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "MATERIAL", "UN", "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "PINCHABLE", 0, "0;1;2;4;6", False)
    logging.debug("       Set 3 Done")    
    arcpy.management.AssignDefaultToField(UPDMLine, "PINCHABLE", 2, "3;5;7;8;9", False)   
    arcpy.management.AssignDefaultToField(UPDMLine, "SPATIALACCURACY", 0, "0;1;2;3;4;5;6;7;8;9", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "WARNINGTAPE", 0, "0;5;7;8;9", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "WARNINGTAPE", 1, "1;2;3;4;6", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "TRACERWIRE", 0, "0;5;7;8;9", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "TRACERWIRE", 1, "1;2;3;4;6", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "DEPTHUNITS", 2, "0;1;2;3;4;5;6;7;8;9", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "DESIGNPRESSUREUNITS", 1, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "DESIGNCLASS", 0, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "PIGGABLE", 0, "0;1;2;3;4;5;6;7", False)
    logging.debug("       Set 4 Done")    
    arcpy.management.AssignDefaultToField(UPDMLine, "STUBBED", 0, "0;1;2;3;4;5;6;7", False)     
    arcpy.management.AssignDefaultToField(UPDMLine, "DESIGNCLASS", 0, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "LONGSEAMTYPE", 0, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "SEAMTYPE", 0, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "LONGITUDINALSEAMORIENTATION", 0, "0;1;2;3;4;5;6;7", False) 
    arcpy.management.AssignDefaultToField(UPDMLine, "INSTALLATIONMETHOD", 6, "5;7", False)    
    arcpy.management.AssignDefaultToField(UPDMLine, "INSTALLATIONMETHOD", 1, "1;2;3;4;6", False)   
    arcpy.management.AssignDefaultToField(UPDMLine, "PIPESURFACEAREA", 0, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "PIPEVOLUME", 0, "0;1;2;3;4;5;6;7", False)    
    arcpy.management.AssignDefaultToField(UPDMLine, "PIPESURFACEAREAUNITS", 1, "0;1;2;3;4;5;6;7", False)
    logging.debug("       Set 5 Done")    
    arcpy.management.AssignDefaultToField(UPDMLine, "PIPEVOLUMEUNITS", 1, "0;1;2;3;4;5;6;7", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "GAUGE", 0, "9: Rectifier Cable", False)
    arcpy.management.AssignDefaultToField(UPDMLine, "CRITICALINDICATOR", 0, "0;1;2;3;4;5;6;7;8;9", False)
    
    logging.debug("     Successfully Added 53 Default Values to Line")
    

    ###############################################################
    # Defining Gas STRUCTURE JUNCTION
    ###############################################################
    logging.debug("Setting up the STRUCTURE JUNCTION class...")

    #Adding Attributes to Gas Structure Junction featureclass
    logging.debug("     Adding 15 Attributes to Structure Junction")
    
    arcpy.management.AddField(structureJunction, "BARCODE", "TEXT", None, None, 16, "Barcode", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureJunction, "MANUFACTURER", "TEXT", None, None, 2, "Manufacturer", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureJunction, "MANUFACTURERLOTNO", "LONG", None, None, None, "Manufacturer Lot#", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureJunction, "MANUFACTUREDATE", "DATE", None, None, None, "Manufacture Date", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureJunction, "MANUFACTURECOMPONENT", "TEXT", None, None, 2, "Manufacturer Component", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureJunction, "MATERIAL", "TEXT", None, None, 2, "Material", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureJunction, "COMMENTS", "TEXT", None, None, 255, "Comments", "NULLABLE", "NON_REQUIRED", None)   
    arcpy.management.AddField(structureJunction, "SPATIALACCURACY", "SHORT", None, None, None, "Spatial Accuracy", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureJunction, "GPSX", "DOUBLE", 16, 8, None, "GPS X", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureJunction, "GPSY", "DOUBLE", 16, 8, None, "GPS Y", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("       Set 1 Done")     
    arcpy.management.AddField(structureJunction, "GPSZ", "DOUBLE", 16, 8, None, "GPS Z", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureJunction, "WORKORDERID", "TEXT", None, None, 38, "Work OrderID", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureJunction, "LEGACYID", "TEXT", None, None, 38, "LegacyID", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureJunction, "PROJECTNUMBER", "TEXT", None, None, 38, "Project Number", "NULLABLE", "NON_REQUIRED", None)   
    arcpy.management.AddField(structureJunction, "ASSETUNITCODE", "TEXT", None, None, 38, "Asset Unit Code", "NULLABLE", "NON_REQUIRED", None)

    logging.debug("     Successfully Added 15 Attributes to Structure Junction")

    #Add Domains to Gas Structure Junction Featureclass
    logging.debug("     Adding 18 Domains to Structure Junction")
    arcpy.management.AssignDomainToField(structureJunction, "LIFECYCLESTATUS", "UPDM_LifeCycleStatus", None)
    arcpy.management.AssignDomainToField(structureJunction, "LIFECYCLESTATUS", "UPDM_LifeCycleStatus", "0;1;2;3;4;5;6")
    arcpy.management.AssignDomainToField(structureJunction, "MANUFACTURER", "UPDM_StrJunctMufflerManuf", "1")
    arcpy.management.AssignDomainToField(structureJunction, "MANUFACTURER", "UPDM_StrJunctAnchorManuf", "2")
    arcpy.management.AssignDomainToField(structureJunction, "MANUFACTURER", "UPDM_StrJunctPipeSuppManuf", "3")
    arcpy.management.AssignDomainToField(structureJunction, "MANUFACTURER", "UPDM_StrJunctPipeHangManuf", "4")
    arcpy.management.AssignDomainToField(structureJunction, "MANUFACTURER", "UPDM_StrJunctEndConnectManuf", "5")
    arcpy.management.AssignDomainToField(structureJunction, "MANUFACTURECOMPONENT", "UPDM_StrJunctMufflerManComp", "1")
    arcpy.management.AssignDomainToField(structureJunction, "MANUFACTURECOMPONENT", "UPDM_StrJunctAnchorManComp", "2")
    arcpy.management.AssignDomainToField(structureJunction, "MANUFACTURECOMPONENT", "UPDM_StrJunctPipeSuppManComp", "3")
    logging.debug("       Set 1 Done")      
    arcpy.management.AssignDomainToField(structureJunction, "MANUFACTURECOMPONENT", "UPDM_StrJunctPipeHangManComp", "4")
    arcpy.management.AssignDomainToField(structureJunction, "MANUFACTURECOMPONENT", "UPDM_StrJunctEndConnectManComp", "5")
    arcpy.management.AssignDomainToField(structureJunction, "MATERIAL", "UPDM_StrJunctMufflerMaterial", "1")
    arcpy.management.AssignDomainToField(structureJunction, "MATERIAL", "UPDM_StrJunctAnchorMaterial", "2")
    arcpy.management.AssignDomainToField(structureJunction, "MATERIAL", "UPDM_StrJunctPipeSuppMaterial", "3")
    arcpy.management.AssignDomainToField(structureJunction, "MATERIAL", "UPDM_StrJunctPipeHangMaterial", "4")
    arcpy.management.AssignDomainToField(structureJunction, "MATERIAL", "UPDM_StrJunctEndConnectMaterial", "5")
    arcpy.management.AssignDomainToField(structureJunction, "SPATIALACCURACY", "UPDM_SpatialAccuracy", "0;1;2;3;4;5;6")
    logging.debug("     Successfully Added 18 Domain to Structure Junction")

    #Add Default Values to Structure Junction Featureclass
    logging.debug("     Adding 4 Default Values to Structure Junction")
    arcpy.management.AssignDefaultToField(structureJunction, "MANUFACTURER", "0", "0;1;2;3;4;5", False)
    arcpy.management.AssignDefaultToField(structureJunction, "MANUFACTURECOMPONENT", "0", "0;1;2;3;4;5", False)
    arcpy.management.AssignDefaultToField(structureJunction, "MATERIAL", "0", "0;1;2;3;4;5", False)
    arcpy.management.AssignDefaultToField(structureJunction, "SPATIALACCURACY", 0, "0;1;2", False)
    logging.debug("     Successfully Added 4 Default value to Structure Junction")

    ##############################################################
    # Defining Gas STRUCTURE LINE
    ##############################################################
    logging.debug("Setting up the STRUCTURE LINE class...")

    #Adding Attributes to Gas Structure Edge featureclass
    logging.debug("     Adding 14 Attribute to Structure Line")
    arcpy.management.AddField(structureEdge, "MATERIAL", "TEXT", None, None, 2, "Material", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureEdge, "DIAMETERUNITS", "SHORT", None, None, None, "Diameter Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureEdge, "COATINGTYPE", "TEXT", None, None, 30, "Coating Type", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureEdge, "ENDCONNECTION", "TEXT", None, None, 15, "End Connection", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureEdge, "FILLDATE", "DATE", None, None, None, "Fill Date", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureEdge, "FILLCAPACITY", "LONG", None, None, None, "Fill Capacity", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureEdge, "PROTECTIONTYPE", "TEXT", None, None, 20, "Protection Type", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureEdge, "REASON", "TEXT", None, None, 20, "Reason", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureEdge, "LENGTHUNITS", "SHORT", None, None, None, "Length Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureEdge, "SPATIALACCURACY", "SHORT", None, None, None, "Spatial Accuracy", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("      Set 1 Done")
    arcpy.management.AddField(structureEdge, "LEGACYID", "TEXT", None, None, 38, "Reason", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureEdge, "WORKORDERID", "TEXT", None, None, 38, "Work OrderID", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureEdge, "PROJECTNUMBER", "TEXT", None, None, 38, "Project Number", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureEdge, "ASSETUNITCODE", "TEXT", None, None, 38, "Asset Unit Code", "NULLABLE", "NON_REQUIRED", None)

    logging.debug("     Successfully Added 14 Attribute to Structure Line")

    #Add Domains to Structure Line Featureclass
    logging.debug("     Adding 18 Domains to Structure Line")
    arcpy.management.AssignDomainToField(structureEdge, "LIFECYCLESTATUS", "UPDM_LifeCycleStatus", None)    
    arcpy.management.AssignDomainToField(structureEdge, "LIFECYCLESTATUS", "UPDM_LifeCycleStatus", "0;1")
    arcpy.management.AssignDomainToField(structureEdge, "DIAMETERUNITS", "UPDM_UnitsForDiameter", None)
    arcpy.management.AssignDomainToField(structureEdge, "DIAMETERUNITS", "UPDM_UnitsForDiameter", "0;1")
    arcpy.management.AssignDomainToField(structureEdge, "FILLTYPE", "UPDM_PipeCasingFillType", None)
    arcpy.management.AssignDomainToField(structureEdge, "FILLTYPE", "UPDM_PipeCasingFillType", "0;1")
    arcpy.management.AssignDomainToField(structureEdge, "CASINGDIAMETER", "UPDM_StrLinePCasingDiameter", None)
    arcpy.management.AssignDomainToField(structureEdge, "CASINGDIAMETER", "UPDM_StrLinePCasingDiameter", "0;1")    
    arcpy.management.AssignDomainToField(structureEdge, "VENTEDINDICATOR", "UPDM_YesNoIndicator", "0;1")
    arcpy.management.AssignDomainToField(structureEdge, "INSERTINDICATOR", "UPDM_YesNoIndicator", "0;1")
    logging.debug("      Set 1 Done")    
    arcpy.management.AssignDomainToField(structureEdge, "LENGTHUNITS", "UPDM_UnitsForLength", "0;1")
    arcpy.management.AssignDomainToField(structureEdge, "LENGTHSOURCE", "UPDM_LengthSource", None)
    arcpy.management.AssignDomainToField(structureEdge, "LENGTHSOURCE", "UPDM_LengthSource", "0;1")
    arcpy.management.AssignDomainToField(structureEdge, "MATERIAL", "UPDM_ASTMTransPipeMaterial", None)
    arcpy.management.AssignDomainToField(structureEdge, "MATERIAL", "UPDM_ASTMTransPipeMaterial", "0;1")
    arcpy.management.AssignDomainToField(structureEdge, "COATINGTYPE", "UPDM_LineCoatingType", None)
    arcpy.management.AssignDomainToField(structureEdge, "COATINGTYPE", "UPDM_LineCoatingType", "0;1")
    arcpy.management.AssignDomainToField(structureEdge, "SPATIALACCURACY", "UPDM_SpatialAccuracy", "0;1")
    logging.debug("     Successfully Added 18 Domains to Structure Line")

    #Add Default values to Structure Line Featureclass
    logging.debug("     Adding 10 Default values to Structure Line")
    arcpy.management.AssignDefaultToField(structureEdge, "LIFECYCLESTATUS", 3, "0;1", False)
    arcpy.management.AssignDefaultToField(structureEdge, "DIAMETERUNITS", 1, "0;1", False)
    arcpy.management.AssignDefaultToField(structureEdge, "VENTEDINDICATOR", 0, "0;1", False)
    arcpy.management.AssignDefaultToField(structureEdge, "INSERTINDICATOR", 0, "0;1", False)
    arcpy.management.AssignDefaultToField(structureEdge, "LENGTHUNITS", 1, "0;1", False)
    arcpy.management.AssignDefaultToField(structureEdge, "FILLTYPE", 0, "0;1", False)
    arcpy.management.AssignDefaultToField(structureEdge, "CASINGDIAMETER", 0, "0;1", False)
    arcpy.management.AssignDefaultToField(structureEdge, "MATERIAL", 0, "0;1", False)
    arcpy.management.AssignDefaultToField(structureEdge, "COATINGTYPE", 0, "0;1", False)
    arcpy.management.AssignDefaultToField(structureEdge, "SPATIALACCURACY", 0, "0;1", False)
    logging.debug("     Successfully Added 10 Default values to Structure Line")

    ##############################################################
    # Defining Gas STRUCTURE BOUNDARY
    ##############################################################
    logging.debug("Setting up the STRUCTURE BOUNDARY class...")


    #Adding Attributes to Gas Structure Boundary featureclass
    logging.debug("     Adding 16 Attribute to Structure Boundary")
    arcpy.management.AddField(structureBoundary, "STRUCTUREHEIGHT", "DOUBLE", 16, 8, None, "Structure Height", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureBoundary, "STRUCTURELENGTH", "DOUBLE", 16, 8, None, "Structure Length", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureBoundary, "STRUCTUREWIDTH", "DOUBLE", 16, 8, None, "Structure Width", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureBoundary, "STRUCTUREUNITS", "SHORT", None, None, None, "Structure Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureBoundary, "MATERIAL", "TEXT", None, None, 2, "Material", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureBoundary, "COVERDEPTH", "DOUBLE", 16, 8, None, "Cover Depth", "NULLABLE", "NON_REQUIRED", None)    
    arcpy.management.AddField(structureBoundary, "COVERLENGTH", "DOUBLE", 16, 8, None, "Cover Length", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureBoundary, "COVERWIDTH", "DOUBLE", 16, 8, None, "Cover Width", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureBoundary, "COVERUNITS", "SHORT", None, None, None, "Cover Units", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureBoundary, "VENTEDINDICATOR", "SHORT", None, None, None, "Vented Indicator", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("      Set 1 Done")
    arcpy.management.AddField(structureBoundary, "CONTRACTOR", "TEXT", None, None, 50, "Contractor", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureBoundary, "PIPELINEOPERATOR", "TEXT", None, None, 30, "Pipeline Operator", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureBoundary, "PIPELINESTATION", "DOUBLE", 16, 5, None, "Pipeline Station", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureBoundary, "SPATIALACCURACY", "SHORT", None, None, None, "Spatial Accuracy", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureBoundary, "LEGACYID", "TEXT", None, None, 20, "Legacy ID", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureBoundary, "WORKORDERID", "TEXT", None, None, 38, "Work Order ID", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureBoundary, "PROJECTNUMBER", "TEXT", None, None, 38, "Project Number", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(structureBoundary, "ASSETUNITCODE", "TEXT", None, None, 38, "Asset Unit Code", "NULLABLE", "NON_REQUIRED", None)
    logging.debug("     Successfully Added 16 Attribute to Structure Boundary")

    #Add Domains to Structure Boundary Featureclass
    logging.debug("     Adding 6 Domains to Structure Boundary")
    arcpy.management.AssignDomainToField(structureBoundary, "LIFECYCLESTATUS", "UPDM_LifeCycleStatus", None)    
    arcpy.management.AssignDomainToField(structureBoundary, "LIFECYCLESTATUS", "UPDM_LifeCycleStatus", "0;1;2;3")
    arcpy.management.AssignDomainToField(structureBoundary, "STRUCTUREUNITS", "UPDM_UnitsForLength", "0;1;2;3")
    arcpy.management.AssignDomainToField(structureBoundary, "COVERUNITS", "UPDM_UnitsForLength", "0;1;2;3")
    arcpy.management.AssignDomainToField(structureBoundary, "VENTEDINDICATOR", "UPDM_YesNoIndicator", "0;1;2;3")
    arcpy.management.AssignDomainToField(structureBoundary, "SPATIALACCURACY", "UPDM_SpatialAccuracy", "0;1;2;3")
    logging.debug("     Successfully Added 6 Domains to Structure Boundary")

    #Add Default values to Structure Boundary Featureclass
    logging.debug("     Adding 5 Default values to Structure Boundary")
    arcpy.management.AssignDefaultToField(structureBoundary, "LIFECYCLESTATUS", 3, "0;1;2;3", False)
    arcpy.management.AssignDefaultToField(structureBoundary, "STRUCTUREUNITS", 1, "0;1;2", False)
    arcpy.management.AssignDefaultToField(structureBoundary, "COVERUNITS", 1, "0;1;2", False)
    arcpy.management.AssignDefaultToField(structureBoundary, "VENTEDINDICATOR", 0, "0;1;2", False)
    arcpy.management.AssignDefaultToField(structureBoundary, "SPATIALACCURACY", 0, "0;1;2", False)
    logging.debug("     Successfully Added 5 Default values to Structure Boundary")

    #########################################
    #Defining PipeSystemSERVICESUMMARY
    #########################################
    logging.debug("Setting up the SERVICESUMMARY class...")
    ServiceSummaryName = DomainName + "ServiceSummary"
    arcpy.management.CreateFeatureclass(featureDatasetPath, ServiceSummaryName, "POLYLINE", None, "ENABLED", "ENABLED", None, None, 0, 0, 0)

    #Adding attributes to UPDMServiceSummary Featureclass
    arcpy.management.AddField(UPDMServiceSummary, "LIFECYCLESTATUS", "SHORT", None, None, None, "LifeCycleStatus", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMServiceSummary, "BRANCH", "SHORT", None, None, None, "Branch", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMServiceSummary, "SERVICELENGTH", "DOUBLE", 10, 4, None, "Service Length", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMServiceSummary, "MATERIAL", "LONG", None, None, None, "Material", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMServiceSummary, "DIAMETER", "DOUBLE", 10, 5, None, "Diameter", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddField(UPDMServiceSummary, "INSTALLATIONDATE", "DATE", None, None, None, "Install Date", "NULLABLE", "NON_REQUIRED", None)
    arcpy.management.AddGlobalIDs(UPDMServiceSummary)

    #Add Domains to ServiceSummary Featureclass
    arcpy.management.AssignDomainToField(UPDMServiceSummary, "LIFECYCLESTATUS", "UPDM_LifeCycleStatus", None)
    arcpy.management.AssignDomainToField(UPDMServiceSummary, "BRANCH", "UPDM_YesNoIndicator", None)
    arcpy.management.AssignDomainToField(UPDMServiceSummary, "MATERIAL", "UPDM_PipeMaterial", None)
    arcpy.management.AssignDomainToField(UPDMServiceSummary, "DIAMETER", "UPDM_LineDiameterServices", None)

    #Add Default values to Service Summary Featureclass
    logging.debug("     Adding 1 Default values to Service Summary")
    arcpy.management.AssignDefaultToField(UPDMServiceSummary, "LIFECYCLESTATUS", 3, None, False)
    arcpy.management.AssignDefaultToField(UPDMServiceSummary, "BRANCH", 0, None, False)

    #Add Relationship Class
    arcpy.management.CreateRelationshipClass(UPDMServiceSummary, UPDMLine, UPDMServiceSummaryREL, "SIMPLE", "GASPIPELine", "GASPIPEServiceSummary", "NONE", "ONE_TO_MANY", "NONE", "GlobalID", "SERVICESUMMARYID", None, None)
    logging.debug("    Completed setting up the SERVICESUMMARY class...")

    endTime = timer()
    totalElapsedTime = endTime - startTime
    logging.debug("OVERALL PROCESS COMPLETE in : " + str(totalElapsedTime))

    print(arcpy.GetMessages())
    logging.shutdown()
except:
    logging.exception("Exception", exc_info=sys.exc_info())
    logging.shutdown()
